#ifndef _RANGESET_H_
#define _RANGESET_H_

#ifdef _MSC_VER
#pragma warning (disable:4786)
#endif
#include <map>
#include <limits>

template <class Key, class Compare = std::less<Key>, class Alloc = std::allocator<pair<Key, Key> > >
class range_set_iterator
{
public:

	typedef std::bidirectional_iterator_tag	iterator_category;
	typedef Key								value_type;
	typedef size_t							difference_type;
	typedef Key*							pointer;
	typedef Key&							reference;

private:
	typedef std::map<Key, Key, Compare, Alloc>		REP_TYPE;
	typedef typename REP_TYPE::iterator				REP_TYPE_IT;


public:
	REP_TYPE	*m_rep;
	REP_TYPE_IT m_it;
	Key			m_curr;

	range_set_iterator()
	{
		m_rep = 0;
	}

	range_set_iterator(const range_set_iterator &src)
	{
		m_rep = src.m_rep;
		m_it = src.m_it;
		m_curr = src.m_curr;
	}

	range_set_iterator(REP_TYPE *rep, const REP_TYPE_IT &it, const Key &curr)
	{
		m_rep = rep;
		m_it = it;
		m_curr = curr;
	}

	range_set_iterator &operator=(const range_set_iterator &src)
	{
		m_rep = src.m_rep;
		m_it = src.m_it;
		m_curr = src.m_curr;
		return *this;
	}

	inline bool operator==(const range_set_iterator &rhs)
	{
		return (m_rep == rhs.m_rep && 
			m_it == rhs.m_it &&
			((m_it != m_rep->end() &&
			 m_curr == rhs.m_curr)) ||
			m_it == rhs.m_it);
	}

	inline bool operator!=(const range_set_iterator &rhs)
	{
		return (!(m_rep == rhs.m_rep && 
			m_it == rhs.m_it &&
			((m_it != m_rep->end() &&
			 m_curr == rhs.m_curr)) ||
			m_it == rhs.m_it));
	}

	inline Key &operator*()
	{
		return m_curr;
	}

	inline range_set_iterator &operator++()
	{
		if (m_rep && m_it != m_rep->end()) {
			++m_curr;
			if (!(m_curr < m_it->second)) {
				++m_it;
				if (m_it != m_rep->end()) {
					m_curr = m_it->first;
				}
			}
		}
		return *this;
	}

	inline range_set_iterator &operator--()
	{
		if (m_rep && m_rep->begin() != m_rep->end()) {
			if (m_it == m_rep->end()) {
				--m_it;
				m_curr = m_it->second;
				--m_curr;
			} else {
				--m_curr;
				if (m_curr < m_it->first) {
					if (m_it == m_rep->begin()) {
						++m_curr; // do not go before the start
					} else {
						--m_it;
						m_curr = m_it->second;
						--m_curr;
					}
				}
			}
		}
		return *this;
	}

	range_set_iterator operator++(int)
	{
		range_set_iterator tmp = *this;
		if (m_rep && m_it != m_rep->end()) {
			++m_curr;
			if (!(m_curr < m_it->second)) {
				++m_it;
				if (m_it != m_rep->end()) {
					m_curr = m_it->first;
				}
			}
		}
		return tmp;		
	}

	range_set_iterator &operator--(int)
	{
		range_set_iterator tmp = *this;
		if (m_rep && m_rep->begin() != m_rep->end()) {
			if (m_it == m_rep->end()) {
				--m_it;
				m_curr = m_it->second;
				--m_curr;
			} else {
				--m_curr;
				if (m_curr < m_it->first) {
					if (m_it == m_rep->begin()) {
						++m_curr; // do not go before the start
					} else {
						--m_it;
						m_curr = m_it->second;
						--m_curr;
					}
				}
			}
		}
		return tmp;
	}
};

const int RANGE_SET_RES_INRANGE		= 0;
const int RANGE_SET_RES_ONEBEFORE	= 1;
const int RANGE_SET_RES_ONEAFTER	= 2;
const int RANGE_SET_RES_MERGE		= 3;
const int RANGE_SET_RES_BEFORE		= 4;
const int RANGE_SET_RES_AFTER		= 5;
const int RANGE_SET_RES_MIDDLE		= 6;

// This range_set has an index to allow for log(N) access to the nth element
template <class Key, class Compare = std::less<Key>, class Alloc = std::allocator<std::pair<Key, Key> > >
class range_set
{
	// The representation is a map which maps the start of a range to the end of a range
	// This is a somewhat unusual usage of std::map, but very useful in this case
	// since the starts are sorted ascendingly and hence we can determine in log(N) time in which
	// range a give element is
	typedef std::map<Key, Key, Compare, Alloc>						REP_TYPE;
	typedef typename std::map<Key, Key, Compare, Alloc>::iterator	REP_TYPE_IT;

	// The index is a map which maps the count of elements below this iterator's range end to the iterator
	typedef std::map<int, REP_TYPE_IT>								INDEX_TYPE;
	typedef typename std::map<int, REP_TYPE_IT>::iterator			INDEX_IT;

	REP_TYPE	m_rep;
	INDEX_TYPE	m_index;
	size_t		m_total_size;
	
	long compute_index()
	{
		int n = 0;
		REP_TYPE_IT	it = m_rep.begin(), itend = m_rep.end();
		m_index.clear();
		if (it != itend) {
			n += it->second - it->first;
			m_index.insert(std::make_pair(n, it));
			++it;
		}
		while (it != itend) {
			n += it->second - it->first;
			m_index.insert(std::make_pair(n, it));
			++it;
		}
		return n;
	}

	inline void clear_index()
	{
		if (m_total_size) {
			m_index.clear();
			m_total_size = 0;
		}
	}

	inline REP_TYPE_IT internal_find(const Key val, int *resType)
	{
		REP_TYPE_IT it;

		it = m_rep.lower_bound(val);
		if (it == m_rep.end()) { // after
			if (it == m_rep.begin()) { // empty range
				*resType = RANGE_SET_RES_AFTER;
			} else {
				--it;
				if (!(val < it->second)) { // Greater than last range
					if (!(it->second < val)) { // one after
						*resType = RANGE_SET_RES_ONEAFTER;
					} else { // simply greater than everything else
						++it;
						*resType = RANGE_SET_RES_AFTER;
					}
				} else { // contained in range
					*resType = RANGE_SET_RES_INRANGE;
				}
			}
		} else {
			if (it == m_rep.begin()) { // at start or before
				if (val == it->first) { // at the start
					*resType = RANGE_SET_RES_INRANGE;
				} else {
					Key valplus = val;

					++valplus;
					if (valplus == it->first) {
						*resType = RANGE_SET_RES_ONEBEFORE;
					} else { // Before everything else
						*resType = RANGE_SET_RES_BEFORE;
					}
				}
			} else { // in the middle
				if (it->first == val) { // in there
					*resType = RANGE_SET_RES_INRANGE;
				} else {
					Key valplus = val;
					REP_TYPE_IT itprev = it;

					++valplus;
					--itprev;
					if (valplus == it->first) { // prepend or merge
						if (itprev->second == val) { // merge
							*resType = RANGE_SET_RES_MERGE;
						} else { // one before
							*resType = RANGE_SET_RES_ONEBEFORE;
						}
					} else { // in previous, append or create new one
						if (val < itprev->second) {
							it = itprev;
							*resType = RANGE_SET_RES_INRANGE;
						} else { // append or new one
							if (val == itprev->second) {
								it = itprev;
								*resType = RANGE_SET_RES_ONEAFTER;
							} else { // somewhere in the middle, before 'it'
								*resType = RANGE_SET_RES_MIDDLE;
							}
						}
					}
				}
			}
		}
		return it;
	}

public:
	typedef range_set_iterator<Key, Compare, Alloc> iterator;
	typedef Key										value_type;
	typedef Key										key_type;
	typedef Compare									key_compare;
	typedef Compare									value_compare;
	typedef Key										*pointer;
	typedef Key										&reference;
	typedef Key										&const_reference;
	typedef size_t									size_type;
	typedef int										difference_type;

	range_set() : m_rep(), m_index()
	// Create an empty range_set
	{
		m_total_size = 0;
	}

	range_set(const range_set &src) : m_rep(src.m_rep), m_index(src.m_index)
	// copy constructor
	{
		m_total_size = src.m_total_size;
	}

	range_set(const key_compare& comp) : m_rep(comp), m_index()
	// constructor with a given comparison function
	{
		m_total_size = 0;
	}

	range_set &operator=(const range_set &src)
	// Assignment operator
	{
		m_rep.clear();
		m_rep.insert(src.m_rep.begin(), src.m_rep.end());
		m_index.clear();
		m_index.insert(src.m_index.begin(), src.m_index.end());
		m_total_size = src.m_total_size;
		return *this;
	}

	inline bool operator==(const range_set &rhs)
	// equality function (relies on the representation)
	{
		return m_rep == rhs.m_rep;
	}

	inline bool operator!=(const range_set &rhs)
	// inequality function (relies on the representation)
	{
		return m_rep != rhs.m_rep;
	}

	void swap(range_set &other)
	// swap (relies on the representation)
	{
		m_rep.swap(other.m_rep);
	}

	inline iterator begin()
	// returns an iterator to the start of the range_set
	{
		REP_TYPE_IT it = m_rep.begin();
		if (it != m_rep.end()) {
			return iterator(&m_rep, it, it->first);
		} else {
			// don't dereference the iterator if we are at the end
			return iterator(&m_rep, it, Key());
		}
	}

	inline iterator end()
	// returns an iterator to the end of the range_set
	{
		return iterator(&m_rep, m_rep.end(), Key());
	}

	inline void clear()
	// clears the range_set
	{
		m_rep.clear();
	}

	inline size_type size()
	// returns the number of elements in the range_set
	{
		if (m_total_size) {
			// the cached size is up to date, so just return it
			return m_total_size;
		} else {
			// we have to recalculate the number of elements
			return (m_total_size = compute_index());
		}
	}

	inline size_type max_size()
	// yeah, well, just hope we have enough memory for this
	{
		return std::numeric_limits<size_t>::max;
	}

	inline size_type n_ranges()
	// returns the number of ranges stored internally
	{
		return m_rep.size();
	}

	inline bool empty() const
	{
		return m_rep.empty();
	}

	inline key_compare key_comp() const
	{
		return m_rep.key_comp();
	}

	inline value_compare value_comp() const
	{
		return m_rep.value_comp();
	}

	inline iterator find(const Key val)
	// returns an iterator pointing to a given element
	{
		REP_TYPE_IT it;
		int res;

		it = internal_find(val, &res);
		if (res == RES_INRANGE) {
			return iterator(&m_rep, it, val);
		} else {
			return iterator(&m_rep, m_rep.end(), Key());
		}
	}

	inline iterator get_nth(unsigned int n)
	// returns an iterator to the Nth element (this is O(log(N)) in general, O(N) in the worst case)
	// Note that this function requires the Key to be random-access (i.e. you can add and substract Keys)
	// If the index is not needed, the Key only has to be bidirectional (i.e. you can use ++ and --)
	{
		if (m_total_size == 0) { // Need to compute index
			m_total_size = compute_index();
		}

		INDEX_IT	ind_it;
		REP_TYPE_IT	it;

		++n;
		ind_it = m_index.lower_bound(n);
		if (ind_it == m_index.end()) {
			return iterator(&m_rep, m_rep.end(), Key());
		} else {
			it = ind_it->second;
			if (ind_it == m_index.begin()) {
				return iterator(&m_rep, it, n + it->first - 1);
			} else {
				--ind_it;
				return iterator(&m_rep, it, n - ind_it->first + it->first - 1);
			}
		}
	}

	inline iterator insert(const Key val)
	// The standard insertion function
	// it is a bit complicated by the different possibilities that arise with our representation of the data
	{
		REP_TYPE_IT it, itres;
		int			res;
		Key			other;
		
		it = internal_find(val, &res);
		switch (res) {
		case RANGE_SET_RES_INRANGE:
			// It already exists, so we just return the iterator we got from the internal find
			itres = it;
			break;
		case RANGE_SET_RES_ONEBEFORE:
			// It is just before a range already in the set, so we enlarge that range
			other = it->second;
			// Unfortunately we have to erase the old std::map entry.
			// Using our own red-black tree could improve matters here
			m_rep.erase(it);
			itres = m_rep.insert(std::make_pair(val, other)).first;
			break;
		case RANGE_SET_RES_ONEAFTER:
			// It is directly after a range we already have, so we just increase the end of this range
			// This is possible, since std::map's second element is non const
			++(it->second);
			itres = it;
			break;
		case RANGE_SET_RES_MERGE:
			// It's the element missing in between two ranges, so we merge them
			// It would be better to just increase the second element of the first iterator
			// unfortunately, there may be differences in handling this by different STLs
			// so this is a 'safe way'
			{
				REP_TYPE_IT itprev = it;
				Key			start;

				other = it->second;
				--itprev;
				start = itprev->first;
				m_rep.erase(itprev);
				m_rep.erase(it);
				itres = m_rep.insert(std::make_pair(start, other)).first;
			}
			break;
		case RANGE_SET_RES_MIDDLE:
		case RANGE_SET_RES_BEFORE:
		case RANGE_SET_RES_AFTER:
			// There is no range adjacent, so we can just add a pair
			other = val;
			++other;
			itres = m_rep.insert(std::make_pair(val, other)).first;
			break;
		}
		clear_index();
		return iterator(&m_rep, itres, val);
	}

	inline void quick_insert(const Key val)
	// A quick and dirty insert function
	// It is not safe to use this in general
	// Use this only when you are sure that you will not insert adjacent values
	{
		Key maxval(val);
		++maxval;
		m_rep.insert(std::make_pair(val, maxval));
	}

	inline void erase(const Key val)
	// Erases an element by Key
	{
		REP_TYPE_IT it;
		int			res;

		it = internal_find(val, &res);
		if (res == RES_INRANGE) {
			Key valplus = val;

			++valplus;
			if (it->first == val) {
				if (it->second == valplus) {
					m_rep.erase(it);
				} else {
					Key tmp = it->second;
					m_rep.erase(it);
					m_rep.insert(std::make_pair(valplus, tmp));
				}
			} else {
				if (valplus == it->second) {
					Key tmp = it->first;
					m_rep.erase(it);
					m_rep.insert(std::make_pair(tmp, val));
				} else { // Break a range
					Key tmp = it->second;
					it->second = val;
					m_rep.insert(std::make_pair(valplus, tmp));
				}
			}
			clear_index();
		}
	}

	inline void erase(const iterator &thisit)
	// Erases an element by iterator
	{
		REP_TYPE_IT it = thisit.m_it;
		Key val = thisit.m_curr, valplus;

		valplus = val;
		++valplus;

		if (it->first == val) {
			if (it->second == valplus) {
				m_rep.erase(it);
			} else {
				Key tmp = it->second;
				m_rep.erase(it);
				m_rep.insert(std::make_pair(valplus, tmp));
			}
		} else {
			if (valplus == it->second) {
				Key tmp = it->first;
				m_rep.erase(it);
				m_rep.insert(std::make_pair(tmp, val));
			} else { // Break a range
				Key tmp = it->second;
				it->second = val;
				m_rep.insert(std::make_pair(valplus, tmp));
			}
		}
		clear_index();
	}

	inline void insert_range(const Key &val1, const Key &val2)
	// Inserts a range of values in O(log(N)) time independent of the number of items added
	{
		REP_TYPE_IT itstart, itend;
		int			resstart, resend;
		Key			min, max, tmp;

		tmp = val2;
		--tmp;
		itstart = internal_find(val1, &resstart);
		itend = internal_find(tmp, &resend);
		switch (resstart) {
		case RANGE_SET_RES_INRANGE:
		case RANGE_SET_RES_ONEAFTER:
			min = itstart->first;
			break;
		case RANGE_SET_RES_ONEBEFORE:
		case RANGE_SET_RES_BEFORE:
		case RANGE_SET_RES_AFTER:
		case RANGE_SET_RES_MIDDLE:
			min = val1;
			break;
		case RANGE_SET_RES_MERGE:
			--itstart;
			min = itstart->first;
			break;
		}
		switch (resend) {
		case RANGE_SET_RES_INRANGE:
		case RANGE_SET_RES_ONEBEFORE:
		case RANGE_SET_RES_MERGE:
			max = itend->second;
			++itend;
			break;
		case RANGE_SET_RES_ONEAFTER:
			max = val2;
			++itend;
			break;
		case RANGE_SET_RES_AFTER:
			max = val2;
			if (itend != m_rep.end()) {
				++itend;
			}
			break;
		case RANGE_SET_RES_MIDDLE:
		case RANGE_SET_RES_BEFORE:
			max = val2;
			break;
		}
		m_rep.erase(itstart, itend);
		m_rep.insert(std::make_pair(min, max));
		clear_index();
	}

	inline void quick_insert_range(const Key &val1, const Key &val2)
	// Inserts a range of values in O(log(N)) time
	// This is unsafe unless you know what you are doing
	// Use this only when you are sure that the range does not overlap with anything and is not adjacent to other ranges
	{
		m_rep.insert(std::make_pair(val1, val2));
	}

	inline void erase_range(const Key &val1, const Key &val2)
	// Removes a range of values in O(log(N)) time
	// val1 has to be smaller than val2
	{
		REP_TYPE_IT itstart, itend;
		int			resstart, resend;
		Key			max, tmp;
		bool		start_in_range = false, end_in_range = false;

		if (!(val1 < val2))
			return;

		tmp = val2;
		--tmp;
		itstart = internal_find(val1, &resstart);
		itend = internal_find(tmp, &resend);
		switch (resstart) {
		case RANGE_SET_RES_ONEAFTER:
			++itstart;
			break;
		case RANGE_SET_RES_INRANGE:
			start_in_range = true;
			break;
		}
		switch (resend) {
		case RANGE_SET_RES_ONEAFTER:
			++itend;
			break;
		case RANGE_SET_RES_INRANGE:
			end_in_range = true;
			break;
		}
		// Check whether we have to split up ranges
		if (start_in_range && end_in_range) {
			if (itstart == itend) { // break up a single range
				max = itstart->second;
				if (val1 == itstart->first) { // shrink from left
					m_rep.erase(itstart);
					m_rep.insert(std::make_pair(val2, max));
				} else if (max == val2) { // shrink from right
					itstart->second = val1;
				} else { // break into two
					itstart->second = val1;
					m_rep.insert(std::make_pair(val2, max));
				}
			} else { // Two distinct ranges
				REP_TYPE_IT ittmp;

				if (val1 != itstart->first) {
					itstart->second = val1;
					++itstart;
				}
				if (val2 != itend->second) {
					// This entry is after the one pointed to by itend
					m_rep.insert(std::make_pair(val2, itend->second));
				}
				++itend;
				m_rep.erase(itstart, itend);
			}
		} else if (start_in_range) {
			if (val1 != itstart->first) { 
				itstart->second = val1;
				++itstart;
			}
			m_rep.erase(itstart, itend);
		} else if (end_in_range) {
			if (val2 != itend->first) {
				if (val2 != itend->second) {
					m_rep.insert(std::make_pair(val2, itend->second));
				}
				++itend;
			}
			m_rep.erase(itstart, itend);
		} else {
			m_rep.erase(itstart, itend);
		}
		clear_index();
	}

	friend range_set<Key, Compare, Alloc> *intersect_range(range_set<Key, Compare, Alloc> &src, Key val1, Key val2);
};

template <class Key, class Compare, class Alloc>
range_set<Key, Compare, Alloc> *intersect_range(range_set<Key, Compare, Alloc> &src, const Key &val1, const Key &val2)
// This function returns the intersection of two range_sets
// The caller is responsible for freeing the resulting range_set (it is always non-null)
{
	std::map<Key, Key, Compare, Alloc>::iterator itstart, itend;
	range_set<Key, Compare, Alloc>  *rg = new range_set<Key, Compare, Alloc>();
	int resstart, resend;
	Key tmp = val2, t1, t2;
	bool in_start = false, in_end = false;

	if (val1 >= val2)
		return rg;

	--tmp;
	itstart = src.internal_find(val1, &resstart);
	itend = src.internal_find(tmp, &resend);
	if (resstart == RES_ONEAFTER) {
		++itstart;
	} else if (resstart == RES_INRANGE) {
		in_start = true;
	}

	if (resend == RES_ONEAFTER) {
		++itend;
	} else if (resend == RES_INRANGE) {
		in_end = true;
	}
	while (itstart != itend) {
		if (in_start) {
			rg->m_rep.insert(std::make_pair(val1, itstart->second));
			in_start = false;
		} else {
			rg->m_rep.insert(std::make_pair(t1 = itstart->first, t2 = itstart->second));
		}
		++itstart;
	}
	if (in_end) {
		if (in_start) {
			rg->m_rep.insert(std::make_pair(val1, val2));
		} else {
			rg->m_rep.insert(std::make_pair(t1 = itstart->first, t2 = val2));
		}
	}
	return rg;
}

#endif // RANGESET_H_