#include <iostream>
#include <set>
#include "range_set.h"
#include "utils.h"
#include <conio.h> // for _getch()

using namespace std;

// The std::set we compare against
std::set<int> g_set;

// The range_set
range_set<int> g_rset;

template <class T>
void insert_N(int N, T &s)
{
	// inserts N elements from 1 to N to the container s by using insert
	for (int i = 1; i != N; ++i) {
		s.insert(i);
	}
}

void insertion_speed()
{
	cout << "Testing consecutive inserting speed" << endl;
	cout << "-----------------------------------" << endl;
	CPrecisionTimer ct;
	int N = 1000000;
	ct.Start();
	insert_N(N, g_set);
	cout << "Inserting " << N << " consecutive elements into std::set took: " << ct.Stop() << " seconds" << endl;
	ct.Start();
	insert_N(N, g_rset);
	cout << "Inserting " << N << " consecutive elements into range_set took: " << ct.Stop() << " seconds" << endl;
	g_rset.clear();
	ct.Start();
	g_rset.insert_range(1, N + 1);
    cout << "Inserting " << N << " consecutive elements with insert_range into range_set took: " << ct.Stop() << " seconds" << endl;

	cout << "Memory used by std::set  (approx): " << g_set.size() * (sizeof(int) + sizeof(int *) * 2) + sizeof(g_set) << endl;
	cout << "Memory used by range_set (approx): " << g_rset.n_ranges() * (sizeof(int) + sizeof(int *) * 2) * 2 + sizeof(g_rset) << endl;
	ct.Start();
	g_set.clear();
	cout << "Clearing the std::set took: " << ct.Stop() << " seconds" << endl;
	ct.Start();
	g_rset.clear();
	cout << "Clearing the range_set took: " << ct.Stop() << " seconds" << endl;
}

void random_speed()
// Inserts integers randomly into the container
// The range_set is at a penalty here, since it is not meant to be used this way
{
	cout << "Testing random insertion" << endl;
	cout << "------------------------" << endl;
	int i, N = 10000;

	CPrecisionTimer ct;
	srand(87161);
	ct.Start();
	for (i = 0; i < N / 2; ++i) {
		g_set.insert(rand() % N);
	}
	cout << "Inserting " << N / 2 << " elements randomly into std::set took: " << ct.Stop() << " seconds" << endl;
	srand(87161);
	ct.Start();
	for (i = 0; i < N / 2; ++i) {
		g_rset.insert(rand() % N);
	}
	cout << "Inserting " << N / 2 << " elements randomly into range_set took: " << ct.Stop() << " seconds" << endl;
	g_set.clear();
	g_rset.clear();
}

// The following functions are used to insert the Japanese Unicode characters into the appropriate container
// Overloads for std::set
inline void i_single(std::set<int> *s, int val)
{
	s->insert(val);
}

inline void i_range(std::set<int> *s, int a, int b)
{
	for (int i = a; i != b; ++i) {
		s->insert(i);
	}
}

// Overloads for range_set
inline void i_single(range_set<int> *s, int val)
{
	s->quick_insert(val);
}

inline void i_range(range_set<int> *s, int a, int b)
{
	s->quick_insert_range(a, b);
}

template <class T>
void insert_Japanese_character_set(T *s)
{
	i_range(s, 0x0030, 0x0039 + 1);
	i_range(s, 0x3000, 0x3002 + 1);
	i_range(s, 0x3005, 0x3007 + 1);
	i_range(s, 0x3041, 0x3093 + 1);
	i_range(s, 0x309B, 0x309E + 1);
	i_range(s, 0x30A1, 0x30F6 + 1);
	i_range(s, 0x30FB, 0x30FE + 1);
	i_range(s, 0x3231, 0x3232 + 1);
	i_single(s, 0x3239);
	i_range(s, 0x32A4, 0x32A8 + 1);
	i_range(s, 0x337C, 0x337E + 1);
	i_range(s, 0x4E00, 0x4E01 + 1);
	i_single(s, 0x4E03);
	i_range(s, 0x4E07, 0x4E0B + 1);
	i_range(s, 0x4E0D, 0x4E0E + 1);
	i_range(s, 0x4E10, 0x4E11 + 1);
	i_range(s, 0x4E14, 0x4E19 + 1);
	i_single(s, 0x4E1E);
	i_single(s, 0x4E21);
	i_single(s, 0x4E26);
	i_single(s, 0x4E28);
	i_single(s, 0x4E2A);
	i_single(s, 0x4E2D);
	i_range(s, 0x4E31, 0x4E32 + 1);
	i_single(s, 0x4E36);
	i_range(s, 0x4E38, 0x4E39 + 1);
	i_range(s, 0x4E3B, 0x4E3C + 1);
	i_single(s, 0x4E3F);
	i_range(s, 0x4E42, 0x4E43 + 1);
	i_single(s, 0x4E45);
	i_single(s, 0x4E4B);
	i_range(s, 0x4E4D, 0x4E4F + 1);
	i_range(s, 0x4E55, 0x4E59 + 1);
	i_range(s, 0x4E5D, 0x4E5F + 1);
	i_single(s, 0x4E62);
	i_single(s, 0x4E71);
	i_single(s, 0x4E73);
	i_single(s, 0x4E7E);
	i_single(s, 0x4E80);
	i_single(s, 0x4E82);
	i_range(s, 0x4E85, 0x4E86 + 1);
	i_range(s, 0x4E88, 0x4E8C + 1);
	i_single(s, 0x4E8E);
	i_range(s, 0x4E91, 0x4E92 + 1);
	i_range(s, 0x4E94, 0x4E95 + 1);
	i_range(s, 0x4E98, 0x4E99 + 1);
	i_range(s, 0x4E9B, 0x4E9C + 1);
	i_range(s, 0x4E9E, 0x4EA2 + 1);
	i_range(s, 0x4EA4, 0x4EA6 + 1);
	i_single(s, 0x4EA8);
	i_range(s, 0x4EAB, 0x4EAE + 1);
	i_single(s, 0x4EB0);
	i_single(s, 0x4EB3);
	i_single(s, 0x4EB6);
	i_single(s, 0x4EBA);
	i_range(s, 0x4EC0, 0x4EC2 + 1);
	i_single(s, 0x4EC4);
	i_range(s, 0x4EC6, 0x4EC7 + 1);
	i_range(s, 0x4ECA, 0x4ECB + 1);
	i_range(s, 0x4ECD, 0x4ECF + 1);
	i_range(s, 0x4ED4, 0x4ED9 + 1);
	i_range(s, 0x4EDD, 0x4EDF + 1);
	i_single(s, 0x4EE1);
	i_range(s, 0x4EE3, 0x4EE5 + 1);
	i_range(s, 0x4EED, 0x4EEE + 1);
	i_single(s, 0x4EF0);
	i_single(s, 0x4EF2);
	i_range(s, 0x4EF6, 0x4EF7 + 1);
	i_range(s, 0x4EFB, 0x4EFC + 1);
	i_range(s, 0x4F00, 0x4F01 + 1);
	i_single(s, 0x4F03);
	i_range(s, 0x4F09, 0x4F0A + 1);
	i_range(s, 0x4F0D, 0x4F11 + 1);
	i_single(s, 0x4F1A);
	i_range(s, 0x4F1C, 0x4F1D + 1);
	i_range(s, 0x4F2F, 0x4F30 + 1);
	i_single(s, 0x4F34);
	i_single(s, 0x4F36);
	i_range(s, 0x4F38, 0x4F3A + 1);
	i_range(s, 0x4F3C, 0x4F3D + 1);
	i_single(s, 0x4F43);
	i_range(s, 0x4F46, 0x4F47 + 1);
	i_range(s, 0x4F4D, 0x4F51 + 1);
	i_single(s, 0x4F53);
	i_range(s, 0x4F55, 0x4F57 + 1);
	i_range(s, 0x4F59, 0x4F5E + 1);
	i_single(s, 0x4F69);
	i_range(s, 0x4F6F, 0x4F70 + 1);
	i_single(s, 0x4F73);
	i_range(s, 0x4F75, 0x4F76 + 1);
	i_range(s, 0x4F7B, 0x4F7C + 1);
	i_single(s, 0x4F7F);
	i_single(s, 0x4F83);
	i_single(s, 0x4F86);
	i_single(s, 0x4F88);
	i_range(s, 0x4F8A, 0x4F8B + 1);
	i_single(s, 0x4F8D);
	i_single(s, 0x4F8F);
	i_range(s, 0x4F91, 0x4F92 + 1);
	i_single(s, 0x4F94);
	i_single(s, 0x4F96);
	i_single(s, 0x4F98);
	i_range(s, 0x4F9A, 0x4F9B + 1);
	i_single(s, 0x4F9D);
	i_range(s, 0x4FA0, 0x4FA1 + 1);
	i_single(s, 0x4FAB);
	i_range(s, 0x4FAD, 0x4FAF + 1);
	i_range(s, 0x4FB5, 0x4FB6 + 1);
	i_single(s, 0x4FBF);
	i_range(s, 0x4FC2, 0x4FC4 + 1);
	i_range(s, 0x4FC9, 0x4FCA + 1);
	i_range(s, 0x4FCD, 0x4FCE + 1);
	i_range(s, 0x4FD0, 0x4FD1 + 1);
	i_single(s, 0x4FD4);
	i_range(s, 0x4FD7, 0x4FD8 + 1);
	i_range(s, 0x4FDA, 0x4FDB + 1);
	i_single(s, 0x4FDD);
	i_single(s, 0x4FDF);
	i_single(s, 0x4FE1);
	i_range(s, 0x4FE3, 0x4FE5 + 1);
	i_range(s, 0x4FEE, 0x4FEF + 1);
	i_single(s, 0x4FF3);
	i_range(s, 0x4FF5, 0x4FF6 + 1);
	i_single(s, 0x4FF8);
	i_single(s, 0x4FFA);
	i_range(s, 0x4FFE, 0x4FFF + 1);
	i_range(s, 0x5005, 0x5006 + 1);
	i_single(s, 0x5009);
	i_single(s, 0x500B);
	i_single(s, 0x500D);
	i_single(s, 0x500F);
	i_range(s, 0x5011, 0x5012 + 1);
	i_single(s, 0x5014);
	i_single(s, 0x5016);
	i_range(s, 0x5019, 0x501A + 1);
	i_range(s, 0x501E, 0x501F + 1);
	i_range(s, 0x5021, 0x5026 + 1);
	i_range(s, 0x5028, 0x502D + 1);
	i_single(s, 0x5036);
	i_single(s, 0x5039);
	i_single(s, 0x5040);
	i_range(s, 0x5042, 0x5043 + 1);
	i_range(s, 0x5046, 0x5049 + 1);
	i_range(s, 0x504F, 0x5050 + 1);
	i_range(s, 0x5055, 0x5056 + 1);
	i_single(s, 0x505A);
	i_single(s, 0x505C);
	i_single(s, 0x5065);
	i_single(s, 0x506C);
	i_single(s, 0x5070);
	i_single(s, 0x5072);
	i_range(s, 0x5074, 0x5076 + 1);
	i_single(s, 0x5078);
	i_single(s, 0x507D);
	i_single(s, 0x5080);
	i_single(s, 0x5085);
	i_single(s, 0x508D);
	i_single(s, 0x5091);
	i_single(s, 0x5094);
	i_range(s, 0x5098, 0x509A + 1);
	i_range(s, 0x50AC, 0x50AD + 1);
	i_range(s, 0x50B2, 0x50B5 + 1);
	i_single(s, 0x50B7);
	i_single(s, 0x50BE);
	i_single(s, 0x50C2);
	i_single(s, 0x50C5);
	i_range(s, 0x50C9, 0x50CA + 1);
	i_single(s, 0x50CD);
	i_single(s, 0x50CF);
	i_single(s, 0x50D1);
	i_range(s, 0x50D5, 0x50D6 + 1);
	i_single(s, 0x50D8);
	i_single(s, 0x50DA);
	i_single(s, 0x50DE);
	i_single(s, 0x50E3);
	i_single(s, 0x50E5);
	i_single(s, 0x50E7);
	i_range(s, 0x50ED, 0x50EE + 1);
	i_range(s, 0x50F4, 0x50F5 + 1);
	i_single(s, 0x50F9);
	i_single(s, 0x50FB);
	i_range(s, 0x5100, 0x5102 + 1);
	i_single(s, 0x5104);
	i_single(s, 0x5109);
	i_single(s, 0x5112);
	i_range(s, 0x5114, 0x5116 + 1);
	i_single(s, 0x5118);
	i_single(s, 0x511A);
	i_single(s, 0x511F);
	i_single(s, 0x5121);
	i_single(s, 0x512A);
	i_single(s, 0x5132);
	i_single(s, 0x5137);
	i_range(s, 0x513A, 0x513C + 1);
	i_range(s, 0x513F, 0x5141 + 1);
	i_range(s, 0x5143, 0x514E + 1);
	i_single(s, 0x5150);
	i_single(s, 0x5152);
	i_single(s, 0x5154);
	i_single(s, 0x515A);
	i_single(s, 0x515C);
	i_single(s, 0x5162);
	i_range(s, 0x5164, 0x5165 + 1);
	i_range(s, 0x5168, 0x516E + 1);
	i_single(s, 0x5171);
	i_range(s, 0x5175, 0x5178 + 1);
	i_single(s, 0x517C);
	i_single(s, 0x5180);
	i_single(s, 0x5182);
	i_range(s, 0x5185, 0x5186 + 1);
	i_range(s, 0x5189, 0x518A + 1);
	i_range(s, 0x518C, 0x518D + 1);
	i_range(s, 0x518F, 0x5193 + 1);
	i_range(s, 0x5195, 0x5197 + 1);
	i_single(s, 0x5199);
	i_single(s, 0x519D);
	i_single(s, 0x51A0);
	i_single(s, 0x51A2);
	i_range(s, 0x51A4, 0x51A6 + 1);
	i_range(s, 0x51A8, 0x51AC + 1);
	i_range(s, 0x51B0, 0x51B7 + 1);
	i_range(s, 0x51BD, 0x51BE + 1);
	i_range(s, 0x51C4, 0x51C6 + 1);
	i_single(s, 0x51C9);
	i_range(s, 0x51CB, 0x51CD + 1);
	i_single(s, 0x51D6);
	i_range(s, 0x51DB, 0x51DD + 1);
	i_range(s, 0x51E0, 0x51E1 + 1);
	i_range(s, 0x51E6, 0x51E7 + 1);
	i_range(s, 0x51E9, 0x51EA + 1);
	i_range(s, 0x51EC, 0x51ED + 1);
	i_range(s, 0x51F0, 0x51F1 + 1);
	i_range(s, 0x51F5, 0x51F6 + 1);
	i_range(s, 0x51F8, 0x51FA + 1);
	i_range(s, 0x51FD, 0x51FE + 1);
	i_single(s, 0x5200);
	i_range(s, 0x5203, 0x5204 + 1);
	i_range(s, 0x5206, 0x5208 + 1);
	i_range(s, 0x520A, 0x520B + 1);
	i_single(s, 0x520E);
	i_single(s, 0x5211);
	i_range(s, 0x5214, 0x5215 + 1);
	i_single(s, 0x5217);
	i_single(s, 0x521D);
	i_range(s, 0x5224, 0x5225 + 1);
	i_single(s, 0x5227);
	i_range(s, 0x5229, 0x522A + 1);
	i_single(s, 0x522E);
	i_single(s, 0x5230);
	i_single(s, 0x5233);
	i_range(s, 0x5236, 0x523B + 1);
	i_range(s, 0x5243, 0x5244 + 1);
	i_single(s, 0x5247);
	i_range(s, 0x524A, 0x524D + 1);
	i_single(s, 0x524F);
	i_single(s, 0x5254);
	i_single(s, 0x5256);
	i_single(s, 0x525B);
	i_single(s, 0x525E);
	i_range(s, 0x5263, 0x5265 + 1);
	i_range(s, 0x5269, 0x526A + 1);
	i_range(s, 0x526F, 0x5275 + 1);
	i_single(s, 0x527D);
	i_single(s, 0x527F);
	i_single(s, 0x5283);
	i_range(s, 0x5287, 0x5289 + 1);
	i_single(s, 0x528D);
	i_range(s, 0x5291, 0x5292 + 1);
	i_single(s, 0x5294);
	i_range(s, 0x529B, 0x529C + 1);
	i_range(s, 0x529F, 0x52A0 + 1);
	i_single(s, 0x52A3);
	i_single(s, 0x52A6);
	i_range(s, 0x52A9, 0x52AD + 1);
	i_single(s, 0x52AF);
	i_single(s, 0x52B1);
	i_range(s, 0x52B4, 0x52B5 + 1);
	i_single(s, 0x52B9);
	i_single(s, 0x52BC);
	i_single(s, 0x52BE);
	i_range(s, 0x52C0, 0x52C1 + 1);
	i_single(s, 0x52C3);
	i_single(s, 0x52C5);
	i_single(s, 0x52C7);
	i_single(s, 0x52C9);
	i_single(s, 0x52CD);
	i_single(s, 0x52D2);
	i_single(s, 0x52D5);
	i_range(s, 0x52D7, 0x52D9 + 1);
	i_single(s, 0x52DB);
	i_range(s, 0x52DD, 0x52E0 + 1);
	i_range(s, 0x52E2, 0x52E4 + 1);
	i_range(s, 0x52E6, 0x52E7 + 1);
	i_range(s, 0x52F2, 0x52F3 + 1);
	i_single(s, 0x52F5);
	i_range(s, 0x52F8, 0x52FA + 1);
	i_range(s, 0x52FE, 0x5302 + 1);
	i_range(s, 0x5305, 0x5308 + 1);
	i_single(s, 0x530D);
	i_range(s, 0x530F, 0x5310 + 1);
	i_range(s, 0x5315, 0x5317 + 1);
	i_range(s, 0x5319, 0x531A + 1);
	i_single(s, 0x531D);
	i_range(s, 0x5320, 0x5321 + 1);
	i_range(s, 0x5323, 0x5324 + 1);
	i_single(s, 0x532A);
	i_single(s, 0x532F);
	i_single(s, 0x5331);
	i_single(s, 0x5333);
	i_range(s, 0x5338, 0x533B + 1);
	i_range(s, 0x533F, 0x5341 + 1);
	i_single(s, 0x5343);
	i_range(s, 0x5345, 0x534A + 1);
	i_single(s, 0x534D);
	i_range(s, 0x5351, 0x5354 + 1);
	i_range(s, 0x5357, 0x5358 + 1);
	i_single(s, 0x535A);
	i_single(s, 0x535C);
	i_single(s, 0x535E);
	i_single(s, 0x5360);
	i_single(s, 0x5366);
	i_single(s, 0x5369);
	i_range(s, 0x536E, 0x5375 + 1);
	i_range(s, 0x5377, 0x5378 + 1);
	i_single(s, 0x537B);
	i_single(s, 0x537F);
	i_single(s, 0x5382);
	i_single(s, 0x5384);
	i_single(s, 0x5393);
	i_single(s, 0x5396);
	i_single(s, 0x5398);
	i_single(s, 0x539A);
	i_range(s, 0x539F, 0x53A0 + 1);
	i_range(s, 0x53A5, 0x53A6 + 1);
	i_range(s, 0x53A8, 0x53A9 + 1);
	i_range(s, 0x53AD, 0x53AE + 1);
	i_single(s, 0x53B0);
	i_range(s, 0x53B2, 0x53B3 + 1);
	i_single(s, 0x53B6);
	i_single(s, 0x53BB);
	i_range(s, 0x53C2, 0x53C3 + 1);
	i_range(s, 0x53C8, 0x53CE + 1);
	i_single(s, 0x53D4);
	i_range(s, 0x53D6, 0x53D7 + 1);
	i_single(s, 0x53D9);
	i_single(s, 0x53DB);
	i_single(s, 0x53DD);
	i_single(s, 0x53DF);
	i_range(s, 0x53E1, 0x53E5 + 1);
	i_range(s, 0x53E8, 0x53F3 + 1);
	i_range(s, 0x53F6, 0x53F8 + 1);
	i_single(s, 0x53FA);
	i_single(s, 0x5401);
	i_range(s, 0x5403, 0x5404 + 1);
	i_range(s, 0x5408, 0x5411 + 1);
	i_single(s, 0x541B);
	i_single(s, 0x541D);
	i_range(s, 0x541F, 0x5420 + 1);
	i_single(s, 0x5426);
	i_single(s, 0x5429);
	i_range(s, 0x542B, 0x542E + 1);
	i_single(s, 0x5436);
	i_range(s, 0x5438, 0x5439 + 1);
	i_range(s, 0x543B, 0x543E + 1);
	i_single(s, 0x5440);
	i_single(s, 0x5442);
	i_single(s, 0x5446);
	i_range(s, 0x5448, 0x544A + 1);
	i_single(s, 0x544E);
	i_single(s, 0x5451);
	i_single(s, 0x545F);
	i_single(s, 0x5468);
	i_single(s, 0x546A);
	i_range(s, 0x5470, 0x5471 + 1);
	i_single(s, 0x5473);
	i_range(s, 0x5475, 0x5477 + 1);
	i_range(s, 0x547B, 0x547D + 1);
	i_single(s, 0x5480);
	i_single(s, 0x5484);
	i_single(s, 0x5486);
	i_range(s, 0x548A, 0x548C + 1);
	i_range(s, 0x548E, 0x5490 + 1);
	i_single(s, 0x5492);
	i_single(s, 0x549C);
	i_single(s, 0x54A2);
	i_range(s, 0x54A4, 0x54A5 + 1);
	i_range(s, 0x54A8, 0x54A9 + 1);
	i_range(s, 0x54AB, 0x54AC + 1);
	i_single(s, 0x54AF);
	i_range(s, 0x54B2, 0x54B3 + 1);
	i_single(s, 0x54B8);
	i_range(s, 0x54BC, 0x54BE + 1);
	i_range(s, 0x54C0, 0x54C2 + 1);
	i_single(s, 0x54C4);
	i_range(s, 0x54C7, 0x54C9 + 1);
	i_single(s, 0x54D8);
	i_range(s, 0x54E1, 0x54E2 + 1);
	i_range(s, 0x54E5, 0x54E6 + 1);
	i_range(s, 0x54E8, 0x54E9 + 1);
	i_range(s, 0x54ED, 0x54EE + 1);
	i_single(s, 0x54F2);
	i_single(s, 0x54FA);
	i_single(s, 0x54FD);
	i_single(s, 0x54FF);
	i_single(s, 0x5504);
	i_range(s, 0x5506, 0x5507 + 1);
	i_range(s, 0x550F, 0x5510 + 1);
	i_single(s, 0x5514);
	i_single(s, 0x5516);
	i_range(s, 0x552E, 0x552F + 1);
	i_single(s, 0x5531);
	i_single(s, 0x5533);
	i_range(s, 0x5538, 0x5539 + 1);
	i_single(s, 0x553E);
	i_single(s, 0x5540);
	i_range(s, 0x5544, 0x5546 + 1);
	i_single(s, 0x554C);
	i_single(s, 0x554F);
	i_single(s, 0x5553);
	i_range(s, 0x5556, 0x5557 + 1);
	i_range(s, 0x555C, 0x555D + 1);
	i_single(s, 0x5563);
	i_range(s, 0x557B, 0x557C + 1);
	i_single(s, 0x557E);
	i_single(s, 0x5580);
	i_range(s, 0x5583, 0x5584 + 1);
	i_range(s, 0x5586, 0x5587 + 1);
	i_range(s, 0x5589, 0x558B + 1);
	i_range(s, 0x5598, 0x559A + 1);
	i_range(s, 0x559C, 0x559F + 1);
	i_range(s, 0x55A7, 0x55AC + 1);
	i_single(s, 0x55AE);
	i_single(s, 0x55B0);
	i_single(s, 0x55B6);
	i_range(s, 0x55C4, 0x55C5 + 1);
	i_single(s, 0x55C7);
	i_single(s, 0x55D4);
	i_single(s, 0x55DA);
	i_single(s, 0x55DC);
	i_single(s, 0x55DF);
	i_range(s, 0x55E3, 0x55E4 + 1);
	i_single(s, 0x55F7);
	i_single(s, 0x55F9);
	i_range(s, 0x55FD, 0x55FE + 1);
	i_single(s, 0x5606);
	i_single(s, 0x5609);
	i_single(s, 0x5614);
	i_range(s, 0x5616, 0x5618 + 1);
	i_single(s, 0x561B);
	i_single(s, 0x5629);
	i_single(s, 0x562F);
	i_range(s, 0x5631, 0x5632 + 1);
	i_single(s, 0x5634);
	i_single(s, 0x5636);
	i_single(s, 0x5638);
	i_single(s, 0x5642);
	i_single(s, 0x564C);
	i_single(s, 0x564E);
	i_single(s, 0x5650);
	i_single(s, 0x565B);
	i_single(s, 0x5664);
	i_single(s, 0x5668);
	i_range(s, 0x566A, 0x566C + 1);
	i_single(s, 0x5674);
	i_single(s, 0x5678);
	i_single(s, 0x567A);
	i_single(s, 0x5680);
	i_range(s, 0x5686, 0x5687 + 1);
	i_single(s, 0x568A);
	i_single(s, 0x568F);
	i_single(s, 0x5694);
	i_single(s, 0x56A0);
	i_single(s, 0x56A2);
	i_single(s, 0x56A5);
	i_single(s, 0x56AE);
	i_single(s, 0x56B4);
	i_single(s, 0x56B6);
	i_single(s, 0x56BC);
	i_range(s, 0x56C0, 0x56C3 + 1);
	i_single(s, 0x56C8);
	i_single(s, 0x56CE);
	i_single(s, 0x56D1);
	i_single(s, 0x56D3);
	i_range(s, 0x56D7, 0x56D8 + 1);
	i_range(s, 0x56DA, 0x56DB + 1);
	i_single(s, 0x56DE);
	i_single(s, 0x56E0);
	i_single(s, 0x56E3);
	i_single(s, 0x56EE);
	i_single(s, 0x56F0);
	i_range(s, 0x56F2, 0x56F3 + 1);
	i_range(s, 0x56F9, 0x56FA + 1);
	i_single(s, 0x56FD);
	i_range(s, 0x56FF, 0x5700 + 1);
	i_range(s, 0x5703, 0x5704 + 1);
	i_range(s, 0x5708, 0x5709 + 1);
	i_single(s, 0x570B);
	i_single(s, 0x570D);
	i_single(s, 0x570F);
	i_range(s, 0x5712, 0x5713 + 1);
	i_single(s, 0x5716);
	i_single(s, 0x5718);
	i_single(s, 0x571C);
	i_single(s, 0x571F);
	i_range(s, 0x5726, 0x5728 + 1);
	i_single(s, 0x572D);
	i_single(s, 0x5730);
	i_range(s, 0x5737, 0x5738 + 1);
	i_single(s, 0x573B);
	i_single(s, 0x5740);
	i_single(s, 0x5742);
	i_single(s, 0x5747);
	i_single(s, 0x574A);
	i_range(s, 0x574E, 0x5751 + 1);
	i_single(s, 0x5759);
	i_single(s, 0x5761);
	i_range(s, 0x5764, 0x5766 + 1);
	i_range(s, 0x5769, 0x576A + 1);
	i_single(s, 0x577F);
	i_single(s, 0x5782);
	i_range(s, 0x5788, 0x5789 + 1);
	i_single(s, 0x578B);
	i_single(s, 0x5793);
	i_single(s, 0x57A0);
	i_range(s, 0x57A2, 0x57A4 + 1);
	i_single(s, 0x57AA);
	i_single(s, 0x57AC);
	i_single(s, 0x57B0);
	i_single(s, 0x57B3);
	i_single(s, 0x57C0);
	i_single(s, 0x57C3);
	i_range(s, 0x57C6, 0x57C8 + 1);
	i_single(s, 0x57CB);
	i_single(s, 0x57CE);
	i_range(s, 0x57D2, 0x57D4 + 1);
	i_single(s, 0x57D6);
	i_single(s, 0x57DC);
	i_range(s, 0x57DF, 0x57E0 + 1);
	i_single(s, 0x57E3);
	i_single(s, 0x57F4);
	i_single(s, 0x57F7);
	i_range(s, 0x57F9, 0x57FA + 1);
	i_single(s, 0x57FC);
	i_single(s, 0x5800);
	i_single(s, 0x5802);
	i_range(s, 0x5805, 0x5806 + 1);
	i_range(s, 0x580A, 0x580B + 1);
	i_single(s, 0x5815);
	i_single(s, 0x5819);
	i_single(s, 0x581D);
	i_single(s, 0x5821);
	i_single(s, 0x5824);
	i_single(s, 0x582A);
	i_range(s, 0x582F, 0x5831 + 1);
	i_range(s, 0x5834, 0x5835 + 1);
	i_single(s, 0x583A);
	i_single(s, 0x583D);
	i_range(s, 0x5840, 0x5841 + 1);
	i_range(s, 0x584A, 0x584B + 1);
	i_range(s, 0x5851, 0x5852 + 1);
	i_single(s, 0x5854);
	i_range(s, 0x5857, 0x585A + 1);
	i_single(s, 0x585E);
	i_single(s, 0x5862);
	i_single(s, 0x5869);
	i_single(s, 0x586B);
	i_single(s, 0x5870);
	i_single(s, 0x5872);
	i_single(s, 0x5875);
	i_single(s, 0x5879);
	i_single(s, 0x587E);
	i_single(s, 0x5883);
	i_single(s, 0x5885);
	i_single(s, 0x5893);
	i_single(s, 0x5897);
	i_single(s, 0x589C);
	i_range(s, 0x589E, 0x589F + 1);
	i_single(s, 0x58A8);
	i_single(s, 0x58AB);
	i_single(s, 0x58AE);
	i_range(s, 0x58B2, 0x58B3 + 1);
	i_range(s, 0x58B8, 0x58BB + 1);
	i_single(s, 0x58BE);
	i_single(s, 0x58C1);
	i_single(s, 0x58C5);
	i_single(s, 0x58C7);
	i_single(s, 0x58CA);
	i_single(s, 0x58CC);
	i_single(s, 0x58D1);
	i_single(s, 0x58D3);
	i_single(s, 0x58D5);
	i_range(s, 0x58D7, 0x58D9 + 1);
	i_single(s, 0x58DC);
	i_range(s, 0x58DE, 0x58DF + 1);
	i_range(s, 0x58E4, 0x58E5 + 1);
	i_range(s, 0x58EB, 0x58EC + 1);
	i_range(s, 0x58EE, 0x58F2 + 1);
	i_single(s, 0x58F7);
	i_range(s, 0x58F9, 0x58FD + 1);
	i_single(s, 0x5902);
	i_range(s, 0x5909, 0x590B + 1);
	i_range(s, 0x590F, 0x5910 + 1);
	i_range(s, 0x5915, 0x5916 + 1);
	i_range(s, 0x5918, 0x591C + 1);
	i_single(s, 0x5922);
	i_single(s, 0x5925);
	i_single(s, 0x5927);
	i_range(s, 0x5929, 0x592E + 1);
	i_range(s, 0x5931, 0x5932 + 1);
	i_range(s, 0x5937, 0x5938 + 1);
	i_single(s, 0x593E);
	i_single(s, 0x5944);
	i_range(s, 0x5947, 0x5949 + 1);
	i_range(s, 0x594E, 0x5951 + 1);
	i_range(s, 0x5953, 0x5955 + 1);
	i_range(s, 0x5957, 0x5958 + 1);
	i_range(s, 0x595A, 0x595B + 1);
	i_single(s, 0x595D);
	i_single(s, 0x5960);
	i_range(s, 0x5962, 0x5963 + 1);
	i_single(s, 0x5965);
	i_range(s, 0x5967, 0x596A + 1);
	i_single(s, 0x596C);
	i_single(s, 0x596E);
	i_range(s, 0x5973, 0x5974 + 1);
	i_single(s, 0x5978);
	i_single(s, 0x597D);
	i_range(s, 0x5981, 0x5984 + 1);
	i_single(s, 0x598A);
	i_single(s, 0x598D);
	i_single(s, 0x5993);
	i_single(s, 0x5996);
	i_single(s, 0x5999);
	i_single(s, 0x599B);
	i_single(s, 0x599D);
	i_range(s, 0x59A3, 0x59A5 + 1);
	i_single(s, 0x59A8);
	i_single(s, 0x59AC);
	i_single(s, 0x59B2);
	i_range(s, 0x59B9, 0x59BB + 1);
	i_single(s, 0x59BE);
	i_single(s, 0x59C6);
	i_single(s, 0x59C9);
	i_single(s, 0x59CB);
	i_range(s, 0x59D0, 0x59D1 + 1);
	i_range(s, 0x59D3, 0x59D4 + 1);
	i_range(s, 0x59D9, 0x59DA + 1);
	i_single(s, 0x59DC);
	i_range(s, 0x59E5, 0x59E6 + 1);
	i_single(s, 0x59E8);
	i_range(s, 0x59EA, 0x59EB + 1);
	i_single(s, 0x59F6);
	i_single(s, 0x59FB);
	i_single(s, 0x59FF);
	i_single(s, 0x5A01);
	i_single(s, 0x5A03);
	i_single(s, 0x5A09);
	i_single(s, 0x5A11);
	i_single(s, 0x5A18);
	i_single(s, 0x5A1A);
	i_single(s, 0x5A1C);
	i_range(s, 0x5A1F, 0x5A20 + 1);
	i_single(s, 0x5A25);
	i_single(s, 0x5A29);
	i_single(s, 0x5A2F);
	i_range(s, 0x5A35, 0x5A36 + 1);
	i_single(s, 0x5A3C);
	i_range(s, 0x5A40, 0x5A41 + 1);
	i_single(s, 0x5A46);
	i_single(s, 0x5A49);
	i_single(s, 0x5A5A);
	i_single(s, 0x5A62);
	i_single(s, 0x5A66);
	i_single(s, 0x5A6A);
	i_single(s, 0x5A6C);
	i_single(s, 0x5A7F);
	i_single(s, 0x5A92);
	i_range(s, 0x5A9A, 0x5A9B + 1);
	i_range(s, 0x5ABC, 0x5ABE + 1);
	i_range(s, 0x5AC1, 0x5AC2 + 1);
	i_single(s, 0x5AC9);
	i_range(s, 0x5ACB, 0x5ACC + 1);
	i_single(s, 0x5AD0);
	i_range(s, 0x5AD6, 0x5AD7 + 1);
	i_single(s, 0x5AE1);
	i_single(s, 0x5AE3);
	i_single(s, 0x5AE6);
	i_single(s, 0x5AE9);
	i_range(s, 0x5AFA, 0x5AFB + 1);
	i_single(s, 0x5B09);
	i_range(s, 0x5B0B, 0x5B0C + 1);
	i_single(s, 0x5B16);
	i_single(s, 0x5B22);
	i_single(s, 0x5B2A);
	i_single(s, 0x5B2C);
	i_single(s, 0x5B30);
	i_single(s, 0x5B32);
	i_single(s, 0x5B36);
	i_single(s, 0x5B3E);
	i_single(s, 0x5B40);
	i_single(s, 0x5B43);
	i_single(s, 0x5B45);
	i_range(s, 0x5B50, 0x5B51 + 1);
	i_range(s, 0x5B54, 0x5B58 + 1);
	i_range(s, 0x5B5A, 0x5B5D + 1);
	i_single(s, 0x5B5F);
	i_range(s, 0x5B63, 0x5B66 + 1);
	i_single(s, 0x5B69);
	i_single(s, 0x5B6B);
	i_range(s, 0x5B70, 0x5B71 + 1);
	i_single(s, 0x5B73);
	i_single(s, 0x5B75);
	i_single(s, 0x5B78);
	i_single(s, 0x5B7A);
	i_single(s, 0x5B80);
	i_single(s, 0x5B83);
	i_single(s, 0x5B85);
	i_range(s, 0x5B87, 0x5B89 + 1);
	i_range(s, 0x5B8B, 0x5B8D + 1);
	i_single(s, 0x5B8F);
	i_single(s, 0x5B95);
	i_range(s, 0x5B97, 0x5B9D + 1);
	i_single(s, 0x5B9F);
	i_range(s, 0x5BA2, 0x5BA6 + 1);
	i_single(s, 0x5BAE);
	i_single(s, 0x5BB0);
	i_range(s, 0x5BB3, 0x5BB6 + 1);
	i_range(s, 0x5BB8, 0x5BB9 + 1);
	i_range(s, 0x5BBF, 0x5BC0 + 1);
	i_range(s, 0x5BC2, 0x5BC7 + 1);
	i_single(s, 0x5BC9);
	i_single(s, 0x5BCC);
	i_single(s, 0x5BD0);
	i_range(s, 0x5BD2, 0x5BD4 + 1);
	i_single(s, 0x5BD8);
	i_single(s, 0x5BDB);
	i_range(s, 0x5BDD, 0x5BDF + 1);
	i_range(s, 0x5BE1, 0x5BE2 + 1);
	i_range(s, 0x5BE4, 0x5BE9 + 1);
	i_range(s, 0x5BEB, 0x5BEC + 1);
	i_single(s, 0x5BEE);
	i_single(s, 0x5BF0);
	i_single(s, 0x5BF3);
	i_range(s, 0x5BF5, 0x5BF6 + 1);
	i_single(s, 0x5BF8);
	i_single(s, 0x5BFA);
	i_range(s, 0x5BFE, 0x5BFF + 1);
	i_range(s, 0x5C01, 0x5C02 + 1);
	i_range(s, 0x5C04, 0x5C0B + 1);
	i_range(s, 0x5C0D, 0x5C0F + 1);
	i_single(s, 0x5C11);
	i_single(s, 0x5C13);
	i_single(s, 0x5C16);
	i_single(s, 0x5C1A);
	i_single(s, 0x5C1E);
	i_single(s, 0x5C20);
	i_single(s, 0x5C22);
	i_single(s, 0x5C24);
	i_single(s, 0x5C28);
	i_single(s, 0x5C2D);
	i_single(s, 0x5C31);
	i_range(s, 0x5C38, 0x5C41 + 1);
	i_range(s, 0x5C45, 0x5C46 + 1);
	i_single(s, 0x5C48);
	i_range(s, 0x5C4A, 0x5C4B + 1);
	i_range(s, 0x5C4D, 0x5C51 + 1);
	i_single(s, 0x5C53);
	i_single(s, 0x5C55);
	i_single(s, 0x5C5E);
	i_range(s, 0x5C60, 0x5C61 + 1);
	i_range(s, 0x5C64, 0x5C65 + 1);
	i_single(s, 0x5C6C);
	i_range(s, 0x5C6E, 0x5C6F + 1);
	i_single(s, 0x5C71);
	i_single(s, 0x5C76);
	i_single(s, 0x5C79);
	i_single(s, 0x5C8C);
	i_range(s, 0x5C90, 0x5C91 + 1);
	i_single(s, 0x5C94);
	i_single(s, 0x5CA1);
	i_single(s, 0x5CA6);
	i_range(s, 0x5CA8, 0x5CA9 + 1);
	i_range(s, 0x5CAB, 0x5CAC + 1);
	i_single(s, 0x5CB1);
	i_single(s, 0x5CB3);
	i_range(s, 0x5CB6, 0x5CB8 + 1);
	i_range(s, 0x5CBA, 0x5CBC + 1);
	i_single(s, 0x5CBE);
	i_single(s, 0x5CC5);
	i_single(s, 0x5CC7);
	i_single(s, 0x5CD9);
	i_range(s, 0x5CE0, 0x5CE1 + 1);
	i_range(s, 0x5CE8, 0x5CEA + 1);
	i_single(s, 0x5CED);
	i_range(s, 0x5CEF, 0x5CF0 + 1);
	i_range(s, 0x5CF5, 0x5CF6 + 1);
	i_range(s, 0x5CFA, 0x5CFB + 1);
	i_single(s, 0x5CFD);
	i_single(s, 0x5D07);
	i_single(s, 0x5D0B);
	i_single(s, 0x5D0E);
	i_single(s, 0x5D11);
	i_range(s, 0x5D14, 0x5D1B + 1);
	i_single(s, 0x5D1F);
	i_single(s, 0x5D22);
	i_single(s, 0x5D27);
	i_single(s, 0x5D29);
	i_single(s, 0x5D42);
	i_range(s, 0x5D4B, 0x5D4C + 1);
	i_single(s, 0x5D4E);
	i_single(s, 0x5D50);
	i_range(s, 0x5D52, 0x5D53 + 1);
	i_single(s, 0x5D5C);
	i_single(s, 0x5D69);
	i_range(s, 0x5D6C, 0x5D6D + 1);
	i_single(s, 0x5D6F);
	i_single(s, 0x5D73);
	i_single(s, 0x5D76);
	i_single(s, 0x5D82);
	i_single(s, 0x5D84);
	i_single(s, 0x5D87);
	i_range(s, 0x5D8B, 0x5D8C + 1);
	i_single(s, 0x5D90);
	i_single(s, 0x5D9D);
	i_single(s, 0x5DA2);
	i_single(s, 0x5DAC);
	i_single(s, 0x5DAE);
	i_range(s, 0x5DB7, 0x5DBA + 1);
	i_range(s, 0x5DBC, 0x5DBD + 1);
	i_single(s, 0x5DC9);
	i_range(s, 0x5DCC, 0x5DCD + 1);
	i_single(s, 0x5DD0);
	i_range(s, 0x5DD2, 0x5DD3 + 1);
	i_single(s, 0x5DD6);
	i_single(s, 0x5DDB);
	i_range(s, 0x5DDD, 0x5DDE + 1);
	i_single(s, 0x5DE1);
	i_single(s, 0x5DE3);
	i_range(s, 0x5DE5, 0x5DE8 + 1);
	i_single(s, 0x5DEB);
	i_single(s, 0x5DEE);
	i_range(s, 0x5DF1, 0x5DF5 + 1);
	i_single(s, 0x5DF7);
	i_single(s, 0x5DFB);
	i_range(s, 0x5DFD, 0x5DFE + 1);
	i_range(s, 0x5E02, 0x5E03 + 1);
	i_single(s, 0x5E06);
	i_range(s, 0x5E0B, 0x5E0C + 1);
	i_single(s, 0x5E11);
	i_single(s, 0x5E16);
	i_range(s, 0x5E19, 0x5E1B + 1);
	i_single(s, 0x5E1D);
	i_single(s, 0x5E25);
	i_single(s, 0x5E2B);
	i_single(s, 0x5E2D);
	i_range(s, 0x5E2F, 0x5E30 + 1);
	i_single(s, 0x5E33);
	i_range(s, 0x5E36, 0x5E38 + 1);
	i_single(s, 0x5E3D);
	i_single(s, 0x5E40);
	i_range(s, 0x5E43, 0x5E45 + 1);
	i_single(s, 0x5E47);
	i_single(s, 0x5E4C);
	i_single(s, 0x5E4E);
	i_range(s, 0x5E54, 0x5E55 + 1);
	i_single(s, 0x5E57);
	i_single(s, 0x5E5F);
	i_range(s, 0x5E61, 0x5E64 + 1);
	i_range(s, 0x5E72, 0x5E76 + 1);
	i_range(s, 0x5E78, 0x5E7F + 1);
	i_single(s, 0x5E81);
	i_range(s, 0x5E83, 0x5E84 + 1);
	i_single(s, 0x5E87);
	i_single(s, 0x5E8A);
	i_single(s, 0x5E8F);
	i_range(s, 0x5E95, 0x5E97 + 1);
	i_single(s, 0x5E9A);
	i_single(s, 0x5E9C);
	i_single(s, 0x5EA0);
	i_range(s, 0x5EA6, 0x5EA7 + 1);
	i_single(s, 0x5EAB);
	i_single(s, 0x5EAD);
	i_range(s, 0x5EB5, 0x5EB8 + 1);
	i_range(s, 0x5EC1, 0x5EC3 + 1);
	i_range(s, 0x5EC8, 0x5ECA + 1);
	i_range(s, 0x5ECF, 0x5ED0 + 1);
	i_single(s, 0x5ED3);
	i_single(s, 0x5ED6);
	i_range(s, 0x5EDA, 0x5EDB + 1);
	i_single(s, 0x5EDD);
	i_range(s, 0x5EDF, 0x5EE3 + 1);
	i_range(s, 0x5EE8, 0x5EE9 + 1);
	i_single(s, 0x5EEC);
	i_range(s, 0x5EF0, 0x5EF1 + 1);
	i_range(s, 0x5EF3, 0x5EF4 + 1);
	i_range(s, 0x5EF6, 0x5EF8 + 1);
	i_range(s, 0x5EFA, 0x5EFC + 1);
	i_range(s, 0x5EFE, 0x5EFF + 1);
	i_single(s, 0x5F01);
	i_range(s, 0x5F03, 0x5F04 + 1);
	i_range(s, 0x5F09, 0x5F0D + 1);
	i_range(s, 0x5F0F, 0x5F11 + 1);
	i_range(s, 0x5F13, 0x5F18 + 1);
	i_single(s, 0x5F1B);
	i_single(s, 0x5F1F);
	i_single(s, 0x5F21);
	i_range(s, 0x5F25, 0x5F27 + 1);
	i_single(s, 0x5F29);
	i_single(s, 0x5F2D);
	i_single(s, 0x5F2F);
	i_single(s, 0x5F31);
	i_range(s, 0x5F34, 0x5F35 + 1);
	i_range(s, 0x5F37, 0x5F38 + 1);
	i_single(s, 0x5F3C);
	i_single(s, 0x5F3E);
	i_single(s, 0x5F41);
	i_single(s, 0x5F45);
	i_single(s, 0x5F48);
	i_single(s, 0x5F4A);
	i_single(s, 0x5F4C);
	i_single(s, 0x5F4E);
	i_single(s, 0x5F51);
	i_single(s, 0x5F53);
	i_range(s, 0x5F56, 0x5F57 + 1);
	i_single(s, 0x5F59);
	i_range(s, 0x5F5C, 0x5F5D + 1);
	i_range(s, 0x5F61, 0x5F62 + 1);
	i_range(s, 0x5F66, 0x5F67 + 1);
	i_range(s, 0x5F69, 0x5F6D + 1);
	i_range(s, 0x5F70, 0x5F71 + 1);
	i_single(s, 0x5F73);
	i_single(s, 0x5F77);
	i_single(s, 0x5F79);
	i_single(s, 0x5F7C);
	i_range(s, 0x5F7F, 0x5F85 + 1);
	i_range(s, 0x5F87, 0x5F88 + 1);
	i_range(s, 0x5F8A, 0x5F8C + 1);
	i_range(s, 0x5F90, 0x5F93 + 1);
	i_range(s, 0x5F97, 0x5F99 + 1);
	i_single(s, 0x5F9E);
	i_range(s, 0x5FA0, 0x5FA1 + 1);
	i_range(s, 0x5FA8, 0x5FAA + 1);
	i_range(s, 0x5FAD, 0x5FAE + 1);
	i_range(s, 0x5FB3, 0x5FB4 + 1);
	i_single(s, 0x5FB7);
	i_single(s, 0x5FB9);
	i_range(s, 0x5FBC, 0x5FBD + 1);
	i_single(s, 0x5FC3);
	i_single(s, 0x5FC5);
	i_range(s, 0x5FCC, 0x5FCD + 1);
	i_range(s, 0x5FD6, 0x5FD9 + 1);
	i_range(s, 0x5FDC, 0x5FDE + 1);
	i_single(s, 0x5FE0);
	i_single(s, 0x5FE4);
	i_single(s, 0x5FEB);
	i_range(s, 0x5FF0, 0x5FF1 + 1);
	i_single(s, 0x5FF5);
	i_single(s, 0x5FF8);
	i_single(s, 0x5FFB);
	i_single(s, 0x5FFD);
	i_single(s, 0x5FFF);
	i_range(s, 0x600E, 0x6010 + 1);
	i_single(s, 0x6012);
	i_range(s, 0x6015, 0x6016 + 1);
	i_single(s, 0x6019);
	i_range(s, 0x601B, 0x601D + 1);
	i_range(s, 0x6020, 0x6021 + 1);
	i_range(s, 0x6025, 0x602B + 1);
	i_single(s, 0x602F);
	i_single(s, 0x6031);
	i_single(s, 0x603A);
	i_range(s, 0x6041, 0x6043 + 1);
	i_single(s, 0x6046);
	i_range(s, 0x604A, 0x604B + 1);
	i_single(s, 0x604D);
	i_single(s, 0x6050);
	i_single(s, 0x6052);
	i_single(s, 0x6055);
	i_range(s, 0x6059, 0x605A + 1);
	i_single(s, 0x605D);
	i_range(s, 0x605F, 0x6060 + 1);
	i_range(s, 0x6062, 0x6065 + 1);
	i_range(s, 0x6068, 0x606D + 1);
	i_range(s, 0x606F, 0x6070 + 1);
	i_single(s, 0x6075);
	i_single(s, 0x6077);
	i_single(s, 0x6081);
	i_range(s, 0x6083, 0x6085 + 1);
	i_range(s, 0x6089, 0x608D + 1);
	i_single(s, 0x6092);
	i_single(s, 0x6094);
	i_range(s, 0x6096, 0x6097 + 1);
	i_range(s, 0x609A, 0x609B + 1);
	i_range(s, 0x609F, 0x60A0 + 1);
	i_single(s, 0x60A3);
	i_range(s, 0x60A6, 0x60A7 + 1);
	i_range(s, 0x60A9, 0x60AA + 1);
	i_range(s, 0x60B2, 0x60B6 + 1);
	i_single(s, 0x60B8);
	i_range(s, 0x60BC, 0x60BD + 1);
	i_range(s, 0x60C5, 0x60C7 + 1);
	i_single(s, 0x60D1);
	i_single(s, 0x60D3);
	i_single(s, 0x60D5);
	i_single(s, 0x60D8);
	i_single(s, 0x60DA);
	i_single(s, 0x60DC);
	i_range(s, 0x60DE, 0x60E1 + 1);
	i_single(s, 0x60E3);
	i_range(s, 0x60E7, 0x60E8 + 1);
	i_range(s, 0x60F0, 0x60F4 + 1);
	i_range(s, 0x60F6, 0x60F7 + 1);
	i_range(s, 0x60F9, 0x60FB + 1);
	i_range(s, 0x6100, 0x6101 + 1);
	i_single(s, 0x6103);
	i_single(s, 0x6106);
	i_range(s, 0x6108, 0x6109 + 1);
	i_range(s, 0x610D, 0x610F + 1);
	i_single(s, 0x6111);
	i_single(s, 0x6115);
	i_range(s, 0x611A, 0x611B + 1);
	i_range(s, 0x611F, 0x6121 + 1);
	i_range(s, 0x6127, 0x6128 + 1);
	i_single(s, 0x612C);
	i_single(s, 0x6130);
	i_single(s, 0x6134);
	i_single(s, 0x6137);
	i_range(s, 0x613C, 0x613F + 1);
	i_single(s, 0x6142);
	i_single(s, 0x6144);
	i_range(s, 0x6147, 0x6148 + 1);
	i_range(s, 0x614A, 0x614E + 1);
	i_single(s, 0x6153);
	i_single(s, 0x6155);
	i_range(s, 0x6158, 0x615A + 1);
	i_single(s, 0x615D);
	i_single(s, 0x615F);
	i_range(s, 0x6162, 0x6163 + 1);
	i_single(s, 0x6165);
	i_range(s, 0x6167, 0x6168 + 1);
	i_single(s, 0x616B);
	i_range(s, 0x616E, 0x6171 + 1);
	i_range(s, 0x6173, 0x6177 + 1);
	i_single(s, 0x617E);
	i_single(s, 0x6182);
	i_single(s, 0x6187);
	i_single(s, 0x618A);
	i_single(s, 0x618E);
	i_range(s, 0x6190, 0x6191 + 1);
	i_single(s, 0x6194);
	i_single(s, 0x6196);
	i_range(s, 0x6198, 0x619A + 1);
	i_single(s, 0x61A4);
	i_single(s, 0x61A7);
	i_single(s, 0x61A9);
	i_range(s, 0x61AB, 0x61AC + 1);
	i_single(s, 0x61AE);
	i_single(s, 0x61B2);
	i_single(s, 0x61B6);
	i_single(s, 0x61BA);
	i_single(s, 0x61BE);
	i_single(s, 0x61C3);
	i_range(s, 0x61C6, 0x61CD + 1);
	i_single(s, 0x61D0);
	i_single(s, 0x61E3);
	i_single(s, 0x61E6);
	i_single(s, 0x61F2);
	i_single(s, 0x61F4);
	i_range(s, 0x61F6, 0x61F8 + 1);
	i_single(s, 0x61FA);
	i_range(s, 0x61FC, 0x6200 + 1);
	i_range(s, 0x6208, 0x620A + 1);
	i_range(s, 0x620C, 0x620E + 1);
	i_range(s, 0x6210, 0x6214 + 1);
	i_single(s, 0x6216);
	i_range(s, 0x621A, 0x621B + 1);
	i_range(s, 0x621D, 0x621F + 1);
	i_single(s, 0x6221);
	i_single(s, 0x6226);
	i_single(s, 0x622A);
	i_range(s, 0x622E, 0x6230 + 1);
	i_range(s, 0x6232, 0x6234 + 1);
	i_single(s, 0x6238);
	i_single(s, 0x623B);
	i_range(s, 0x623F, 0x6241 + 1);
	i_range(s, 0x6247, 0x6249 + 1);
	i_single(s, 0x624B);
	i_range(s, 0x624D, 0x624E + 1);
	i_single(s, 0x6253);
	i_single(s, 0x6255);
	i_single(s, 0x6258);
	i_single(s, 0x625B);
	i_single(s, 0x625E);
	i_single(s, 0x6260);
	i_single(s, 0x6263);
	i_single(s, 0x6268);
	i_single(s, 0x626E);
	i_single(s, 0x6271);
	i_single(s, 0x6276);
	i_single(s, 0x6279);
	i_single(s, 0x627C);
	i_range(s, 0x627E, 0x6280 + 1);
	i_range(s, 0x6282, 0x6284 + 1);
	i_range(s, 0x6289, 0x628A + 1);
	i_range(s, 0x6291, 0x6298 + 1);
	i_range(s, 0x629B, 0x629C + 1);
	i_single(s, 0x629E);
	i_single(s, 0x62A6);
	i_range(s, 0x62AB, 0x62AC + 1);
	i_single(s, 0x62B1);
	i_single(s, 0x62B5);
	i_single(s, 0x62B9);
	i_range(s, 0x62BB, 0x62BD + 1);
	i_single(s, 0x62C2);
	i_range(s, 0x62C5, 0x62CA + 1);
	i_range(s, 0x62CC, 0x62CD + 1);
	i_range(s, 0x62CF, 0x62D4 + 1);
	i_range(s, 0x62D7, 0x62D9 + 1);
	i_range(s, 0x62DB, 0x62DD + 1);
	i_range(s, 0x62E0, 0x62E1 + 1);
	i_range(s, 0x62EC, 0x62EF + 1);
	i_single(s, 0x62F1);
	i_single(s, 0x62F3);
	i_range(s, 0x62F5, 0x62F7 + 1);
	i_range(s, 0x62FE, 0x62FF + 1);
	i_range(s, 0x6301, 0x6302 + 1);
	i_range(s, 0x6307, 0x6309 + 1);
	i_single(s, 0x630C);
	i_single(s, 0x6311);
	i_single(s, 0x6319);
	i_single(s, 0x631F);
	i_range(s, 0x6327, 0x6328 + 1);
	i_single(s, 0x632B);
	i_single(s, 0x632F);
	i_single(s, 0x633A);
	i_range(s, 0x633D, 0x633F + 1);
	i_single(s, 0x6349);
	i_range(s, 0x634C, 0x634D + 1);
	i_range(s, 0x634F, 0x6350 + 1);
	i_single(s, 0x6355);
	i_single(s, 0x6357);
	i_single(s, 0x635C);
	i_range(s, 0x6367, 0x6369 + 1);
	i_single(s, 0x636B);
	i_single(s, 0x636E);
	i_single(s, 0x6372);
	i_range(s, 0x6376, 0x6377 + 1);
	i_range(s, 0x637A, 0x637B + 1);
	i_single(s, 0x6380);
	i_single(s, 0x6383);
	i_range(s, 0x6388, 0x6389 + 1);
	i_single(s, 0x638C);
	i_range(s, 0x638E, 0x638F + 1);
	i_single(s, 0x6392);
	i_single(s, 0x6396);
	i_single(s, 0x6398);
	i_single(s, 0x639B);
	i_range(s, 0x639F, 0x63A3 + 1);
	i_single(s, 0x63A5);
	i_range(s, 0x63A7, 0x63AC + 1);
	i_single(s, 0x63B2);
	i_range(s, 0x63B4, 0x63B5 + 1);
	i_single(s, 0x63BB);
	i_single(s, 0x63BE);
	i_single(s, 0x63C0);
	i_range(s, 0x63C3, 0x63C4 + 1);
	i_single(s, 0x63C6);
	i_single(s, 0x63C9);
	i_range(s, 0x63CF, 0x63D0 + 1);
	i_single(s, 0x63D2);
	i_single(s, 0x63D6);
	i_range(s, 0x63DA, 0x63DB + 1);
	i_single(s, 0x63E1);
	i_single(s, 0x63E3);
	i_single(s, 0x63E9);
	i_single(s, 0x63EE);
	i_range(s, 0x63F4, 0x63F6 + 1);
	i_single(s, 0x63FA);
	i_single(s, 0x6406);
	i_single(s, 0x640D);
	i_single(s, 0x640F);
	i_single(s, 0x6413);
	i_range(s, 0x6416, 0x6417 + 1);
	i_single(s, 0x641C);
	i_single(s, 0x6426);
	i_single(s, 0x6428);
	i_range(s, 0x642C, 0x642D + 1);
	i_single(s, 0x6434);
	i_single(s, 0x6436);
	i_single(s, 0x643A);
	i_single(s, 0x643E);
	i_single(s, 0x6442);
	i_single(s, 0x644E);
	i_single(s, 0x6458);
	i_single(s, 0x6460);
	i_single(s, 0x6467);
	i_single(s, 0x6469);
	i_single(s, 0x646F);
	i_single(s, 0x6476);
	i_single(s, 0x6478);
	i_single(s, 0x647A);
	i_single(s, 0x6483);
	i_single(s, 0x6488);
	i_range(s, 0x6492, 0x6493 + 1);
	i_single(s, 0x6495);
	i_single(s, 0x649A);
	i_range(s, 0x649D, 0x649E + 1);
	i_range(s, 0x64A4, 0x64A5 + 1);
	i_single(s, 0x64A9);
	i_single(s, 0x64AB);
	i_range(s, 0x64AD, 0x64AE + 1);
	i_single(s, 0x64B0);
	i_single(s, 0x64B2);
	i_single(s, 0x64B9);
	i_range(s, 0x64BB, 0x64BC + 1);
	i_range(s, 0x64C1, 0x64C2 + 1);
	i_single(s, 0x64C5);
	i_single(s, 0x64C7);
	i_range(s, 0x64CD, 0x64CE + 1);
	i_single(s, 0x64D2);
	i_single(s, 0x64D4);
	i_single(s, 0x64D8);
	i_single(s, 0x64DA);
	i_range(s, 0x64E0, 0x64E3 + 1);
	i_range(s, 0x64E6, 0x64E7 + 1);
	i_single(s, 0x64EC);
	i_single(s, 0x64EF);
	i_range(s, 0x64F1, 0x64F2 + 1);
	i_single(s, 0x64F4);
	i_single(s, 0x64F6);
	i_single(s, 0x64FA);
	i_range(s, 0x64FD, 0x64FE + 1);
	i_single(s, 0x6500);
	i_single(s, 0x6505);
	i_single(s, 0x6518);
	i_range(s, 0x651C, 0x651D + 1);
	i_range(s, 0x6523, 0x6524 + 1);
	i_range(s, 0x652A, 0x652C + 1);
	i_single(s, 0x652F);
	i_range(s, 0x6534, 0x6539 + 1);
	i_single(s, 0x653B);
	i_range(s, 0x653E, 0x653F + 1);
	i_single(s, 0x6545);
	i_single(s, 0x6548);
	i_range(s, 0x654D, 0x654F + 1);
	i_single(s, 0x6551);
	i_range(s, 0x6555, 0x6559 + 1);
	i_range(s, 0x655D, 0x655E + 1);
	i_range(s, 0x6562, 0x6563 + 1);
	i_single(s, 0x6566);
	i_single(s, 0x656C);
	i_single(s, 0x6570);
	i_single(s, 0x6572);
	i_range(s, 0x6574, 0x6575 + 1);
	i_range(s, 0x6577, 0x6578 + 1);
	i_range(s, 0x6582, 0x6583 + 1);
	i_range(s, 0x6587, 0x6589 + 1);
	i_single(s, 0x658C);
	i_single(s, 0x658E);
	i_range(s, 0x6590, 0x6591 + 1);
	i_single(s, 0x6597);
	i_single(s, 0x6599);
	i_range(s, 0x659B, 0x659C + 1);
	i_single(s, 0x659F);
	i_single(s, 0x65A1);
	i_range(s, 0x65A4, 0x65A5 + 1);
	i_single(s, 0x65A7);
	i_range(s, 0x65AB, 0x65AD + 1);
	i_range(s, 0x65AF, 0x65B0 + 1);
	i_single(s, 0x65B7);
	i_single(s, 0x65B9);
	i_range(s, 0x65BC, 0x65BD + 1);
	i_single(s, 0x65C1);
	i_range(s, 0x65C3, 0x65C6 + 1);
	i_range(s, 0x65CB, 0x65CC + 1);
	i_single(s, 0x65CF);
	i_single(s, 0x65D2);
	i_single(s, 0x65D7);
	i_single(s, 0x65D9);
	i_single(s, 0x65DB);
	i_range(s, 0x65E0, 0x65E2 + 1);
	i_range(s, 0x65E5, 0x65E9 + 1);
	i_range(s, 0x65EC, 0x65ED + 1);
	i_single(s, 0x65F1);
	i_range(s, 0x65FA, 0x65FB + 1);
	i_single(s, 0x6600);
	i_range(s, 0x6602, 0x6603 + 1);
	i_range(s, 0x6606, 0x6607 + 1);
	i_range(s, 0x6609, 0x660A + 1);
	i_single(s, 0x660C);
	i_range(s, 0x660E, 0x660F + 1);
	i_range(s, 0x6613, 0x6615 + 1);
	i_single(s, 0x661C);
	i_range(s, 0x661E, 0x6620 + 1);
	i_range(s, 0x6624, 0x6625 + 1);
	i_range(s, 0x6627, 0x6628 + 1);
	i_range(s, 0x662D, 0x662F + 1);
	i_single(s, 0x6631);
	i_range(s, 0x6634, 0x6636 + 1);
	i_range(s, 0x663B, 0x663C + 1);
	i_single(s, 0x663F);
	i_range(s, 0x6641, 0x6644 + 1);
	i_single(s, 0x6649);
	i_single(s, 0x664B);
	i_single(s, 0x664F);
	i_single(s, 0x6652);
	i_single(s, 0x6657);
	i_single(s, 0x6659);
	i_range(s, 0x665D, 0x665F + 1);
	i_single(s, 0x6662);
	i_range(s, 0x6664, 0x6669 + 1);
	i_range(s, 0x666E, 0x6670 + 1);
	i_range(s, 0x6673, 0x6674 + 1);
	i_single(s, 0x6676);
	i_single(s, 0x667A);
	i_single(s, 0x6681);
	i_range(s, 0x6683, 0x6684 + 1);
	i_range(s, 0x6687, 0x6689 + 1);
	i_single(s, 0x668E);
	i_single(s, 0x6691);
	i_range(s, 0x6696, 0x6699 + 1);
	i_single(s, 0x669D);
	i_single(s, 0x66A0);
	i_single(s, 0x66A2);
	i_single(s, 0x66A6);
	i_single(s, 0x66AB);
	i_single(s, 0x66AE);
	i_single(s, 0x66B2);
	i_single(s, 0x66B4);
	i_range(s, 0x66B8, 0x66B9 + 1);
	i_single(s, 0x66BC);
	i_range(s, 0x66BE, 0x66BF + 1);
	i_single(s, 0x66C1);
	i_single(s, 0x66C4);
	i_single(s, 0x66C7);
	i_single(s, 0x66C9);
	i_single(s, 0x66D6);
	i_range(s, 0x66D9, 0x66DA + 1);
	i_range(s, 0x66DC, 0x66DD + 1);
	i_single(s, 0x66E0);
	i_single(s, 0x66E6);
	i_single(s, 0x66E9);
	i_single(s, 0x66F0);
	i_range(s, 0x66F2, 0x66F5 + 1);
	i_range(s, 0x66F7, 0x6700 + 1);
	i_single(s, 0x6703);
	i_range(s, 0x6708, 0x6709 + 1);
	i_single(s, 0x670B);
	i_range(s, 0x670D, 0x670F + 1);
	i_range(s, 0x6714, 0x6717 + 1);
	i_single(s, 0x671B);
	i_range(s, 0x671D, 0x671F + 1);
	i_range(s, 0x6726, 0x6728 + 1);
	i_range(s, 0x672A, 0x672E + 1);
	i_single(s, 0x6731);
	i_single(s, 0x6734);
	i_range(s, 0x6736, 0x6738 + 1);
	i_single(s, 0x673A);
	i_single(s, 0x673D);
	i_single(s, 0x673F);
	i_single(s, 0x6741);
	i_single(s, 0x6746);
	i_single(s, 0x6749);
	i_range(s, 0x674E, 0x6751 + 1);
	i_single(s, 0x6753);
	i_single(s, 0x6756);
	i_single(s, 0x6759);
	i_single(s, 0x675C);
	i_range(s, 0x675E, 0x6766 + 1);
	i_single(s, 0x676A);
	i_single(s, 0x676D);
	i_range(s, 0x676F, 0x6773 + 1);
	i_single(s, 0x6775);
	i_single(s, 0x6777);
	i_single(s, 0x677C);
	i_range(s, 0x677E, 0x677F + 1);
	i_single(s, 0x6785);
	i_single(s, 0x6787);
	i_single(s, 0x6789);
	i_range(s, 0x678B, 0x678C + 1);
	i_single(s, 0x6790);
	i_single(s, 0x6795);
	i_single(s, 0x6797);
	i_single(s, 0x679A);
	i_range(s, 0x679C, 0x679D + 1);
	i_range(s, 0x67A0, 0x67A2 + 1);
	i_single(s, 0x67A6);
	i_single(s, 0x67A9);
	i_single(s, 0x67AF);
	i_range(s, 0x67B3, 0x67B4 + 1);
	i_range(s, 0x67B6, 0x67B9 + 1);
	i_single(s, 0x67BB);
	i_range(s, 0x67C0, 0x67C1 + 1);
	i_single(s, 0x67C4);
	i_single(s, 0x67C6);
	i_single(s, 0x67CA);
	i_range(s, 0x67CE, 0x67D1 + 1);
	i_range(s, 0x67D3, 0x67D4 + 1);
	i_single(s, 0x67D8);
	i_single(s, 0x67DA);
	i_range(s, 0x67DD, 0x67DE + 1);
	i_single(s, 0x67E2);
	i_single(s, 0x67E4);
	i_single(s, 0x67E7);
	i_single(s, 0x67E9);
	i_single(s, 0x67EC);
	i_range(s, 0x67EE, 0x67EF + 1);
	i_single(s, 0x67F1);
	i_range(s, 0x67F3, 0x67F5 + 1);
	i_single(s, 0x67FB);
	i_range(s, 0x67FE, 0x67FF + 1);
	i_range(s, 0x6801, 0x6804 + 1);
	i_single(s, 0x6813);
	i_range(s, 0x6816, 0x6817 + 1);
	i_single(s, 0x681E);
	i_range(s, 0x6821, 0x6822 + 1);
	i_range(s, 0x6829, 0x682B + 1);
	i_single(s, 0x6832);
	i_single(s, 0x6834);
	i_range(s, 0x6838, 0x6839 + 1);
	i_range(s, 0x683C, 0x683D + 1);
	i_range(s, 0x6840, 0x6844 + 1);
	i_single(s, 0x6846);
	i_single(s, 0x6848);
	i_range(s, 0x684D, 0x684E + 1);
	i_range(s, 0x6850, 0x6854 + 1);
	i_single(s, 0x6859);
	i_range(s, 0x685C, 0x685D + 1);
	i_single(s, 0x685F);
	i_single(s, 0x6863);
	i_single(s, 0x6867);
	i_single(s, 0x6874);
	i_range(s, 0x6876, 0x6877 + 1);
	i_range(s, 0x687E, 0x687F + 1);
	i_single(s, 0x6881);
	i_single(s, 0x6883);
	i_single(s, 0x6885);
	i_single(s, 0x688D);
	i_single(s, 0x688F);
	i_range(s, 0x6893, 0x6894 + 1);
	i_single(s, 0x6897);
	i_single(s, 0x689B);
	i_single(s, 0x689D);
	i_range(s, 0x689F, 0x68A0 + 1);
	i_single(s, 0x68A2);
	i_range(s, 0x68A6, 0x68A8 + 1);
	i_single(s, 0x68AD);
	i_range(s, 0x68AF, 0x68B1 + 1);
	i_single(s, 0x68B3);
	i_range(s, 0x68B5, 0x68B6 + 1);
	i_range(s, 0x68B9, 0x68BA + 1);
	i_single(s, 0x68BC);
	i_single(s, 0x68C4);
	i_single(s, 0x68C6);
	i_range(s, 0x68C8, 0x68CB + 1);
	i_single(s, 0x68CD);
	i_single(s, 0x68CF);
	i_single(s, 0x68D2);
	i_range(s, 0x68D4, 0x68D5 + 1);
	i_range(s, 0x68D7, 0x68D8 + 1);
	i_single(s, 0x68DA);
	i_range(s, 0x68DF, 0x68E1 + 1);
	i_single(s, 0x68E3);
	i_single(s, 0x68E7);
	i_range(s, 0x68EE, 0x68EF + 1);
	i_single(s, 0x68F2);
	i_range(s, 0x68F9, 0x68FA + 1);
	i_range(s, 0x6900, 0x6901 + 1);
	i_range(s, 0x6904, 0x6905 + 1);
	i_single(s, 0x6908);
	i_range(s, 0x690B, 0x690F + 1);
	i_single(s, 0x6912);
	i_range(s, 0x6919, 0x691C + 1);
	i_range(s, 0x6921, 0x6923 + 1);
	i_range(s, 0x6925, 0x6926 + 1);
	i_single(s, 0x6928);
	i_single(s, 0x692A);
	i_single(s, 0x6930);
	i_single(s, 0x6934);
	i_single(s, 0x6936);
	i_single(s, 0x6939);
	i_single(s, 0x693D);
	i_single(s, 0x693F);
	i_single(s, 0x694A);
	i_range(s, 0x6953, 0x6955 + 1);
	i_range(s, 0x6959, 0x695A + 1);
	i_range(s, 0x695C, 0x695E + 1);
	i_range(s, 0x6960, 0x6962 + 1);
	i_single(s, 0x6968);
	i_range(s, 0x696A, 0x696B + 1);
	i_range(s, 0x696D, 0x696F + 1);
	i_range(s, 0x6973, 0x6975 + 1);
	i_range(s, 0x6977, 0x6979 + 1);
	i_range(s, 0x697C, 0x697E + 1);
	i_range(s, 0x6981, 0x6982 + 1);
	i_single(s, 0x698A);
	i_single(s, 0x698E);
	i_single(s, 0x6991);
	i_range(s, 0x6994, 0x6995 + 1);
	i_single(s, 0x6998);
	i_range(s, 0x699B, 0x699C + 1);
	i_single(s, 0x69A0);
	i_single(s, 0x69A7);
	i_single(s, 0x69AE);
	i_range(s, 0x69B1, 0x69B2 + 1);
	i_single(s, 0x69B4);
	i_single(s, 0x69BB);
	i_range(s, 0x69BE, 0x69BF + 1);
	i_single(s, 0x69C1);
	i_single(s, 0x69C3);
	i_single(s, 0x69C7);
	i_range(s, 0x69CA, 0x69CE + 1);
	i_single(s, 0x69D0);
	i_single(s, 0x69D3);
	i_range(s, 0x69D8, 0x69D9 + 1);
	i_range(s, 0x69DD, 0x69DE + 1);
	i_single(s, 0x69E2);
	i_range(s, 0x69E7, 0x69E8 + 1);
	i_single(s, 0x69EB);
	i_single(s, 0x69ED);
	i_single(s, 0x69F2);
	i_single(s, 0x69F9);
	i_single(s, 0x69FB);
	i_single(s, 0x69FD);
	i_single(s, 0x69FF);
	i_single(s, 0x6A02);
	i_single(s, 0x6A05);
	i_range(s, 0x6A0A, 0x6A0C + 1);
	i_range(s, 0x6A12, 0x6A14 + 1);
	i_single(s, 0x6A17);
	i_single(s, 0x6A19);
	i_single(s, 0x6A1B);
	i_range(s, 0x6A1E, 0x6A1F + 1);
	i_range(s, 0x6A21, 0x6A23 + 1);
	i_range(s, 0x6A29, 0x6A2B + 1);
	i_single(s, 0x6A2E);
	i_single(s, 0x6A30);
	i_range(s, 0x6A35, 0x6A36 + 1);
	i_range(s, 0x6A38, 0x6A3A + 1);
	i_single(s, 0x6A3D);
	i_single(s, 0x6A44);
	i_range(s, 0x6A46, 0x6A48 + 1);
	i_single(s, 0x6A4B);
	i_range(s, 0x6A58, 0x6A59 + 1);
	i_single(s, 0x6A5F);
	i_range(s, 0x6A61, 0x6A62 + 1);
	i_single(s, 0x6A66);
	i_single(s, 0x6A6B);
	i_range(s, 0x6A72, 0x6A73 + 1);
	i_single(s, 0x6A78);
	i_range(s, 0x6A7E, 0x6A80 + 1);
	i_single(s, 0x6A84);
	i_range(s, 0x6A8D, 0x6A8E + 1);
	i_single(s, 0x6A90);
	i_single(s, 0x6A97);
	i_single(s, 0x6A9C);
	i_single(s, 0x6AA0);
	i_range(s, 0x6AA2, 0x6AA3 + 1);
	i_single(s, 0x6AAA);
	i_single(s, 0x6AAC);
	i_single(s, 0x6AAE);
	i_single(s, 0x6AB3);
	i_single(s, 0x6AB8);
	i_single(s, 0x6ABB);
	i_range(s, 0x6AC1, 0x6AC3 + 1);
	i_single(s, 0x6AD1);
	i_single(s, 0x6AD3);
	i_range(s, 0x6ADA, 0x6ADB + 1);
	i_range(s, 0x6ADE, 0x6ADF + 1);
	i_single(s, 0x6AE2);
	i_single(s, 0x6AE4);
	i_single(s, 0x6AE8);
	i_single(s, 0x6AEA);
	i_range(s, 0x6AFA, 0x6AFB + 1);
	i_range(s, 0x6B04, 0x6B05 + 1);
	i_single(s, 0x6B0A);
	i_single(s, 0x6B12);
	i_single(s, 0x6B16);
	i_single(s, 0x6B1D);
	i_range(s, 0x6B1F, 0x6B21 + 1);
	i_single(s, 0x6B23);
	i_single(s, 0x6B27);
	i_single(s, 0x6B32);
	i_range(s, 0x6B37, 0x6B3A + 1);
	i_range(s, 0x6B3D, 0x6B3E + 1);
	i_single(s, 0x6B43);
	i_single(s, 0x6B47);
	i_single(s, 0x6B49);
	i_single(s, 0x6B4C);
	i_single(s, 0x6B4E);
	i_single(s, 0x6B50);
	i_range(s, 0x6B53, 0x6B54 + 1);
	i_single(s, 0x6B59);
	i_single(s, 0x6B5B);
	i_single(s, 0x6B5F);
	i_range(s, 0x6B61, 0x6B64 + 1);
	i_single(s, 0x6B66);
	i_range(s, 0x6B69, 0x6B6A + 1);
	i_single(s, 0x6B6F);
	i_range(s, 0x6B73, 0x6B74 + 1);
	i_range(s, 0x6B78, 0x6B79 + 1);
	i_single(s, 0x6B7B);
	i_range(s, 0x6B7F, 0x6B80 + 1);
	i_range(s, 0x6B83, 0x6B84 + 1);
	i_single(s, 0x6B86);
	i_range(s, 0x6B89, 0x6B8B + 1);
	i_single(s, 0x6B8D);
	i_range(s, 0x6B95, 0x6B96 + 1);
	i_single(s, 0x6B98);
	i_single(s, 0x6B9E);
	i_single(s, 0x6BA4);
	i_range(s, 0x6BAA, 0x6BAB + 1);
	i_single(s, 0x6BAF);
	i_range(s, 0x6BB1, 0x6BB5 + 1);
	i_single(s, 0x6BB7);
	i_range(s, 0x6BBA, 0x6BBC + 1);
	i_range(s, 0x6BBF, 0x6BC0 + 1);
	i_range(s, 0x6BC5, 0x6BC6 + 1);
	i_single(s, 0x6BCB);
	i_range(s, 0x6BCD, 0x6BCE + 1);
	i_range(s, 0x6BD2, 0x6BD4 + 1);
	i_single(s, 0x6BD6);
	i_single(s, 0x6BD8);
	i_single(s, 0x6BDB);
	i_single(s, 0x6BDF);
	i_range(s, 0x6BEB, 0x6BEC + 1);
	i_single(s, 0x6BEF);
	i_single(s, 0x6BF3);
	i_single(s, 0x6C08);
	i_single(s, 0x6C0F);
	i_single(s, 0x6C11);
	i_range(s, 0x6C13, 0x6C14 + 1);
	i_single(s, 0x6C17);
	i_single(s, 0x6C1B);
	i_range(s, 0x6C23, 0x6C24 + 1);
	i_single(s, 0x6C34);
	i_range(s, 0x6C37, 0x6C38 + 1);
	i_range(s, 0x6C3E, 0x6C42 + 1);
	i_single(s, 0x6C4E);
	i_single(s, 0x6C50);
	i_single(s, 0x6C55);
	i_single(s, 0x6C57);
	i_single(s, 0x6C5A);
	i_range(s, 0x6C5C, 0x6C60 + 1);
	i_single(s, 0x6C62);
	i_single(s, 0x6C68);
	i_single(s, 0x6C6A);
	i_range(s, 0x6C6F, 0x6C70 + 1);
	i_range(s, 0x6C72, 0x6C73 + 1);
	i_single(s, 0x6C7A);
	i_range(s, 0x6C7D, 0x6C7E + 1);
	i_range(s, 0x6C81, 0x6C83 + 1);
	i_single(s, 0x6C86);
	i_single(s, 0x6C88);
	i_range(s, 0x6C8C, 0x6C8D + 1);
	i_single(s, 0x6C90);
	i_range(s, 0x6C92, 0x6C93 + 1);
	i_single(s, 0x6C96);
	i_range(s, 0x6C99, 0x6C9B + 1);
	i_range(s, 0x6CA1, 0x6CA2 + 1);
	i_single(s, 0x6CAB);
	i_single(s, 0x6CAE);
	i_single(s, 0x6CB1);
	i_single(s, 0x6CB3);
	i_range(s, 0x6CB8, 0x6CBF + 1);
	i_single(s, 0x6CC1);
	i_range(s, 0x6CC4, 0x6CC5 + 1);
	i_range(s, 0x6CC9, 0x6CCA + 1);
	i_single(s, 0x6CCC);
	i_single(s, 0x6CD3);
	i_single(s, 0x6CD5);
	i_single(s, 0x6CD7);
	i_range(s, 0x6CD9, 0x6CDB + 1);
	i_single(s, 0x6CDD);
	i_range(s, 0x6CE1, 0x6CE3 + 1);
	i_single(s, 0x6CE5);
	i_single(s, 0x6CE8);
	i_single(s, 0x6CEA);
	i_range(s, 0x6CEF, 0x6CF1 + 1);
	i_single(s, 0x6CF3);
	i_single(s, 0x6D04);
	i_range(s, 0x6D0B, 0x6D0C + 1);
	i_single(s, 0x6D12);
	i_single(s, 0x6D17);
	i_single(s, 0x6D19);
	i_single(s, 0x6D1B);
	i_range(s, 0x6D1E, 0x6D1F + 1);
	i_single(s, 0x6D25);
	i_range(s, 0x6D29, 0x6D2B + 1);
	i_range(s, 0x6D32, 0x6D33 + 1);
	i_range(s, 0x6D35, 0x6D36 + 1);
	i_single(s, 0x6D38);
	i_single(s, 0x6D3B);
	i_range(s, 0x6D3D, 0x6D3E + 1);
	i_single(s, 0x6D41);
	i_range(s, 0x6D44, 0x6D45 + 1);
	i_range(s, 0x6D59, 0x6D5A + 1);
	i_single(s, 0x6D5C);
	i_range(s, 0x6D63, 0x6D64 + 1);
	i_single(s, 0x6D66);
	i_range(s, 0x6D69, 0x6D6A + 1);
	i_single(s, 0x6D6C);
	i_range(s, 0x6D6E, 0x6D6F + 1);
	i_single(s, 0x6D74);
	i_range(s, 0x6D77, 0x6D79 + 1);
	i_single(s, 0x6D85);
	i_range(s, 0x6D87, 0x6D88 + 1);
	i_single(s, 0x6D8C);
	i_single(s, 0x6D8E);
	i_single(s, 0x6D93);
	i_range(s, 0x6D95, 0x6D96 + 1);
	i_single(s, 0x6D99);
	i_range(s, 0x6D9B, 0x6D9C + 1);
	i_single(s, 0x6DAC);
	i_single(s, 0x6DAF);
	i_single(s, 0x6DB2);
	i_single(s, 0x6DB5);
	i_single(s, 0x6DB8);
	i_single(s, 0x6DBC);
	i_single(s, 0x6DC0);
	i_range(s, 0x6DC5, 0x6DC7 + 1);
	i_range(s, 0x6DCB, 0x6DCC + 1);
	i_single(s, 0x6DCF);
	i_range(s, 0x6DD1, 0x6DD2 + 1);
	i_single(s, 0x6DD5);
	i_range(s, 0x6DD8, 0x6DD9 + 1);
	i_single(s, 0x6DDE);
	i_single(s, 0x6DE1);
	i_single(s, 0x6DE4);
	i_single(s, 0x6DE6);
	i_single(s, 0x6DE8);
	i_range(s, 0x6DEA, 0x6DEC + 1);
	i_single(s, 0x6DEE);
	i_range(s, 0x6DF1, 0x6DF3 + 1);
	i_single(s, 0x6DF5);
	i_range(s, 0x6DF7, 0x6DFC + 1);
	i_single(s, 0x6E05);
	i_range(s, 0x6E07, 0x6E0B + 1);
	i_single(s, 0x6E13);
	i_single(s, 0x6E15);
	i_range(s, 0x6E19, 0x6E1B + 1);
	i_single(s, 0x6E1D);
	i_range(s, 0x6E1F, 0x6E21 + 1);
	i_range(s, 0x6E23, 0x6E27 + 1);
	i_single(s, 0x6E29);
	i_range(s, 0x6E2B, 0x6E2F + 1);
	i_range(s, 0x6E38, 0x6E3A + 1);
	i_single(s, 0x6E3C);
	i_single(s, 0x6E3E);
	i_single(s, 0x6E43);
	i_single(s, 0x6E4A);
	i_range(s, 0x6E4D, 0x6E4E + 1);
	i_single(s, 0x6E56);
	i_single(s, 0x6E58);
	i_range(s, 0x6E5B, 0x6E5C + 1);
	i_single(s, 0x6E5F);
	i_single(s, 0x6E67);
	i_single(s, 0x6E6B);
	i_range(s, 0x6E6E, 0x6E6F + 1);
	i_single(s, 0x6E72);
	i_single(s, 0x6E76);
	i_range(s, 0x6E7E, 0x6E80 + 1);
	i_single(s, 0x6E82);
	i_single(s, 0x6E8C);
	i_range(s, 0x6E8F, 0x6E90 + 1);
	i_single(s, 0x6E96);
	i_single(s, 0x6E98);
	i_range(s, 0x6E9C, 0x6E9D + 1);
	i_single(s, 0x6E9F);
	i_single(s, 0x6EA2);
	i_single(s, 0x6EA5);
	i_single(s, 0x6EAA);
	i_single(s, 0x6EAF);
	i_single(s, 0x6EB2);
	i_range(s, 0x6EB6, 0x6EB7 + 1);
	i_single(s, 0x6EBA);
	i_single(s, 0x6EBD);
	i_single(s, 0x6EBF);
	i_single(s, 0x6EC2);
	i_range(s, 0x6EC4, 0x6EC5 + 1);
	i_single(s, 0x6EC9);
	i_range(s, 0x6ECB, 0x6ECC + 1);
	i_single(s, 0x6ED1);
	i_range(s, 0x6ED3, 0x6ED5 + 1);
	i_range(s, 0x6EDD, 0x6EDE + 1);
	i_single(s, 0x6EEC);
	i_single(s, 0x6EEF);
	i_single(s, 0x6EF2);
	i_single(s, 0x6EF4);
	i_range(s, 0x6EF7, 0x6EF8 + 1);
	i_range(s, 0x6EFE, 0x6EFF + 1);
	i_range(s, 0x6F01, 0x6F02 + 1);
	i_single(s, 0x6F06);
	i_single(s, 0x6F09);
	i_single(s, 0x6F0F);
	i_single(s, 0x6F11);
	i_range(s, 0x6F13, 0x6F15 + 1);
	i_single(s, 0x6F20);
	i_range(s, 0x6F22, 0x6F23 + 1);
	i_range(s, 0x6F2B, 0x6F2C + 1);
	i_range(s, 0x6F31, 0x6F32 + 1);
	i_single(s, 0x6F38);
	i_range(s, 0x6F3E, 0x6F3F + 1);
	i_single(s, 0x6F41);
	i_single(s, 0x6F45);
	i_single(s, 0x6F54);
	i_single(s, 0x6F58);
	i_range(s, 0x6F5B, 0x6F5C + 1);
	i_single(s, 0x6F5F);
	i_single(s, 0x6F64);
	i_single(s, 0x6F66);
	i_range(s, 0x6F6D, 0x6F70 + 1);
	i_single(s, 0x6F74);
	i_single(s, 0x6F78);
	i_single(s, 0x6F7A);
	i_single(s, 0x6F7C);
	i_range(s, 0x6F80, 0x6F82 + 1);
	i_single(s, 0x6F84);
	i_single(s, 0x6F86);
	i_single(s, 0x6F88);
	i_single(s, 0x6F8E);
	i_single(s, 0x6F91);
	i_single(s, 0x6F97);
	i_single(s, 0x6FA1);
	i_range(s, 0x6FA3, 0x6FA4 + 1);
	i_single(s, 0x6FAA);
	i_single(s, 0x6FB1);
	i_single(s, 0x6FB3);
	i_single(s, 0x6FB5);
	i_single(s, 0x6FB9);
	i_range(s, 0x6FC0, 0x6FC3 + 1);
	i_single(s, 0x6FC6);
	i_range(s, 0x6FD4, 0x6FD5 + 1);
	i_single(s, 0x6FD8);
	i_single(s, 0x6FDB);
	i_range(s, 0x6FDF, 0x6FE1 + 1);
	i_single(s, 0x6FE4);
	i_range(s, 0x6FEB, 0x6FEC + 1);
	i_range(s, 0x6FEE, 0x6FEF + 1);
	i_single(s, 0x6FF1);
	i_single(s, 0x6FF3);
	i_range(s, 0x6FF5, 0x6FF6 + 1);
	i_single(s, 0x6FFA);
	i_single(s, 0x6FFE);
	i_single(s, 0x7001);
	i_single(s, 0x7005);
	i_single(s, 0x7007);
	i_single(s, 0x7009);
	i_single(s, 0x700B);
	i_single(s, 0x700F);
	i_single(s, 0x7011);
	i_single(s, 0x7015);
	i_single(s, 0x7018);
	i_range(s, 0x701A, 0x701B + 1);
	i_range(s, 0x701D, 0x701F + 1);
	i_range(s, 0x7026, 0x7028 + 1);
	i_single(s, 0x702C);
	i_single(s, 0x7030);
	i_single(s, 0x7032);
	i_single(s, 0x703E);
	i_single(s, 0x704C);
	i_single(s, 0x7051);
	i_single(s, 0x7058);
	i_single(s, 0x7063);
	i_single(s, 0x706B);
	i_range(s, 0x706F, 0x7070 + 1);
	i_single(s, 0x7078);
	i_range(s, 0x707C, 0x707D + 1);
	i_single(s, 0x7085);
	i_range(s, 0x7089, 0x708A + 1);
	i_single(s, 0x708E);
	i_single(s, 0x7092);
	i_single(s, 0x7099);
	i_range(s, 0x70AB, 0x70AF + 1);
	i_single(s, 0x70B3);
	i_range(s, 0x70B8, 0x70BB + 1);
	i_single(s, 0x70C8);
	i_single(s, 0x70CB);
	i_single(s, 0x70CF);
	i_single(s, 0x70D9);
	i_single(s, 0x70DD);
	i_single(s, 0x70DF);
	i_single(s, 0x70F1);
	i_single(s, 0x70F9);
	i_single(s, 0x70FD);
	i_single(s, 0x7104);
	i_single(s, 0x7109);
	i_single(s, 0x710F);
	i_single(s, 0x7114);
	i_range(s, 0x7119, 0x711A + 1);
	i_single(s, 0x711C);
	i_single(s, 0x7121);
	i_single(s, 0x7126);
	i_single(s, 0x7136);
	i_single(s, 0x713C);
	i_range(s, 0x7146, 0x7147 + 1);
	i_single(s, 0x7149);
	i_single(s, 0x714C);
	i_single(s, 0x714E);
	i_range(s, 0x7155, 0x7156 + 1);
	i_single(s, 0x7159);
	i_single(s, 0x715C);
	i_single(s, 0x7162);
	i_range(s, 0x7164, 0x7167 + 1);
	i_single(s, 0x7169);
	i_single(s, 0x716C);
	i_single(s, 0x716E);
	i_single(s, 0x717D);
	i_single(s, 0x7184);
	i_single(s, 0x7188);
	i_single(s, 0x718A);
	i_single(s, 0x718F);
	i_range(s, 0x7194, 0x7195 + 1);
	i_single(s, 0x7199);
	i_single(s, 0x719F);
	i_single(s, 0x71A8);
	i_single(s, 0x71AC);
	i_single(s, 0x71B1);
	i_single(s, 0x71B9);
	i_single(s, 0x71BE);
	i_single(s, 0x71C1);
	i_single(s, 0x71C3);
	i_range(s, 0x71C8, 0x71C9 + 1);
	i_single(s, 0x71CE);
	i_single(s, 0x71D0);
	i_single(s, 0x71D2);
	i_range(s, 0x71D4, 0x71D5 + 1);
	i_single(s, 0x71D7);
	i_range(s, 0x71DF, 0x71E0 + 1);
	i_range(s, 0x71E5, 0x71E7 + 1);
	i_range(s, 0x71EC, 0x71EE + 1);
	i_single(s, 0x71F5);
	i_single(s, 0x71F9);
	i_range(s, 0x71FB, 0x71FC + 1);
	i_range(s, 0x71FE, 0x71FF + 1);
	i_single(s, 0x7206);
	i_single(s, 0x720D);
	i_single(s, 0x7210);
	i_single(s, 0x721B);
	i_single(s, 0x7228);
	i_single(s, 0x722A);
	i_range(s, 0x722C, 0x722D + 1);
	i_single(s, 0x7230);
	i_single(s, 0x7232);
	i_range(s, 0x7235, 0x7236 + 1);
	i_range(s, 0x723A, 0x7240 + 1);
	i_range(s, 0x7246, 0x7248 + 1);
	i_range(s, 0x724B, 0x724C + 1);
	i_single(s, 0x7252);
	i_range(s, 0x7258, 0x7259 + 1);
	i_single(s, 0x725B);
	i_single(s, 0x725D);
	i_single(s, 0x725F);
	i_range(s, 0x7261, 0x7262 + 1);
	i_single(s, 0x7267);
	i_single(s, 0x7269);
	i_single(s, 0x7272);
	i_single(s, 0x7274);
	i_single(s, 0x7279);
	i_range(s, 0x727D, 0x727E + 1);
	i_range(s, 0x7280, 0x7282 + 1);
	i_single(s, 0x7287);
	i_single(s, 0x7292);
	i_single(s, 0x7296);
	i_single(s, 0x72A0);
	i_single(s, 0x72A2);
	i_single(s, 0x72A7);
	i_single(s, 0x72AC);
	i_single(s, 0x72AF);
	i_range(s, 0x72B1, 0x72B2 + 1);
	i_single(s, 0x72B6);
	i_single(s, 0x72B9);
	i_single(s, 0x72BE);
	i_range(s, 0x72C2, 0x72C4 + 1);
	i_single(s, 0x72C6);
	i_single(s, 0x72CE);
	i_single(s, 0x72D0);
	i_single(s, 0x72D2);
	i_single(s, 0x72D7);
	i_single(s, 0x72D9);
	i_single(s, 0x72DB);
	i_range(s, 0x72E0, 0x72E2 + 1);
	i_single(s, 0x72E9);
	i_range(s, 0x72EC, 0x72ED + 1);
	i_range(s, 0x72F7, 0x72F9 + 1);
	i_range(s, 0x72FC, 0x72FD + 1);
	i_single(s, 0x730A);
	i_range(s, 0x7316, 0x7317 + 1);
	i_range(s, 0x731B, 0x731D + 1);
	i_single(s, 0x731F);
	i_range(s, 0x7324, 0x7325 + 1);
	i_range(s, 0x7329, 0x732B + 1);
	i_range(s, 0x732E, 0x732F + 1);
	i_single(s, 0x7334);
	i_range(s, 0x7336, 0x7337 + 1);
	i_range(s, 0x733E, 0x733F + 1);
	i_range(s, 0x7344, 0x7345 + 1);
	i_range(s, 0x734E, 0x734F + 1);
	i_single(s, 0x7357);
	i_single(s, 0x7363);
	i_single(s, 0x7368);
	i_single(s, 0x736A);
	i_single(s, 0x7370);
	i_single(s, 0x7372);
	i_single(s, 0x7375);
	i_range(s, 0x7377, 0x7378 + 1);
	i_range(s, 0x737A, 0x737B + 1);
	i_single(s, 0x7384);
	i_single(s, 0x7387);
	i_single(s, 0x7389);
	i_single(s, 0x738B);
	i_single(s, 0x7396);
	i_single(s, 0x73A9);
	i_range(s, 0x73B2, 0x73B3 + 1);
	i_single(s, 0x73BB);
	i_single(s, 0x73BD);
	i_single(s, 0x73C0);
	i_single(s, 0x73C2);
	i_range(s, 0x73C8, 0x73CA + 1);
	i_range(s, 0x73CD, 0x73CE + 1);
	i_single(s, 0x73D2);
	i_single(s, 0x73D6);
	i_single(s, 0x73DE);
	i_single(s, 0x73E0);
	i_single(s, 0x73E3);
	i_single(s, 0x73E5);
	i_single(s, 0x73EA);
	i_range(s, 0x73ED, 0x73EE + 1);
	i_single(s, 0x73F1);
	i_single(s, 0x73F5);
	i_single(s, 0x73F8);
	i_single(s, 0x73FE);
	i_single(s, 0x7403);
	i_range(s, 0x7405, 0x7407 + 1);
	i_single(s, 0x7409);
	i_single(s, 0x7422);
	i_range(s, 0x7425, 0x7426 + 1);
	i_range(s, 0x7429, 0x742A + 1);
	i_single(s, 0x742E);
	i_range(s, 0x7432, 0x7436 + 1);
	i_single(s, 0x743A);
	i_single(s, 0x743F);
	i_single(s, 0x7441);
	i_single(s, 0x7455);
	i_range(s, 0x7459, 0x745C + 1);
	i_range(s, 0x745E, 0x7460 + 1);
	i_range(s, 0x7462, 0x7464 + 1);
	i_range(s, 0x7469, 0x746A + 1);
	i_range(s, 0x746F, 0x7470 + 1);
	i_single(s, 0x7473);
	i_single(s, 0x7476);
	i_single(s, 0x747E);
	i_single(s, 0x7483);
	i_single(s, 0x7489);
	i_single(s, 0x748B);
	i_range(s, 0x749E, 0x749F + 1);
	i_single(s, 0x74A2);
	i_single(s, 0x74A7);
	i_single(s, 0x74B0);
	i_single(s, 0x74BD);
	i_single(s, 0x74CA);
	i_single(s, 0x74CF);
	i_single(s, 0x74D4);
	i_single(s, 0x74DC);
	i_single(s, 0x74E0);
	i_range(s, 0x74E2, 0x74E3 + 1);
	i_range(s, 0x74E6, 0x74E7 + 1);
	i_single(s, 0x74E9);
	i_single(s, 0x74EE);
	i_range(s, 0x74F0, 0x74F2 + 1);
	i_range(s, 0x74F6, 0x74F8 + 1);
	i_single(s, 0x7501);
	i_range(s, 0x7503, 0x7505 + 1);
	i_range(s, 0x750C, 0x750E + 1);
	i_single(s, 0x7511);
	i_single(s, 0x7513);
	i_single(s, 0x7515);
	i_single(s, 0x7518);
	i_single(s, 0x751A);
	i_single(s, 0x751C);
	i_range(s, 0x751E, 0x751F + 1);
	i_single(s, 0x7523);
	i_range(s, 0x7525, 0x7526 + 1);
	i_single(s, 0x7528);
	i_range(s, 0x752B, 0x752C + 1);
	i_range(s, 0x752F, 0x7533 + 1);
	i_range(s, 0x7537, 0x7538 + 1);
	i_range(s, 0x753A, 0x753C + 1);
	i_single(s, 0x7544);
	i_single(s, 0x7546);
	i_range(s, 0x7549, 0x754D + 1);
	i_single(s, 0x754F);
	i_single(s, 0x7551);
	i_single(s, 0x7554);
	i_range(s, 0x7559, 0x755D + 1);
	i_single(s, 0x7560);
	i_single(s, 0x7562);
	i_range(s, 0x7564, 0x7567 + 1);
	i_range(s, 0x7569, 0x756B + 1);
	i_single(s, 0x756D);
	i_range(s, 0x756F, 0x7570 + 1);
	i_range(s, 0x7573, 0x7574 + 1);
	i_range(s, 0x7576, 0x7578 + 1);
	i_single(s, 0x757F);
	i_single(s, 0x7582);
	i_range(s, 0x7586, 0x7587 + 1);
	i_range(s, 0x7589, 0x758B + 1);
	i_range(s, 0x758E, 0x758F + 1);
	i_single(s, 0x7591);
	i_single(s, 0x7594);
	i_single(s, 0x759A);
	i_single(s, 0x759D);
	i_single(s, 0x75A3);
	i_single(s, 0x75A5);
	i_single(s, 0x75AB);
	i_range(s, 0x75B1, 0x75B3 + 1);
	i_single(s, 0x75B5);
	i_range(s, 0x75B8, 0x75B9 + 1);
	i_range(s, 0x75BC, 0x75BE + 1);
	i_range(s, 0x75C2, 0x75C3 + 1);
	i_single(s, 0x75C5);
	i_single(s, 0x75C7);
	i_single(s, 0x75CA);
	i_single(s, 0x75CD);
	i_single(s, 0x75D2);
	i_range(s, 0x75D4, 0x75D5 + 1);
	i_range(s, 0x75D8, 0x75D9 + 1);
	i_single(s, 0x75DB);
	i_single(s, 0x75DE);
	i_range(s, 0x75E2, 0x75E3 + 1);
	i_single(s, 0x75E9);
	i_single(s, 0x75F0);
	i_range(s, 0x75F2, 0x75F4 + 1);
	i_single(s, 0x75FA);
	i_single(s, 0x75FC);
	i_range(s, 0x75FE, 0x75FF + 1);
	i_single(s, 0x7601);
	i_single(s, 0x7609);
	i_single(s, 0x760B);
	i_single(s, 0x760D);
	i_range(s, 0x761F, 0x7622 + 1);
	i_single(s, 0x7624);
	i_single(s, 0x7627);
	i_single(s, 0x7630);
	i_single(s, 0x7634);
	i_single(s, 0x763B);
	i_single(s, 0x7642);
	i_range(s, 0x7646, 0x7648 + 1);
	i_single(s, 0x764C);
	i_single(s, 0x7652);
	i_single(s, 0x7656);
	i_single(s, 0x7658);
	i_single(s, 0x765C);
	i_range(s, 0x7661, 0x7662 + 1);
	i_range(s, 0x7667, 0x766A + 1);
	i_single(s, 0x766C);
	i_single(s, 0x7670);
	i_single(s, 0x7672);
	i_single(s, 0x7676);
	i_single(s, 0x7678);
	i_range(s, 0x767A, 0x767E + 1);
	i_single(s, 0x7680);
	i_range(s, 0x7682, 0x7684 + 1);
	i_range(s, 0x7686, 0x7688 + 1);
	i_single(s, 0x768B);
	i_single(s, 0x768E);
	i_single(s, 0x7690);
	i_single(s, 0x7693);
	i_single(s, 0x7696);
	i_range(s, 0x7699, 0x769C + 1);
	i_single(s, 0x769E);
	i_single(s, 0x76A6);
	i_single(s, 0x76AE);
	i_single(s, 0x76B0);
	i_single(s, 0x76B4);
	i_range(s, 0x76B7, 0x76BA + 1);
	i_single(s, 0x76BF);
	i_range(s, 0x76C2, 0x76C3 + 1);
	i_single(s, 0x76C6);
	i_single(s, 0x76C8);
	i_single(s, 0x76CA);
	i_single(s, 0x76CD);
	i_single(s, 0x76D2);
	i_range(s, 0x76D6, 0x76D7 + 1);
	i_range(s, 0x76DB, 0x76DC + 1);
	i_range(s, 0x76DE, 0x76DF + 1);
	i_single(s, 0x76E1);
	i_range(s, 0x76E3, 0x76E5 + 1);
	i_single(s, 0x76E7);
	i_single(s, 0x76EA);
	i_single(s, 0x76EE);
	i_single(s, 0x76F2);
	i_single(s, 0x76F4);
	i_single(s, 0x76F8);
	i_single(s, 0x76FB);
	i_single(s, 0x76FE);
	i_single(s, 0x7701);
	i_single(s, 0x7704);
	i_range(s, 0x7707, 0x7709 + 1);
	i_range(s, 0x770B, 0x770C + 1);
	i_single(s, 0x771B);
	i_range(s, 0x771E, 0x7720 + 1);
	i_range(s, 0x7724, 0x7726 + 1);
	i_single(s, 0x7729);
	i_range(s, 0x7737, 0x7738 + 1);
	i_single(s, 0x773A);
	i_single(s, 0x773C);
	i_single(s, 0x7740);
	i_range(s, 0x7746, 0x7747 + 1);
	i_range(s, 0x775A, 0x775B + 1);
	i_single(s, 0x7761);
	i_single(s, 0x7763);
	i_range(s, 0x7765, 0x7766 + 1);
	i_single(s, 0x7768);
	i_single(s, 0x776B);
	i_single(s, 0x7779);
	i_range(s, 0x777E, 0x777F + 1);
	i_single(s, 0x778B);
	i_single(s, 0x778E);
	i_single(s, 0x7791);
	i_single(s, 0x779E);
	i_single(s, 0x77A0);
	i_single(s, 0x77A5);
	i_range(s, 0x77AC, 0x77AD + 1);
	i_single(s, 0x77B0);
	i_single(s, 0x77B3);
	i_single(s, 0x77B6);
	i_single(s, 0x77B9);
	i_range(s, 0x77BB, 0x77BD + 1);
	i_single(s, 0x77BF);
	i_single(s, 0x77C7);
	i_single(s, 0x77CD);
	i_single(s, 0x77D7);
	i_range(s, 0x77DA, 0x77DC + 1);
	i_range(s, 0x77E2, 0x77E3 + 1);
	i_single(s, 0x77E5);
	i_single(s, 0x77E7);
	i_single(s, 0x77E9);
	i_range(s, 0x77ED, 0x77EF + 1);
	i_single(s, 0x77F3);
	i_single(s, 0x77FC);
	i_single(s, 0x7802);
	i_single(s, 0x780C);
	i_single(s, 0x7812);
	i_range(s, 0x7814, 0x7815 + 1);
	i_range(s, 0x7820, 0x7821 + 1);
	i_range(s, 0x7825, 0x7827 + 1);
	i_single(s, 0x7832);
	i_single(s, 0x7834);
	i_single(s, 0x783A);
	i_single(s, 0x783F);
	i_single(s, 0x7845);
	i_single(s, 0x784E);
	i_single(s, 0x785D);
	i_single(s, 0x7864);
	i_range(s, 0x786B, 0x786C + 1);
	i_single(s, 0x786F);
	i_single(s, 0x7872);
	i_single(s, 0x7874);
	i_single(s, 0x787A);
	i_single(s, 0x787C);
	i_single(s, 0x7881);
	i_range(s, 0x7886, 0x7887 + 1);
	i_range(s, 0x788C, 0x788E + 1);
	i_single(s, 0x7891);
	i_single(s, 0x7893);
	i_single(s, 0x7895);
	i_single(s, 0x7897);
	i_single(s, 0x789A);
	i_single(s, 0x78A3);
	i_single(s, 0x78A7);
	i_range(s, 0x78A9, 0x78AA + 1);
	i_single(s, 0x78AF);
	i_single(s, 0x78B5);
	i_single(s, 0x78BA);
	i_single(s, 0x78BC);
	i_single(s, 0x78BE);
	i_single(s, 0x78C1);
	i_range(s, 0x78C5, 0x78C6 + 1);
	i_range(s, 0x78CA, 0x78CB + 1);
	i_range(s, 0x78D0, 0x78D1 + 1);
	i_single(s, 0x78D4);
	i_single(s, 0x78DA);
	i_range(s, 0x78E7, 0x78E8 + 1);
	i_single(s, 0x78EC);
	i_single(s, 0x78EF);
	i_single(s, 0x78F4);
	i_single(s, 0x78FD);
	i_single(s, 0x7901);
	i_single(s, 0x7907);
	i_single(s, 0x790E);
	i_range(s, 0x7911, 0x7912 + 1);
	i_single(s, 0x7919);
	i_single(s, 0x7926);
	i_range(s, 0x792A, 0x792C + 1);
	i_single(s, 0x7930);
	i_single(s, 0x793A);
	i_single(s, 0x793C);
	i_single(s, 0x793E);
	i_range(s, 0x7940, 0x7941 + 1);
	i_range(s, 0x7947, 0x7949 + 1);
	i_single(s, 0x7950);
	i_single(s, 0x7953);
	i_range(s, 0x7955, 0x7957 + 1);
	i_single(s, 0x795A);
	i_range(s, 0x795D, 0x7960 + 1);
	i_single(s, 0x7962);
	i_single(s, 0x7965);
	i_single(s, 0x7968);
	i_single(s, 0x796D);
	i_single(s, 0x7977);
	i_single(s, 0x797A);
	i_range(s, 0x797F, 0x7981 + 1);
	i_range(s, 0x7984, 0x7985 + 1);
	i_single(s, 0x798A);
	i_range(s, 0x798D, 0x798F + 1);
	i_single(s, 0x7994);
	i_single(s, 0x799B);
	i_single(s, 0x799D);
	i_range(s, 0x79A6, 0x79A7 + 1);
	i_single(s, 0x79AA);
	i_single(s, 0x79AE);
	i_single(s, 0x79B0);
	i_single(s, 0x79B3);
	i_range(s, 0x79B9, 0x79BA + 1);
	i_range(s, 0x79BD, 0x79C1 + 1);
	i_single(s, 0x79C9);
	i_single(s, 0x79CB);
	i_range(s, 0x79D1, 0x79D2 + 1);
	i_single(s, 0x79D5);
	i_single(s, 0x79D8);
	i_single(s, 0x79DF);
	i_single(s, 0x79E1);
	i_range(s, 0x79E3, 0x79E4 + 1);
	i_range(s, 0x79E6, 0x79E7 + 1);
	i_single(s, 0x79E9);
	i_single(s, 0x79EC);
	i_single(s, 0x79F0);
	i_single(s, 0x79FB);
	i_single(s, 0x7A00);
	i_single(s, 0x7A08);
	i_single(s, 0x7A0B);
	i_range(s, 0x7A0D, 0x7A0E + 1);
	i_single(s, 0x7A14);
	i_range(s, 0x7A17, 0x7A1A + 1);
	i_single(s, 0x7A1C);
	i_range(s, 0x7A1F, 0x7A20 + 1);
	i_single(s, 0x7A2E);
	i_range(s, 0x7A31, 0x7A32 + 1);
	i_single(s, 0x7A37);
	i_range(s, 0x7A3B, 0x7A40 + 1);
	i_range(s, 0x7A42, 0x7A43 + 1);
	i_single(s, 0x7A46);
	i_single(s, 0x7A49);
	i_range(s, 0x7A4D, 0x7A50 + 1);
	i_single(s, 0x7A57);
	i_range(s, 0x7A61, 0x7A63 + 1);
	i_single(s, 0x7A69);
	i_single(s, 0x7A6B);
	i_single(s, 0x7A70);
	i_single(s, 0x7A74);
	i_single(s, 0x7A76);
	i_range(s, 0x7A79, 0x7A7A + 1);
	i_single(s, 0x7A7D);
	i_single(s, 0x7A7F);
	i_single(s, 0x7A81);
	i_range(s, 0x7A83, 0x7A84 + 1);
	i_single(s, 0x7A88);
	i_range(s, 0x7A92, 0x7A93 + 1);
	i_range(s, 0x7A95, 0x7A98 + 1);
	i_single(s, 0x7A9F);
	i_range(s, 0x7AA9, 0x7AAA + 1);
	i_range(s, 0x7AAE, 0x7AB0 + 1);
	i_single(s, 0x7AB6);
	i_single(s, 0x7ABA);
	i_single(s, 0x7ABF);
	i_range(s, 0x7AC3, 0x7AC5 + 1);
	i_range(s, 0x7AC7, 0x7AC8 + 1);
	i_range(s, 0x7ACA, 0x7ACB + 1);
	i_single(s, 0x7ACD);
	i_single(s, 0x7ACF);
	i_range(s, 0x7AD1, 0x7AD3 + 1);
	i_single(s, 0x7AD5);
	i_range(s, 0x7AD9, 0x7ADA + 1);
	i_range(s, 0x7ADC, 0x7ADD + 1);
	i_range(s, 0x7ADF, 0x7AE3 + 1);
	i_range(s, 0x7AE5, 0x7AE7 + 1);
	i_range(s, 0x7AEA, 0x7AEB + 1);
	i_single(s, 0x7AED);
	i_range(s, 0x7AEF, 0x7AF0 + 1);
	i_single(s, 0x7AF6);
	i_range(s, 0x7AF8, 0x7AFA + 1);
	i_single(s, 0x7AFF);
	i_single(s, 0x7B02);
	i_single(s, 0x7B04);
	i_single(s, 0x7B06);
	i_single(s, 0x7B08);
	i_range(s, 0x7B0A, 0x7B0B + 1);
	i_single(s, 0x7B0F);
	i_single(s, 0x7B11);
	i_range(s, 0x7B18, 0x7B19 + 1);
	i_single(s, 0x7B1B);
	i_single(s, 0x7B1E);
	i_single(s, 0x7B20);
	i_range(s, 0x7B25, 0x7B26 + 1);
	i_single(s, 0x7B28);
	i_single(s, 0x7B2C);
	i_single(s, 0x7B33);
	i_range(s, 0x7B35, 0x7B36 + 1);
	i_single(s, 0x7B39);
	i_range(s, 0x7B45, 0x7B46 + 1);
	i_range(s, 0x7B48, 0x7B49 + 1);
	i_range(s, 0x7B4B, 0x7B4D + 1);
	i_range(s, 0x7B4F, 0x7B52 + 1);
	i_single(s, 0x7B54);
	i_single(s, 0x7B56);
	i_single(s, 0x7B5D);
	i_single(s, 0x7B65);
	i_single(s, 0x7B67);
	i_single(s, 0x7B6C);
	i_single(s, 0x7B6E);
	i_range(s, 0x7B70, 0x7B71 + 1);
	i_range(s, 0x7B74, 0x7B75 + 1);
	i_single(s, 0x7B7A);
	i_range(s, 0x7B86, 0x7B87 + 1);
	i_single(s, 0x7B8B);
	i_single(s, 0x7B8D);
	i_single(s, 0x7B8F);
	i_single(s, 0x7B92);
	i_range(s, 0x7B94, 0x7B95 + 1);
	i_range(s, 0x7B97, 0x7B9A + 1);
	i_range(s, 0x7B9C, 0x7B9F + 1);
	i_single(s, 0x7BA1);
	i_single(s, 0x7BAA);
	i_single(s, 0x7BAD);
	i_single(s, 0x7BB1);
	i_single(s, 0x7BB4);
	i_single(s, 0x7BB8);
	i_range(s, 0x7BC0, 0x7BC1 + 1);
	i_single(s, 0x7BC4);
	i_range(s, 0x7BC6, 0x7BC7 + 1);
	i_single(s, 0x7BC9);
	i_range(s, 0x7BCB, 0x7BCC + 1);
	i_single(s, 0x7BCF);
	i_single(s, 0x7BDD);
	i_single(s, 0x7BE0);
	i_range(s, 0x7BE4, 0x7BE6 + 1);
	i_single(s, 0x7BE9);
	i_single(s, 0x7BED);
	i_single(s, 0x7BF3);
	i_range(s, 0x7BF6, 0x7BF7 + 1);
	i_single(s, 0x7C00);
	i_single(s, 0x7C07);
	i_single(s, 0x7C0D);
	i_range(s, 0x7C11, 0x7C14 + 1);
	i_single(s, 0x7C17);
	i_single(s, 0x7C1F);
	i_single(s, 0x7C21);
	i_single(s, 0x7C23);
	i_single(s, 0x7C27);
	i_range(s, 0x7C2A, 0x7C2B + 1);
	i_range(s, 0x7C37, 0x7C38 + 1);
	i_range(s, 0x7C3D, 0x7C40 + 1);
	i_single(s, 0x7C43);
	i_range(s, 0x7C4C, 0x7C4D + 1);
	i_range(s, 0x7C4F, 0x7C50 + 1);
	i_single(s, 0x7C54);
	i_single(s, 0x7C56);
	i_single(s, 0x7C58);
	i_range(s, 0x7C5F, 0x7C60 + 1);
	i_range(s, 0x7C64, 0x7C65 + 1);
	i_single(s, 0x7C6C);
	i_single(s, 0x7C73);
	i_single(s, 0x7C75);
	i_single(s, 0x7C7E);
	i_range(s, 0x7C81, 0x7C83 + 1);
	i_single(s, 0x7C89);
	i_single(s, 0x7C8B);
	i_single(s, 0x7C8D);
	i_single(s, 0x7C90);
	i_single(s, 0x7C92);
	i_single(s, 0x7C95);
	i_range(s, 0x7C97, 0x7C98 + 1);
	i_single(s, 0x7C9B);
	i_single(s, 0x7C9F);
	i_range(s, 0x7CA1, 0x7CA2 + 1);
	i_range(s, 0x7CA4, 0x7CA5 + 1);
	i_range(s, 0x7CA7, 0x7CA8 + 1);
	i_single(s, 0x7CAB);
	i_range(s, 0x7CAD, 0x7CAE + 1);
	i_range(s, 0x7CB1, 0x7CB3 + 1);
	i_single(s, 0x7CB9);
	i_range(s, 0x7CBD, 0x7CBE + 1);
	i_single(s, 0x7CC0);
	i_single(s, 0x7CC2);
	i_single(s, 0x7CC5);
	i_single(s, 0x7CCA);
	i_single(s, 0x7CCE);
	i_single(s, 0x7CD2);
	i_single(s, 0x7CD6);
	i_single(s, 0x7CD8);
	i_single(s, 0x7CDC);
	i_range(s, 0x7CDE, 0x7CE0 + 1);
	i_single(s, 0x7CE2);
	i_single(s, 0x7CE7);
	i_single(s, 0x7CEF);
	i_single(s, 0x7CF2);
	i_single(s, 0x7CF4);
	i_single(s, 0x7CF6);
	i_single(s, 0x7CF8);
	i_range(s, 0x7CFA, 0x7CFB + 1);
	i_single(s, 0x7CFE);
	i_single(s, 0x7D00);
	i_single(s, 0x7D02);
	i_range(s, 0x7D04, 0x7D06 + 1);
	i_range(s, 0x7D0A, 0x7D0B + 1);
	i_single(s, 0x7D0D);
	i_single(s, 0x7D10);
	i_range(s, 0x7D14, 0x7D15 + 1);
	i_range(s, 0x7D17, 0x7D1C + 1);
	i_range(s, 0x7D20, 0x7D22 + 1);
	i_range(s, 0x7D2B, 0x7D2C + 1);
	i_range(s, 0x7D2E, 0x7D30 + 1);
	i_range(s, 0x7D32, 0x7D33 + 1);
	i_single(s, 0x7D35);
	i_range(s, 0x7D39, 0x7D3A + 1);
	i_single(s, 0x7D3F);
	i_range(s, 0x7D42, 0x7D46 + 1);
	i_single(s, 0x7D48);
	i_range(s, 0x7D4B, 0x7D4C + 1);
	i_range(s, 0x7D4E, 0x7D50 + 1);
	i_single(s, 0x7D56);
	i_range(s, 0x7D5B, 0x7D5C + 1);
	i_single(s, 0x7D5E);
	i_range(s, 0x7D61, 0x7D63 + 1);
	i_single(s, 0x7D66);
	i_single(s, 0x7D68);
	i_single(s, 0x7D6E);
	i_range(s, 0x7D71, 0x7D73 + 1);
	i_range(s, 0x7D75, 0x7D76 + 1);
	i_single(s, 0x7D79);
	i_single(s, 0x7D7D);
	i_single(s, 0x7D89);
	i_single(s, 0x7D8F);
	i_single(s, 0x7D93);
	i_range(s, 0x7D99, 0x7D9C + 1);
	i_range(s, 0x7D9F, 0x7DA0 + 1);
	i_range(s, 0x7DA2, 0x7DA3 + 1);
	i_range(s, 0x7DAB, 0x7DB2 + 1);
	i_range(s, 0x7DB4, 0x7DB5 + 1);
	i_range(s, 0x7DB7, 0x7DB8 + 1);
	i_range(s, 0x7DBA, 0x7DBB + 1);
	i_range(s, 0x7DBD, 0x7DBF + 1);
	i_single(s, 0x7DC7);
	i_range(s, 0x7DCA, 0x7DCB + 1);
	i_single(s, 0x7DCF);
	i_range(s, 0x7DD1, 0x7DD2 + 1);
	i_range(s, 0x7DD5, 0x7DD6 + 1);
	i_single(s, 0x7DD8);
	i_single(s, 0x7DDA);
	i_range(s, 0x7DDC, 0x7DDE + 1);
	i_range(s, 0x7DE0, 0x7DE1 + 1);
	i_single(s, 0x7DE4);
	i_range(s, 0x7DE8, 0x7DE9 + 1);
	i_single(s, 0x7DEC);
	i_single(s, 0x7DEF);
	i_single(s, 0x7DF2);
	i_single(s, 0x7DF4);
	i_single(s, 0x7DFB);
	i_single(s, 0x7E01);
	i_range(s, 0x7E04, 0x7E05 + 1);
	i_range(s, 0x7E09, 0x7E0B + 1);
	i_single(s, 0x7E12);
	i_single(s, 0x7E1B);
	i_range(s, 0x7E1E, 0x7E1F + 1);
	i_range(s, 0x7E21, 0x7E23 + 1);
	i_single(s, 0x7E26);
	i_single(s, 0x7E2B);
	i_single(s, 0x7E2E);
	i_range(s, 0x7E31, 0x7E32 + 1);
	i_single(s, 0x7E35);
	i_single(s, 0x7E37);
	i_range(s, 0x7E39, 0x7E3B + 1);
	i_range(s, 0x7E3D, 0x7E3E + 1);
	i_single(s, 0x7E41);
	i_single(s, 0x7E43);
	i_single(s, 0x7E46);
	i_range(s, 0x7E4A, 0x7E4B + 1);
	i_single(s, 0x7E4D);
	i_single(s, 0x7E52);
	i_range(s, 0x7E54, 0x7E56 + 1);
	i_range(s, 0x7E59, 0x7E5A + 1);
	i_range(s, 0x7E5D, 0x7E5E + 1);
	i_range(s, 0x7E66, 0x7E67 + 1);
	i_range(s, 0x7E69, 0x7E6A + 1);
	i_single(s, 0x7E6D);
	i_single(s, 0x7E70);
	i_single(s, 0x7E79);
	i_range(s, 0x7E7B, 0x7E7D + 1);
	i_single(s, 0x7E7F);
	i_range(s, 0x7E82, 0x7E83 + 1);
	i_range(s, 0x7E88, 0x7E8A + 1);
	i_single(s, 0x7E8C);
	i_range(s, 0x7E8E, 0x7E90 + 1);
	i_range(s, 0x7E92, 0x7E94 + 1);
	i_single(s, 0x7E96);
	i_range(s, 0x7E9B, 0x7E9C + 1);
	i_single(s, 0x7F36);
	i_single(s, 0x7F38);
	i_single(s, 0x7F3A);
	i_single(s, 0x7F45);
	i_single(s, 0x7F47);
	i_range(s, 0x7F4C, 0x7F4E + 1);
	i_range(s, 0x7F50, 0x7F51 + 1);
	i_range(s, 0x7F54, 0x7F55 + 1);
	i_single(s, 0x7F58);
	i_range(s, 0x7F5F, 0x7F60 + 1);
	i_range(s, 0x7F67, 0x7F6B + 1);
	i_single(s, 0x7F6E);
	i_single(s, 0x7F70);
	i_single(s, 0x7F72);
	i_single(s, 0x7F75);
	i_range(s, 0x7F77, 0x7F79 + 1);
	i_range(s, 0x7F82, 0x7F83 + 1);
	i_range(s, 0x7F85, 0x7F88 + 1);
	i_single(s, 0x7F8A);
	i_single(s, 0x7F8C);
	i_single(s, 0x7F8E);
	i_single(s, 0x7F94);
	i_single(s, 0x7F9A);
	i_range(s, 0x7F9D, 0x7F9E + 1);
	i_single(s, 0x7FA1);
	i_range(s, 0x7FA3, 0x7FA4 + 1);
	i_range(s, 0x7FA8, 0x7FA9 + 1);
	i_range(s, 0x7FAE, 0x7FAF + 1);
	i_single(s, 0x7FB2);
	i_single(s, 0x7FB6);
	i_range(s, 0x7FB8, 0x7FB9 + 1);
	i_single(s, 0x7FBD);
	i_single(s, 0x7FC1);
	i_range(s, 0x7FC5, 0x7FC6 + 1);
	i_single(s, 0x7FCA);
	i_single(s, 0x7FCC);
	i_single(s, 0x7FD2);
	i_range(s, 0x7FD4, 0x7FD5 + 1);
	i_range(s, 0x7FE0, 0x7FE1 + 1);
	i_single(s, 0x7FE6);
	i_single(s, 0x7FE9);
	i_single(s, 0x7FEB);
	i_single(s, 0x7FF0);
	i_single(s, 0x7FF3);
	i_single(s, 0x7FF9);
	i_range(s, 0x7FFB, 0x7FFC + 1);
	i_range(s, 0x8000, 0x8001 + 1);
	i_range(s, 0x8003, 0x8006 + 1);
	i_range(s, 0x800B, 0x800C + 1);
	i_single(s, 0x8010);
	i_single(s, 0x8012);
	i_single(s, 0x8015);
	i_range(s, 0x8017, 0x8019 + 1);
	i_single(s, 0x801C);
	i_single(s, 0x8021);
	i_single(s, 0x8028);
	i_single(s, 0x8033);
	i_single(s, 0x8036);
	i_single(s, 0x803B);
	i_single(s, 0x803D);
	i_single(s, 0x803F);
	i_single(s, 0x8046);
	i_single(s, 0x804A);
	i_single(s, 0x8052);
	i_single(s, 0x8056);
	i_single(s, 0x8058);
	i_single(s, 0x805A);
	i_range(s, 0x805E, 0x805F + 1);
	i_range(s, 0x8061, 0x8062 + 1);
	i_single(s, 0x8068);
	i_range(s, 0x806F, 0x8070 + 1);
	i_range(s, 0x8072, 0x8074 + 1);
	i_range(s, 0x8076, 0x8077 + 1);
	i_single(s, 0x8079);
	i_range(s, 0x807D, 0x807F + 1);
	i_range(s, 0x8084, 0x8087 + 1);
	i_single(s, 0x8089);
	i_range(s, 0x808B, 0x808C + 1);
	i_single(s, 0x8093);
	i_single(s, 0x8096);
	i_single(s, 0x8098);
	i_range(s, 0x809A, 0x809B + 1);
	i_single(s, 0x809D);
	i_range(s, 0x80A1, 0x80A2 + 1);
	i_single(s, 0x80A5);
	i_range(s, 0x80A9, 0x80AA + 1);
	i_range(s, 0x80AC, 0x80AD + 1);
	i_single(s, 0x80AF);
	i_range(s, 0x80B1, 0x80B2 + 1);
	i_single(s, 0x80B4);
	i_single(s, 0x80BA);
	i_range(s, 0x80C3, 0x80C4 + 1);
	i_single(s, 0x80C6);
	i_single(s, 0x80CC);
	i_single(s, 0x80CE);
	i_single(s, 0x80D6);
	i_range(s, 0x80D9, 0x80DB + 1);
	i_range(s, 0x80DD, 0x80DE + 1);
	i_single(s, 0x80E1);
	i_range(s, 0x80E4, 0x80E5 + 1);
	i_single(s, 0x80EF);
	i_single(s, 0x80F1);
	i_single(s, 0x80F4);
	i_single(s, 0x80F8);
	i_range(s, 0x80FC, 0x80FD + 1);
	i_single(s, 0x8102);
	i_range(s, 0x8105, 0x810A + 1);
	i_range(s, 0x811A, 0x811B + 1);
	i_single(s, 0x8123);
	i_single(s, 0x8129);
	i_single(s, 0x812F);
	i_single(s, 0x8131);
	i_single(s, 0x8133);
	i_single(s, 0x8139);
	i_single(s, 0x813E);
	i_single(s, 0x8146);
	i_single(s, 0x814B);
	i_single(s, 0x814E);
	i_range(s, 0x8150, 0x8151 + 1);
	i_range(s, 0x8153, 0x8155 + 1);
	i_single(s, 0x815F);
	i_range(s, 0x8165, 0x8166 + 1);
	i_single(s, 0x816B);
	i_single(s, 0x816E);
	i_range(s, 0x8170, 0x8171 + 1);
	i_single(s, 0x8174);
	i_range(s, 0x8178, 0x817A + 1);
	i_range(s, 0x817F, 0x8180 + 1);
	i_range(s, 0x8182, 0x8183 + 1);
	i_single(s, 0x8188);
	i_single(s, 0x818A);
	i_single(s, 0x818F);
	i_single(s, 0x8193);
	i_single(s, 0x8195);
	i_single(s, 0x819A);
	i_range(s, 0x819C, 0x819D + 1);
	i_single(s, 0x81A0);
	i_range(s, 0x81A3, 0x81A4 + 1);
	i_range(s, 0x81A8, 0x81A9 + 1);
	i_single(s, 0x81B0);
	i_single(s, 0x81B3);
	i_single(s, 0x81B5);
	i_single(s, 0x81B8);
	i_single(s, 0x81BA);
	i_range(s, 0x81BD, 0x81C0 + 1);
	i_single(s, 0x81C2);
	i_single(s, 0x81C6);
	i_range(s, 0x81C8, 0x81C9 + 1);
	i_single(s, 0x81CD);
	i_single(s, 0x81D1);
	i_single(s, 0x81D3);
	i_range(s, 0x81D8, 0x81DA + 1);
	i_range(s, 0x81DF, 0x81E0 + 1);
	i_single(s, 0x81E3);
	i_single(s, 0x81E5);
	i_range(s, 0x81E7, 0x81E8 + 1);
	i_single(s, 0x81EA);
	i_single(s, 0x81ED);
	i_range(s, 0x81F3, 0x81F4 + 1);
	i_range(s, 0x81FA, 0x81FC + 1);
	i_single(s, 0x81FE);
	i_range(s, 0x8201, 0x8202 + 1);
	i_single(s, 0x8205);
	i_range(s, 0x8207, 0x820A + 1);
	i_range(s, 0x820C, 0x820E + 1);
	i_single(s, 0x8210);
	i_single(s, 0x8212);
	i_range(s, 0x8216, 0x8218 + 1);
	i_range(s, 0x821B, 0x821C + 1);
	i_range(s, 0x821E, 0x821F + 1);
	i_range(s, 0x8229, 0x822C + 1);
	i_single(s, 0x822E);
	i_single(s, 0x8233);
	i_range(s, 0x8235, 0x8239 + 1);
	i_single(s, 0x8240);
	i_single(s, 0x8247);
	i_range(s, 0x8258, 0x825A + 1);
	i_single(s, 0x825D);
	i_single(s, 0x825F);
	i_single(s, 0x8262);
	i_single(s, 0x8264);
	i_single(s, 0x8266);
	i_single(s, 0x8268);
	i_range(s, 0x826A, 0x826B + 1);
	i_range(s, 0x826E, 0x826F + 1);
	i_range(s, 0x8271, 0x8272 + 1);
	i_range(s, 0x8276, 0x8278 + 1);
	i_single(s, 0x827E);
	i_single(s, 0x828B);
	i_single(s, 0x828D);
	i_single(s, 0x8292);
	i_single(s, 0x8299);
	i_single(s, 0x829D);
	i_single(s, 0x829F);
	i_range(s, 0x82A5, 0x82A6 + 1);
	i_range(s, 0x82AB, 0x82AD + 1);
	i_single(s, 0x82AF);
	i_single(s, 0x82B1);
	i_single(s, 0x82B3);
	i_range(s, 0x82B8, 0x82B9 + 1);
	i_single(s, 0x82BB);
	i_single(s, 0x82BD);
	i_single(s, 0x82C5);
	i_range(s, 0x82D1, 0x82D4 + 1);
	i_single(s, 0x82D7);
	i_single(s, 0x82D9);
	i_range(s, 0x82DB, 0x82DC + 1);
	i_range(s, 0x82DE, 0x82DF + 1);
	i_single(s, 0x82E1);
	i_single(s, 0x82E3);
	i_range(s, 0x82E5, 0x82E7 + 1);
	i_single(s, 0x82EB);
	i_single(s, 0x82F1);
	i_range(s, 0x82F3, 0x82F4 + 1);
	i_range(s, 0x82F9, 0x82FB + 1);
	i_range(s, 0x8301, 0x8306 + 1);
	i_single(s, 0x8309);
	i_single(s, 0x830E);
	i_range(s, 0x8316, 0x8318 + 1);
	i_single(s, 0x831C);
	i_single(s, 0x8323);
	i_single(s, 0x8328);
	i_single(s, 0x832B);
	i_single(s, 0x832F);
	i_range(s, 0x8331, 0x8332 + 1);
	i_range(s, 0x8334, 0x8336 + 1);
	i_range(s, 0x8338, 0x8339 + 1);
	i_single(s, 0x8340);
	i_single(s, 0x8345);
	i_range(s, 0x8349, 0x834A + 1);
	i_range(s, 0x834F, 0x8350 + 1);
	i_single(s, 0x8352);
	i_single(s, 0x8358);
	i_single(s, 0x8362);
	i_single(s, 0x8373);
	i_single(s, 0x8375);
	i_single(s, 0x8377);
	i_range(s, 0x837B, 0x837C + 1);
	i_single(s, 0x837F);
	i_single(s, 0x8385);
	i_single(s, 0x8387);
	i_range(s, 0x8389, 0x838A + 1);
	i_single(s, 0x838E);
	i_single(s, 0x8393);
	i_single(s, 0x8396);
	i_single(s, 0x839A);
	i_range(s, 0x839E, 0x83A0 + 1);
	i_single(s, 0x83A2);
	i_single(s, 0x83A8);
	i_range(s, 0x83AA, 0x83AB + 1);
	i_single(s, 0x83B1);
	i_single(s, 0x83B5);
	i_single(s, 0x83BD);
	i_single(s, 0x83C1);
	i_single(s, 0x83C5);
	i_single(s, 0x83C7);
	i_single(s, 0x83CA);
	i_single(s, 0x83CC);
	i_single(s, 0x83CE);
	i_single(s, 0x83D3);
	i_single(s, 0x83D6);
	i_single(s, 0x83D8);
	i_single(s, 0x83DC);
	i_range(s, 0x83DF, 0x83E0 + 1);
	i_single(s, 0x83E9);
	i_single(s, 0x83EB);
	i_range(s, 0x83EF, 0x83F2 + 1);
	i_single(s, 0x83F4);
	i_range(s, 0x83F6, 0x83F7 + 1);
	i_single(s, 0x83FB);
	i_single(s, 0x83FD);
	i_range(s, 0x8403, 0x8404 + 1);
	i_single(s, 0x8407);
	i_range(s, 0x840B, 0x840E + 1);
	i_single(s, 0x8413);
	i_single(s, 0x8420);
	i_single(s, 0x8422);
	i_range(s, 0x8429, 0x842A + 1);
	i_single(s, 0x842C);
	i_single(s, 0x8431);
	i_single(s, 0x8435);
	i_single(s, 0x8438);
	i_range(s, 0x843C, 0x843D + 1);
	i_single(s, 0x8446);
	i_range(s, 0x8448, 0x8449 + 1);
	i_single(s, 0x844E);
	i_single(s, 0x8457);
	i_single(s, 0x845B);
	i_range(s, 0x8461, 0x8463 + 1);
	i_single(s, 0x8466);
	i_single(s, 0x8469);
	i_range(s, 0x846B, 0x846F + 1);
	i_single(s, 0x8471);
	i_single(s, 0x8475);
	i_single(s, 0x8477);
	i_range(s, 0x8479, 0x847A + 1);
	i_single(s, 0x8482);
	i_single(s, 0x8484);
	i_single(s, 0x848B);
	i_single(s, 0x8490);
	i_single(s, 0x8494);
	i_single(s, 0x8499);
	i_single(s, 0x849C);
	i_single(s, 0x849F);
	i_single(s, 0x84A1);
	i_single(s, 0x84AD);
	i_single(s, 0x84B2);
	i_single(s, 0x84B4);
	i_range(s, 0x84B8, 0x84B9 + 1);
	i_range(s, 0x84BB, 0x84BC + 1);
	i_single(s, 0x84BF);
	i_single(s, 0x84C1);
	i_single(s, 0x84C4);
	i_single(s, 0x84C6);
	i_range(s, 0x84C9, 0x84CB + 1);
	i_single(s, 0x84CD);
	i_range(s, 0x84D0, 0x84D1 + 1);
	i_single(s, 0x84D6);
	i_range(s, 0x84D9, 0x84DA + 1);
	i_single(s, 0x84DC);
	i_single(s, 0x84EC);
	i_single(s, 0x84EE);
	i_single(s, 0x84F4);
	i_single(s, 0x84FC);
	i_range(s, 0x84FF, 0x8500 + 1);
	i_single(s, 0x8506);
	i_single(s, 0x8511);
	i_range(s, 0x8513, 0x8515 + 1);
	i_range(s, 0x8517, 0x8518 + 1);
	i_single(s, 0x851A);
	i_single(s, 0x851F);
	i_single(s, 0x8521);
	i_single(s, 0x8526);
	i_range(s, 0x852C, 0x852D + 1);
	i_single(s, 0x8535);
	i_single(s, 0x853D);
	i_range(s, 0x8540, 0x8541 + 1);
	i_single(s, 0x8543);
	i_range(s, 0x8548, 0x854B + 1);
	i_single(s, 0x854E);
	i_single(s, 0x8553);
	i_single(s, 0x8555);
	i_range(s, 0x8557, 0x855A + 1);
	i_single(s, 0x8563);
	i_range(s, 0x8568, 0x856B + 1);
	i_single(s, 0x856D);
	i_single(s, 0x8577);
	i_single(s, 0x857E);
	i_single(s, 0x8580);
	i_single(s, 0x8584);
	i_range(s, 0x8587, 0x8588 + 1);
	i_single(s, 0x858A);
	i_range(s, 0x8590, 0x8591 + 1);
	i_single(s, 0x8594);
	i_single(s, 0x8597);
	i_single(s, 0x8599);
	i_range(s, 0x859B, 0x859C + 1);
	i_single(s, 0x85A4);
	i_single(s, 0x85A6);
	i_range(s, 0x85A8, 0x85AC + 1);
	i_range(s, 0x85AE, 0x85B0 + 1);
	i_range(s, 0x85B9, 0x85BA + 1);
	i_single(s, 0x85C1);
	i_single(s, 0x85C9);
	i_single(s, 0x85CD);
	i_range(s, 0x85CF, 0x85D0 + 1);
	i_single(s, 0x85D5);
	i_range(s, 0x85DC, 0x85DD + 1);
	i_range(s, 0x85E4, 0x85E5 + 1);
	i_range(s, 0x85E9, 0x85EA + 1);
	i_single(s, 0x85F7);
	i_range(s, 0x85F9, 0x85FB + 1);
	i_single(s, 0x85FE);
	i_single(s, 0x8602);
	i_range(s, 0x8606, 0x8607 + 1);
	i_range(s, 0x860A, 0x860B + 1);
	i_single(s, 0x8613);
	i_range(s, 0x8616, 0x8617 + 1);
	i_single(s, 0x861A);
	i_single(s, 0x8622);
	i_single(s, 0x862D);
	i_range(s, 0x862F, 0x8630 + 1);
	i_single(s, 0x863F);
	i_range(s, 0x864D, 0x864E + 1);
	i_single(s, 0x8650);
	i_range(s, 0x8654, 0x8655 + 1);
	i_single(s, 0x865A);
	i_single(s, 0x865C);
	i_range(s, 0x865E, 0x865F + 1);
	i_single(s, 0x8667);
	i_single(s, 0x866B);
	i_single(s, 0x8671);
	i_single(s, 0x8679);
	i_single(s, 0x867B);
	i_range(s, 0x868A, 0x868C + 1);
	i_single(s, 0x8693);
	i_single(s, 0x8695);
	i_range(s, 0x86A3, 0x86A4 + 1);
	i_range(s, 0x86A9, 0x86AB + 1);
	i_range(s, 0x86AF, 0x86B0 + 1);
	i_single(s, 0x86B6);
	i_single(s, 0x86C4);
	i_range(s, 0x86C6, 0x86C7 + 1);
	i_single(s, 0x86C9);
	i_single(s, 0x86CB);
	i_range(s, 0x86CD, 0x86CE + 1);
	i_single(s, 0x86D4);
	i_single(s, 0x86D9);
	i_single(s, 0x86DB);
	i_range(s, 0x86DE, 0x86DF + 1);
	i_single(s, 0x86E4);
	i_single(s, 0x86E9);
	i_range(s, 0x86EC, 0x86EF + 1);
	i_range(s, 0x86F8, 0x86F9 + 1);
	i_single(s, 0x86FB);
	i_single(s, 0x86FE);
	i_single(s, 0x8700);
	i_range(s, 0x8702, 0x8703 + 1);
	i_single(s, 0x8706);
	i_range(s, 0x8708, 0x870A + 1);
	i_single(s, 0x870D);
	i_range(s, 0x8711, 0x8712 + 1);
	i_single(s, 0x8718);
	i_single(s, 0x871A);
	i_single(s, 0x871C);
	i_single(s, 0x8725);
	i_single(s, 0x8729);
	i_single(s, 0x8734);
	i_single(s, 0x8737);
	i_single(s, 0x873B);
	i_single(s, 0x873F);
	i_single(s, 0x8749);
	i_range(s, 0x874B, 0x874C + 1);
	i_single(s, 0x874E);
	i_single(s, 0x8753);
	i_single(s, 0x8755);
	i_single(s, 0x8757);
	i_single(s, 0x8759);
	i_range(s, 0x875F, 0x8760 + 1);
	i_single(s, 0x8763);
	i_single(s, 0x8766);
	i_single(s, 0x8768);
	i_single(s, 0x876A);
	i_single(s, 0x876E);
	i_single(s, 0x8774);
	i_single(s, 0x8776);
	i_single(s, 0x8778);
	i_single(s, 0x877F);
	i_single(s, 0x8782);
	i_single(s, 0x878D);
	i_single(s, 0x879F);
	i_single(s, 0x87A2);
	i_single(s, 0x87AB);
	i_single(s, 0x87AF);
	i_single(s, 0x87B3);
	i_range(s, 0x87BA, 0x87BB + 1);
	i_single(s, 0x87BD);
	i_single(s, 0x87C0);
	i_single(s, 0x87C4);
	i_range(s, 0x87C6, 0x87C7 + 1);
	i_single(s, 0x87CB);
	i_single(s, 0x87D0);
	i_single(s, 0x87D2);
	i_single(s, 0x87E0);
	i_single(s, 0x87EF);
	i_single(s, 0x87F2);
	i_range(s, 0x87F6, 0x87F7 + 1);
	i_single(s, 0x87F9);
	i_single(s, 0x87FB);
	i_single(s, 0x87FE);
	i_single(s, 0x8805);
	i_single(s, 0x8807);
	i_range(s, 0x880D, 0x880F + 1);
	i_single(s, 0x8811);
	i_range(s, 0x8815, 0x8816 + 1);
	i_range(s, 0x8821, 0x8823 + 1);
	i_single(s, 0x8827);
	i_single(s, 0x8831);
	i_single(s, 0x8836);
	i_single(s, 0x8839);
	i_single(s, 0x883B);
	i_single(s, 0x8840);
	i_single(s, 0x8842);
	i_single(s, 0x8844);
	i_single(s, 0x8846);
	i_range(s, 0x884C, 0x884D + 1);
	i_range(s, 0x8852, 0x8853 + 1);
	i_single(s, 0x8857);
	i_single(s, 0x8859);
	i_single(s, 0x885B);
	i_range(s, 0x885D, 0x885E + 1);
	i_range(s, 0x8861, 0x8863 + 1);
	i_single(s, 0x8868);
	i_single(s, 0x886B);
	i_single(s, 0x8870);
	i_single(s, 0x8872);
	i_single(s, 0x8875);
	i_single(s, 0x8877);
	i_range(s, 0x887D, 0x887F + 1);
	i_range(s, 0x8881, 0x8882 + 1);
	i_single(s, 0x8888);
	i_single(s, 0x888B);
	i_single(s, 0x888D);
	i_single(s, 0x8892);
	i_range(s, 0x8896, 0x8897 + 1);
	i_single(s, 0x8899);
	i_single(s, 0x889E);
	i_single(s, 0x88A2);
	i_single(s, 0x88A4);
	i_single(s, 0x88AB);
	i_single(s, 0x88AE);
	i_range(s, 0x88B0, 0x88B1 + 1);
	i_range(s, 0x88B4, 0x88B5 + 1);
	i_single(s, 0x88B7);
	i_single(s, 0x88BF);
	i_range(s, 0x88C1, 0x88C5 + 1);
	i_single(s, 0x88CF);
	i_range(s, 0x88D4, 0x88D5 + 1);
	i_range(s, 0x88D8, 0x88D9 + 1);
	i_range(s, 0x88DC, 0x88DD + 1);
	i_single(s, 0x88DF);
	i_single(s, 0x88E1);
	i_single(s, 0x88E8);
	i_range(s, 0x88F2, 0x88F5 + 1);
	i_range(s, 0x88F8, 0x88F9 + 1);
	i_range(s, 0x88FC, 0x88FE + 1);
	i_single(s, 0x8902);
	i_single(s, 0x8904);
	i_single(s, 0x8907);
	i_single(s, 0x890A);
	i_single(s, 0x890C);
	i_single(s, 0x8910);
	i_range(s, 0x8912, 0x8913 + 1);
	i_range(s, 0x891C, 0x891E + 1);
	i_single(s, 0x8925);
	i_range(s, 0x892A, 0x892B + 1);
	i_single(s, 0x8936);
	i_single(s, 0x8938);
	i_single(s, 0x893B);
	i_single(s, 0x8941);
	i_range(s, 0x8943, 0x8944 + 1);
	i_range(s, 0x894C, 0x894D + 1);
	i_single(s, 0x8956);
	i_range(s, 0x895E, 0x8960 + 1);
	i_single(s, 0x8964);
	i_single(s, 0x8966);
	i_single(s, 0x896A);
	i_single(s, 0x896D);
	i_single(s, 0x896F);
	i_single(s, 0x8972);
	i_single(s, 0x8974);
	i_single(s, 0x8977);
	i_range(s, 0x897E, 0x897F + 1);
	i_single(s, 0x8981);
	i_single(s, 0x8983);
	i_range(s, 0x8986, 0x8988 + 1);
	i_range(s, 0x898A, 0x898B + 1);
	i_single(s, 0x898F);
	i_single(s, 0x8993);
	i_range(s, 0x8996, 0x8998 + 1);
	i_single(s, 0x899A);
	i_single(s, 0x89A1);
	i_range(s, 0x89A6, 0x89A7 + 1);
	i_range(s, 0x89A9, 0x89AA + 1);
	i_single(s, 0x89AC);
	i_single(s, 0x89AF);
	i_range(s, 0x89B2, 0x89B3 + 1);
	i_single(s, 0x89BA);
	i_single(s, 0x89BD);
	i_range(s, 0x89BF, 0x89C0 + 1);
	i_single(s, 0x89D2);
	i_single(s, 0x89DA);
	i_range(s, 0x89DC, 0x89DD + 1);
	i_single(s, 0x89E3);
	i_range(s, 0x89E6, 0x89E7 + 1);
	i_single(s, 0x89F4);
	i_single(s, 0x89F8);
	i_single(s, 0x8A00);
	i_range(s, 0x8A02, 0x8A03 + 1);
	i_single(s, 0x8A08);
	i_single(s, 0x8A0A);
	i_single(s, 0x8A0C);
	i_single(s, 0x8A0E);
	i_single(s, 0x8A10);
	i_range(s, 0x8A12, 0x8A13 + 1);
	i_range(s, 0x8A16, 0x8A18 + 1);
	i_single(s, 0x8A1B);
	i_single(s, 0x8A1D);
	i_single(s, 0x8A1F);
	i_single(s, 0x8A23);
	i_single(s, 0x8A25);
	i_single(s, 0x8A2A);
	i_single(s, 0x8A2D);
	i_single(s, 0x8A31);
	i_range(s, 0x8A33, 0x8A34 + 1);
	i_range(s, 0x8A36, 0x8A37 + 1);
	i_range(s, 0x8A3A, 0x8A3C + 1);
	i_single(s, 0x8A41);
	i_single(s, 0x8A46);
	i_single(s, 0x8A48);
	i_range(s, 0x8A50, 0x8A52 + 1);
	i_range(s, 0x8A54, 0x8A55 + 1);
	i_single(s, 0x8A5B);
	i_single(s, 0x8A5E);
	i_single(s, 0x8A60);
	i_range(s, 0x8A62, 0x8A63 + 1);
	i_single(s, 0x8A66);
	i_single(s, 0x8A69);
	i_range(s, 0x8A6B, 0x8A6E + 1);
	i_range(s, 0x8A70, 0x8A73 + 1);
	i_single(s, 0x8A79);
	i_single(s, 0x8A7C);
	i_single(s, 0x8A82);
	i_range(s, 0x8A84, 0x8A85 + 1);
	i_single(s, 0x8A87);
	i_single(s, 0x8A89);
	i_range(s, 0x8A8C, 0x8A8D + 1);
	i_single(s, 0x8A91);
	i_single(s, 0x8A93);
	i_single(s, 0x8A95);
	i_single(s, 0x8A98);
	i_single(s, 0x8A9A);
	i_single(s, 0x8A9E);
	i_range(s, 0x8AA0, 0x8AA1 + 1);
	i_range(s, 0x8AA3, 0x8AA8 + 1);
	i_range(s, 0x8AAC, 0x8AAD + 1);
	i_single(s, 0x8AB0);
	i_single(s, 0x8AB2);
	i_single(s, 0x8AB9);
	i_single(s, 0x8ABC);
	i_range(s, 0x8ABE, 0x8ABF + 1);
	i_single(s, 0x8AC2);
	i_single(s, 0x8AC4);
	i_single(s, 0x8AC7);
	i_range(s, 0x8ACB, 0x8ACD + 1);
	i_single(s, 0x8ACF);
	i_single(s, 0x8AD2);
	i_single(s, 0x8AD6);
	i_range(s, 0x8ADA, 0x8ADC + 1);
	i_range(s, 0x8ADE, 0x8AE2 + 1);
	i_single(s, 0x8AE4);
	i_range(s, 0x8AE6, 0x8AE7 + 1);
	i_single(s, 0x8AEB);
	i_range(s, 0x8AED, 0x8AEE + 1);
	i_single(s, 0x8AF1);
	i_single(s, 0x8AF3);
	i_range(s, 0x8AF6, 0x8AF8 + 1);
	i_single(s, 0x8AFA);
	i_single(s, 0x8AFE);
	i_range(s, 0x8B00, 0x8B02 + 1);
	i_single(s, 0x8B04);
	i_single(s, 0x8B07);
	i_single(s, 0x8B0C);
	i_single(s, 0x8B0E);
	i_single(s, 0x8B10);
	i_single(s, 0x8B14);
	i_range(s, 0x8B16, 0x8B17 + 1);
	i_range(s, 0x8B19, 0x8B1B + 1);
	i_single(s, 0x8B1D);
	i_range(s, 0x8B20, 0x8B21 + 1);
	i_single(s, 0x8B26);
	i_single(s, 0x8B28);
	i_range(s, 0x8B2B, 0x8B2C + 1);
	i_single(s, 0x8B33);
	i_single(s, 0x8B39);
	i_single(s, 0x8B3E);
	i_single(s, 0x8B41);
	i_single(s, 0x8B49);
	i_single(s, 0x8B4C);
	i_range(s, 0x8B4E, 0x8B4F + 1);
	i_single(s, 0x8B53);
	i_single(s, 0x8B56);
	i_single(s, 0x8B58);
	i_range(s, 0x8B5A, 0x8B5C + 1);
	i_single(s, 0x8B5F);
	i_single(s, 0x8B66);
	i_range(s, 0x8B6B, 0x8B6C + 1);
	i_range(s, 0x8B6F, 0x8B72 + 1);
	i_single(s, 0x8B74);
	i_single(s, 0x8B77);
	i_single(s, 0x8B7D);
	i_range(s, 0x8B7F, 0x8B80 + 1);
	i_single(s, 0x8B83);
	i_single(s, 0x8B8A);
	i_single(s, 0x8B8C);
	i_single(s, 0x8B8E);
	i_single(s, 0x8B90);
	i_range(s, 0x8B92, 0x8B93 + 1);
	i_single(s, 0x8B96);
	i_range(s, 0x8B99, 0x8B9A + 1);
	i_single(s, 0x8C37);
	i_single(s, 0x8C3A);
	i_single(s, 0x8C3F);
	i_single(s, 0x8C41);
	i_single(s, 0x8C46);
	i_single(s, 0x8C48);
	i_single(s, 0x8C4A);
	i_single(s, 0x8C4C);
	i_single(s, 0x8C4E);
	i_single(s, 0x8C50);
	i_single(s, 0x8C55);
	i_single(s, 0x8C5A);
	i_range(s, 0x8C61, 0x8C62 + 1);
	i_range(s, 0x8C6A, 0x8C6C + 1);
	i_range(s, 0x8C78, 0x8C7A + 1);
	i_single(s, 0x8C7C);
	i_single(s, 0x8C82);
	i_single(s, 0x8C85);
	i_range(s, 0x8C89, 0x8C8A + 1);
	i_range(s, 0x8C8C, 0x8C8E + 1);
	i_single(s, 0x8C94);
	i_single(s, 0x8C98);
	i_range(s, 0x8C9D, 0x8C9E + 1);
	i_range(s, 0x8CA0, 0x8CA2 + 1);
	i_range(s, 0x8CA7, 0x8CB0 + 1);
	i_range(s, 0x8CB2, 0x8CB4 + 1);
	i_range(s, 0x8CB6, 0x8CB8 + 1);
	i_range(s, 0x8CBB, 0x8CBD + 1);
	i_range(s, 0x8CBF, 0x8CC4 + 1);
	i_range(s, 0x8CC7, 0x8CC8 + 1);
	i_single(s, 0x8CCA);
	i_range(s, 0x8CCD, 0x8CCE + 1);
	i_single(s, 0x8CD1);
	i_single(s, 0x8CD3);
	i_range(s, 0x8CDA, 0x8CDC + 1);
	i_single(s, 0x8CDE);
	i_single(s, 0x8CE0);
	i_range(s, 0x8CE2, 0x8CE4 + 1);
	i_single(s, 0x8CE6);
	i_single(s, 0x8CEA);
	i_single(s, 0x8CED);
	i_single(s, 0x8CF0);
	i_single(s, 0x8CF4);
	i_range(s, 0x8CFA, 0x8CFD + 1);
	i_range(s, 0x8D04, 0x8D05 + 1);
	i_range(s, 0x8D07, 0x8D08 + 1);
	i_range(s, 0x8D0A, 0x8D0B + 1);
	i_single(s, 0x8D0D);
	i_range(s, 0x8D0F, 0x8D10 + 1);
	i_range(s, 0x8D12, 0x8D14 + 1);
	i_single(s, 0x8D16);
	i_single(s, 0x8D64);
	i_range(s, 0x8D66, 0x8D67 + 1);
	i_single(s, 0x8D6B);
	i_single(s, 0x8D6D);
	i_range(s, 0x8D70, 0x8D71 + 1);
	i_range(s, 0x8D73, 0x8D74 + 1);
	i_range(s, 0x8D76, 0x8D77 + 1);
	i_single(s, 0x8D81);
	i_single(s, 0x8D85);
	i_single(s, 0x8D8A);
	i_single(s, 0x8D99);
	i_single(s, 0x8DA3);
	i_single(s, 0x8DA8);
	i_single(s, 0x8DB3);
	i_single(s, 0x8DBA);
	i_single(s, 0x8DBE);
	i_single(s, 0x8DC2);
	i_range(s, 0x8DCB, 0x8DCC + 1);
	i_single(s, 0x8DCF);
	i_single(s, 0x8DD6);
	i_range(s, 0x8DDA, 0x8DDB + 1);
	i_single(s, 0x8DDD);
	i_single(s, 0x8DDF);
	i_single(s, 0x8DE1);
	i_single(s, 0x8DE3);
	i_single(s, 0x8DE8);
	i_range(s, 0x8DEA, 0x8DEB + 1);
	i_single(s, 0x8DEF);
	i_single(s, 0x8DF3);
	i_single(s, 0x8DF5);
	i_single(s, 0x8DFC);
	i_single(s, 0x8DFF);
	i_range(s, 0x8E08, 0x8E0A + 1);
	i_range(s, 0x8E0F, 0x8E10 + 1);
	i_range(s, 0x8E1D, 0x8E1F + 1);
	i_single(s, 0x8E2A);
	i_single(s, 0x8E30);
	i_range(s, 0x8E34, 0x8E35 + 1);
	i_single(s, 0x8E42);
	i_single(s, 0x8E44);
	i_range(s, 0x8E47, 0x8E4A + 1);
	i_single(s, 0x8E4C);
	i_single(s, 0x8E50);
	i_single(s, 0x8E55);
	i_single(s, 0x8E59);
	i_range(s, 0x8E5F, 0x8E60 + 1);
	i_range(s, 0x8E63, 0x8E64 + 1);
	i_single(s, 0x8E72);
	i_single(s, 0x8E74);
	i_single(s, 0x8E76);
	i_single(s, 0x8E7C);
	i_single(s, 0x8E81);
	i_range(s, 0x8E84, 0x8E85 + 1);
	i_single(s, 0x8E87);
	i_range(s, 0x8E8A, 0x8E8B + 1);
	i_single(s, 0x8E8D);
	i_single(s, 0x8E91);
	i_range(s, 0x8E93, 0x8E94 + 1);
	i_single(s, 0x8E99);
	i_single(s, 0x8EA1);
	i_range(s, 0x8EAA, 0x8EAC + 1);
	i_range(s, 0x8EAF, 0x8EB1 + 1);
	i_single(s, 0x8EBE);
	i_range(s, 0x8EC5, 0x8EC6 + 1);
	i_single(s, 0x8EC8);
	i_range(s, 0x8ECA, 0x8ECD + 1);
	i_single(s, 0x8ECF);
	i_single(s, 0x8ED2);
	i_single(s, 0x8EDB);
	i_single(s, 0x8EDF);
	i_range(s, 0x8EE2, 0x8EE3 + 1);
	i_single(s, 0x8EEB);
	i_single(s, 0x8EF8);
	i_range(s, 0x8EFB, 0x8EFE + 1);
	i_single(s, 0x8F03);
	i_single(s, 0x8F05);
	i_range(s, 0x8F09, 0x8F0A + 1);
	i_single(s, 0x8F0C);
	i_range(s, 0x8F12, 0x8F15 + 1);
	i_single(s, 0x8F19);
	i_range(s, 0x8F1B, 0x8F1D + 1);
	i_single(s, 0x8F1F);
	i_single(s, 0x8F26);
	i_range(s, 0x8F29, 0x8F2A + 1);
	i_single(s, 0x8F2F);
	i_single(s, 0x8F33);
	i_range(s, 0x8F38, 0x8F39 + 1);
	i_single(s, 0x8F3B);
	i_range(s, 0x8F3E, 0x8F3F + 1);
	i_single(s, 0x8F42);
	i_range(s, 0x8F44, 0x8F46 + 1);
	i_single(s, 0x8F49);
	i_range(s, 0x8F4C, 0x8F4E + 1);
	i_single(s, 0x8F57);
	i_single(s, 0x8F5C);
	i_single(s, 0x8F5F);
	i_range(s, 0x8F61, 0x8F64 + 1);
	i_range(s, 0x8F9B, 0x8F9C + 1);
	i_range(s, 0x8F9E, 0x8F9F + 1);
	i_single(s, 0x8FA3);
	i_range(s, 0x8FA7, 0x8FA8 + 1);
	i_range(s, 0x8FAD, 0x8FB2 + 1);
	i_single(s, 0x8FB7);
	i_range(s, 0x8FBA, 0x8FBC + 1);
	i_single(s, 0x8FBF);
	i_single(s, 0x8FC2);
	i_range(s, 0x8FC4, 0x8FC5 + 1);
	i_single(s, 0x8FCE);
	i_single(s, 0x8FD1);
	i_single(s, 0x8FD4);
	i_single(s, 0x8FDA);
	i_single(s, 0x8FE2);
	i_range(s, 0x8FE5, 0x8FE6 + 1);
	i_range(s, 0x8FE9, 0x8FEB + 1);
	i_single(s, 0x8FED);
	i_range(s, 0x8FEF, 0x8FF0 + 1);
	i_single(s, 0x8FF4);
	i_range(s, 0x8FF7, 0x8FFA + 1);
	i_single(s, 0x8FFD);
	i_range(s, 0x9000, 0x9001 + 1);
	i_single(s, 0x9003);
	i_range(s, 0x9005, 0x9006 + 1);
	i_single(s, 0x900B);
	i_range(s, 0x900D, 0x9011 + 1);
	i_range(s, 0x9013, 0x9017 + 1);
	i_range(s, 0x9019, 0x901A + 1);
	i_range(s, 0x901D, 0x9023 + 1);
	i_single(s, 0x9027);
	i_single(s, 0x902E);
	i_range(s, 0x9031, 0x9032 + 1);
	i_range(s, 0x9035, 0x9036 + 1);
	i_range(s, 0x9038, 0x9039 + 1);
	i_single(s, 0x903C);
	i_single(s, 0x903E);
	i_range(s, 0x9041, 0x9042 + 1);
	i_single(s, 0x9045);
	i_single(s, 0x9047);
	i_range(s, 0x9049, 0x904B + 1);
	i_range(s, 0x904D, 0x9056 + 1);
	i_range(s, 0x9058, 0x9059 + 1);
	i_single(s, 0x905C);
	i_single(s, 0x905E);
	i_range(s, 0x9060, 0x9061 + 1);
	i_single(s, 0x9063);
	i_single(s, 0x9065);
	i_range(s, 0x9067, 0x9069 + 1);
	i_range(s, 0x906D, 0x906F + 1);
	i_single(s, 0x9072);
	i_range(s, 0x9075, 0x9078 + 1);
	i_single(s, 0x907A);
	i_range(s, 0x907C, 0x907D + 1);
	i_range(s, 0x907F, 0x9084 + 1);
	i_single(s, 0x9087);
	i_range(s, 0x9089, 0x908A + 1);
	i_single(s, 0x908F);
	i_single(s, 0x9091);
	i_single(s, 0x90A3);
	i_single(s, 0x90A6);
	i_single(s, 0x90A8);
	i_single(s, 0x90AA);
	i_single(s, 0x90AF);
	i_single(s, 0x90B1);
	i_single(s, 0x90B5);
	i_single(s, 0x90B8);
	i_single(s, 0x90C1);
	i_single(s, 0x90CA);
	i_single(s, 0x90CE);
	i_single(s, 0x90DB);
	i_single(s, 0x90DE);
	i_range(s, 0x90E1, 0x90E2 + 1);
	i_single(s, 0x90E4);
	i_single(s, 0x90E8);
	i_single(s, 0x90ED);
	i_single(s, 0x90F5);
	i_single(s, 0x90F7);
	i_single(s, 0x90FD);
	i_single(s, 0x9102);
	i_single(s, 0x9112);
	i_single(s, 0x9115);
	i_single(s, 0x9119);
	i_single(s, 0x9127);
	i_single(s, 0x912D);
	i_single(s, 0x9130);
	i_single(s, 0x9132);
	i_range(s, 0x9149, 0x914E + 1);
	i_single(s, 0x9152);
	i_single(s, 0x9154);
	i_single(s, 0x9156);
	i_single(s, 0x9158);
	i_range(s, 0x9162, 0x9163 + 1);
	i_single(s, 0x9165);
	i_range(s, 0x9169, 0x916A + 1);
	i_single(s, 0x916C);
	i_range(s, 0x9172, 0x9173 + 1);
	i_single(s, 0x9175);
	i_range(s, 0x9177, 0x9178 + 1);
	i_single(s, 0x9182);
	i_single(s, 0x9187);
	i_single(s, 0x9189);
	i_single(s, 0x918B);
	i_single(s, 0x918D);
	i_single(s, 0x9190);
	i_single(s, 0x9192);
	i_single(s, 0x9197);
	i_single(s, 0x919C);
	i_single(s, 0x91A2);
	i_single(s, 0x91A4);
	i_range(s, 0x91AA, 0x91AB + 1);
	i_single(s, 0x91AF);
	i_range(s, 0x91B4, 0x91B5 + 1);
	i_single(s, 0x91B8);
	i_single(s, 0x91BA);
	i_range(s, 0x91C0, 0x91C1 + 1);
	i_range(s, 0x91C6, 0x91C9 + 1);
	i_range(s, 0x91CB, 0x91D1 + 1);
	i_range(s, 0x91D6, 0x91D8 + 1);
	i_range(s, 0x91DA, 0x91DF + 1);
	i_single(s, 0x91E1);
	i_range(s, 0x91E3, 0x91E7 + 1);
	i_range(s, 0x91ED, 0x91EE + 1);
	i_range(s, 0x91F5, 0x91F6 + 1);
	i_single(s, 0x91FC);
	i_single(s, 0x91FF);
	i_single(s, 0x9206);
	i_single(s, 0x920A);
	i_range(s, 0x920D, 0x920E + 1);
	i_range(s, 0x9210, 0x9211 + 1);
	i_range(s, 0x9214, 0x9215 + 1);
	i_single(s, 0x921E);
	i_single(s, 0x9229);
	i_single(s, 0x922C);
	i_single(s, 0x9234);
	i_single(s, 0x9237);
	i_range(s, 0x9239, 0x923A + 1);
	i_single(s, 0x923C);
	i_range(s, 0x923F, 0x9240 + 1);
	i_range(s, 0x9244, 0x9245 + 1);
	i_range(s, 0x9248, 0x9249 + 1);
	i_single(s, 0x924B);
	i_single(s, 0x924E);
	i_range(s, 0x9250, 0x9251 + 1);
	i_single(s, 0x9257);
	i_range(s, 0x9259, 0x925B + 1);
	i_single(s, 0x925E);
	i_single(s, 0x9262);
	i_single(s, 0x9264);
	i_range(s, 0x9266, 0x9267 + 1);
	i_single(s, 0x9271);
	i_range(s, 0x9277, 0x9278 + 1);
	i_single(s, 0x927E);
	i_single(s, 0x9280);
	i_single(s, 0x9283);
	i_single(s, 0x9285);
	i_single(s, 0x9288);
	i_single(s, 0x9291);
	i_single(s, 0x9293);
	i_range(s, 0x9295, 0x9296 + 1);
	i_single(s, 0x9298);
	i_range(s, 0x929A, 0x929C + 1);
	i_single(s, 0x92A7);
	i_single(s, 0x92AD);
	i_single(s, 0x92B7);
	i_single(s, 0x92B9);
	i_range(s, 0x92CF, 0x92D0 + 1);
	i_range(s, 0x92D2, 0x92D3 + 1);
	i_single(s, 0x92D5);
	i_single(s, 0x92D7);
	i_single(s, 0x92D9);
	i_single(s, 0x92E0);
	i_single(s, 0x92E4);
	i_single(s, 0x92E7);
	i_range(s, 0x92E9, 0x92EA + 1);
	i_single(s, 0x92ED);
	i_range(s, 0x92F2, 0x92F3 + 1);
	i_range(s, 0x92F8, 0x92FC + 1);
	i_single(s, 0x92FF);
	i_single(s, 0x9302);
	i_single(s, 0x9306);
	i_range(s, 0x930F, 0x9310 + 1);
	i_range(s, 0x9318, 0x931A + 1);
	i_range(s, 0x931D, 0x931E + 1);
	i_range(s, 0x9320, 0x9323 + 1);
	i_range(s, 0x9325, 0x9326 + 1);
	i_single(s, 0x9328);
	i_range(s, 0x932B, 0x932C + 1);
	i_range(s, 0x932E, 0x932F + 1);
	i_single(s, 0x9332);
	i_single(s, 0x9335);
	i_range(s, 0x933A, 0x933B + 1);
	i_single(s, 0x9344);
	i_single(s, 0x9348);
	i_single(s, 0x934B);
	i_single(s, 0x934D);
	i_single(s, 0x9354);
	i_range(s, 0x9356, 0x9357 + 1);
	i_range(s, 0x935B, 0x935C + 1);
	i_single(s, 0x9360);
	i_single(s, 0x936C);
	i_single(s, 0x936E);
	i_single(s, 0x9370);
	i_single(s, 0x9375);
	i_single(s, 0x937C);
	i_single(s, 0x937E);
	i_single(s, 0x938C);
	i_single(s, 0x9394);
	i_range(s, 0x9396, 0x9397 + 1);
	i_single(s, 0x939A);
	i_single(s, 0x93A4);
	i_single(s, 0x93A7);
	i_range(s, 0x93AC, 0x93AE + 1);
	i_single(s, 0x93B0);
	i_single(s, 0x93B9);
	i_single(s, 0x93C3);
	i_single(s, 0x93C6);
	i_single(s, 0x93C8);
	i_range(s, 0x93D0, 0x93D1 + 1);
	i_range(s, 0x93D6, 0x93D8 + 1);
	i_range(s, 0x93DD, 0x93DE + 1);
	i_single(s, 0x93E1);
	i_range(s, 0x93E4, 0x93E5 + 1);
	i_single(s, 0x93E8);
	i_single(s, 0x93F8);
	i_single(s, 0x9403);
	i_single(s, 0x9407);
	i_single(s, 0x9410);
	i_range(s, 0x9413, 0x9414 + 1);
	i_range(s, 0x9418, 0x941A + 1);
	i_single(s, 0x9421);
	i_single(s, 0x942B);
	i_single(s, 0x9431);
	i_range(s, 0x9435, 0x9436 + 1);
	i_single(s, 0x9438);
	i_single(s, 0x943A);
	i_single(s, 0x9441);
	i_range(s, 0x9444, 0x9445 + 1);
	i_single(s, 0x9448);
	i_range(s, 0x9451, 0x9453 + 1);
	i_range(s, 0x945A, 0x945B + 1);
	i_single(s, 0x945E);
	i_single(s, 0x9460);
	i_single(s, 0x9462);
	i_single(s, 0x946A);
	i_single(s, 0x9470);
	i_single(s, 0x9475);
	i_single(s, 0x9477);
	i_range(s, 0x947C, 0x947F + 1);
	i_single(s, 0x9481);
	i_single(s, 0x9577);
	i_single(s, 0x9580);
	i_range(s, 0x9582, 0x9583 + 1);
	i_single(s, 0x9587);
	i_range(s, 0x9589, 0x958B + 1);
	i_single(s, 0x958F);
	i_range(s, 0x9591, 0x9594 + 1);
	i_single(s, 0x9596);
	i_range(s, 0x9598, 0x9599 + 1);
	i_single(s, 0x95A0);
	i_range(s, 0x95A2, 0x95A5 + 1);
	i_range(s, 0x95A7, 0x95A8 + 1);
	i_single(s, 0x95AD);
	i_single(s, 0x95B2);
	i_single(s, 0x95B9);
	i_range(s, 0x95BB, 0x95BC + 1);
	i_single(s, 0x95BE);
	i_single(s, 0x95C3);
	i_single(s, 0x95C7);
	i_single(s, 0x95CA);
	i_range(s, 0x95CC, 0x95CD + 1);
	i_range(s, 0x95D4, 0x95D6 + 1);
	i_single(s, 0x95D8);
	i_single(s, 0x95DC);
	i_range(s, 0x95E1, 0x95E2 + 1);
	i_single(s, 0x95E5);
	i_single(s, 0x961C);
	i_single(s, 0x9621);
	i_single(s, 0x9628);
	i_single(s, 0x962A);
	i_range(s, 0x962E, 0x962F + 1);
	i_single(s, 0x9632);
	i_single(s, 0x963B);
	i_range(s, 0x963F, 0x9640 + 1);
	i_single(s, 0x9642);
	i_single(s, 0x9644);
	i_range(s, 0x964B, 0x964D + 1);
	i_range(s, 0x964F, 0x9650 + 1);
	i_range(s, 0x965B, 0x965F + 1);
	i_range(s, 0x9662, 0x9666 + 1);
	i_single(s, 0x966A);
	i_single(s, 0x966C);
	i_single(s, 0x9670);
	i_range(s, 0x9672, 0x9673 + 1);
	i_range(s, 0x9675, 0x9678 + 1);
	i_single(s, 0x967A);
	i_single(s, 0x967D);
	i_range(s, 0x9685, 0x9686 + 1);
	i_single(s, 0x9688);
	i_range(s, 0x968A, 0x968B + 1);
	i_range(s, 0x968D, 0x968F + 1);
	i_range(s, 0x9694, 0x9695 + 1);
	i_range(s, 0x9697, 0x9699 + 1);
	i_range(s, 0x969B, 0x969D + 1);
	i_single(s, 0x96A0);
	i_single(s, 0x96A3);
	i_range(s, 0x96A7, 0x96A8 + 1);
	i_single(s, 0x96AA);
	i_range(s, 0x96AF, 0x96B2 + 1);
	i_single(s, 0x96B4);
	i_range(s, 0x96B6, 0x96B9 + 1);
	i_range(s, 0x96BB, 0x96BC + 1);
	i_range(s, 0x96C0, 0x96C1 + 1);
	i_range(s, 0x96C4, 0x96C7 + 1);
	i_single(s, 0x96C9);
	i_range(s, 0x96CB, 0x96CE + 1);
	i_single(s, 0x96D1);
	i_range(s, 0x96D5, 0x96D6 + 1);
	i_single(s, 0x96D9);
	i_range(s, 0x96DB, 0x96DC + 1);
	i_range(s, 0x96E2, 0x96E3 + 1);
	i_single(s, 0x96E8);
	i_range(s, 0x96EA, 0x96EB + 1);
	i_single(s, 0x96F0);
	i_single(s, 0x96F2);
	i_range(s, 0x96F6, 0x96F7 + 1);
	i_single(s, 0x96F9);
	i_single(s, 0x96FB);
	i_single(s, 0x9700);
	i_single(s, 0x9704);
	i_range(s, 0x9706, 0x9708 + 1);
	i_single(s, 0x970A);
	i_range(s, 0x970D, 0x970F + 1);
	i_single(s, 0x9711);
	i_single(s, 0x9713);
	i_single(s, 0x9716);
	i_single(s, 0x9719);
	i_single(s, 0x971C);
	i_single(s, 0x971E);
	i_single(s, 0x9724);
	i_single(s, 0x9727);
	i_single(s, 0x972A);
	i_single(s, 0x9730);
	i_range(s, 0x9732, 0x9733 + 1);
	i_range(s, 0x9738, 0x9739 + 1);
	i_single(s, 0x973B);
	i_range(s, 0x973D, 0x973E + 1);
	i_range(s, 0x9742, 0x9744 + 1);
	i_single(s, 0x9746);
	i_range(s, 0x9748, 0x9749 + 1);
	i_single(s, 0x974D);
	i_single(s, 0x974F);
	i_range(s, 0x9751, 0x9752 + 1);
	i_range(s, 0x9755, 0x9756 + 1);
	i_single(s, 0x9759);
	i_single(s, 0x975C);
	i_single(s, 0x975E);
	i_range(s, 0x9760, 0x9762 + 1);
	i_single(s, 0x9764);
	i_single(s, 0x9766);
	i_range(s, 0x9768, 0x9769 + 1);
	i_single(s, 0x976B);
	i_single(s, 0x976D);
	i_single(s, 0x9771);
	i_single(s, 0x9774);
	i_range(s, 0x9779, 0x977A + 1);
	i_single(s, 0x977C);
	i_single(s, 0x9781);
	i_range(s, 0x9784, 0x9786 + 1);
	i_single(s, 0x978B);
	i_single(s, 0x978D);
	i_range(s, 0x978F, 0x9790 + 1);
	i_single(s, 0x9798);
	i_single(s, 0x979C);
	i_single(s, 0x97A0);
	i_single(s, 0x97A3);
	i_single(s, 0x97A6);
	i_single(s, 0x97A8);
	i_single(s, 0x97AB);
	i_single(s, 0x97AD);
	i_range(s, 0x97B3, 0x97B4 + 1);
	i_single(s, 0x97C3);
	i_single(s, 0x97C6);
	i_single(s, 0x97C8);
	i_single(s, 0x97CB);
	i_single(s, 0x97D3);
	i_single(s, 0x97DC);
	i_range(s, 0x97ED, 0x97EE + 1);
	i_range(s, 0x97F2, 0x97F3 + 1);
	i_range(s, 0x97F5, 0x97F6 + 1);
	i_single(s, 0x97FB);
	i_single(s, 0x97FF);
	i_range(s, 0x9801, 0x9803 + 1);
	i_range(s, 0x9805, 0x9806 + 1);
	i_single(s, 0x9808);
	i_single(s, 0x980C);
	i_range(s, 0x980F, 0x9813 + 1);
	i_range(s, 0x9817, 0x9818 + 1);
	i_single(s, 0x981A);
	i_single(s, 0x9821);
	i_single(s, 0x9824);
	i_range(s, 0x982C, 0x982D + 1);
	i_single(s, 0x9834);
	i_range(s, 0x9837, 0x9838 + 1);
	i_range(s, 0x983B, 0x983D + 1);
	i_single(s, 0x9846);
	i_range(s, 0x984B, 0x984F + 1);
	i_range(s, 0x9854, 0x9855 + 1);
	i_range(s, 0x9857, 0x9858 + 1);
	i_single(s, 0x985B);
	i_single(s, 0x985E);
	i_single(s, 0x9865);
	i_single(s, 0x9867);
	i_single(s, 0x986B);
	i_range(s, 0x986F, 0x9871 + 1);
	i_range(s, 0x9873, 0x9874 + 1);
	i_single(s, 0x98A8);
	i_single(s, 0x98AA);
	i_single(s, 0x98AF);
	i_single(s, 0x98B1);
	i_single(s, 0x98B6);
	i_range(s, 0x98C3, 0x98C4 + 1);
	i_single(s, 0x98C6);
	i_range(s, 0x98DB, 0x98DC + 1);
	i_single(s, 0x98DF);
	i_single(s, 0x98E2);
	i_single(s, 0x98E9);
	i_single(s, 0x98EB);
	i_range(s, 0x98ED, 0x98EF + 1);
	i_single(s, 0x98F2);
	i_single(s, 0x98F4);
	i_range(s, 0x98FC, 0x98FE + 1);
	i_single(s, 0x9903);
	i_single(s, 0x9905);
	i_range(s, 0x9909, 0x990A + 1);
	i_single(s, 0x990C);
	i_single(s, 0x9910);
	i_range(s, 0x9912, 0x9914 + 1);
	i_single(s, 0x9918);
	i_range(s, 0x991D, 0x991E + 1);
	i_range(s, 0x9920, 0x9921 + 1);
	i_single(s, 0x9924);
	i_range(s, 0x9927, 0x9928 + 1);
	i_single(s, 0x992C);
	i_single(s, 0x992E);
	i_range(s, 0x993D, 0x993E + 1);
	i_single(s, 0x9942);
	i_single(s, 0x9945);
	i_single(s, 0x9949);
	i_range(s, 0x994B, 0x994C + 1);
	i_range(s, 0x9950, 0x9952 + 1);
	i_single(s, 0x9955);
	i_single(s, 0x9957);
	i_range(s, 0x9996, 0x9999 + 1);
	i_single(s, 0x999E);
	i_single(s, 0x99A5);
	i_single(s, 0x99A8);
	i_range(s, 0x99AC, 0x99AE + 1);
	i_range(s, 0x99B3, 0x99B4 + 1);
	i_single(s, 0x99BC);
	i_single(s, 0x99C1);
	i_range(s, 0x99C4, 0x99C6 + 1);
	i_single(s, 0x99C8);
	i_range(s, 0x99D0, 0x99D2 + 1);
	i_single(s, 0x99D5);
	i_single(s, 0x99D8);
	i_single(s, 0x99DB);
	i_single(s, 0x99DD);
	i_single(s, 0x99DF);
	i_single(s, 0x99E2);
	i_range(s, 0x99ED, 0x99EE + 1);
	i_range(s, 0x99F1, 0x99F2 + 1);
	i_single(s, 0x99F8);
	i_single(s, 0x99FB);
	i_single(s, 0x99FF);
	i_single(s, 0x9A01);
	i_single(s, 0x9A05);
	i_range(s, 0x9A0E, 0x9A0F + 1);
	i_range(s, 0x9A12, 0x9A13 + 1);
	i_single(s, 0x9A19);
	i_single(s, 0x9A28);
	i_single(s, 0x9A2B);
	i_single(s, 0x9A30);
	i_single(s, 0x9A37);
	i_single(s, 0x9A3E);
	i_single(s, 0x9A40);
	i_range(s, 0x9A42, 0x9A43 + 1);
	i_single(s, 0x9A45);
	i_range(s, 0x9A4D, 0x9A4E + 1);
	i_single(s, 0x9A55);
	i_single(s, 0x9A57);
	i_range(s, 0x9A5A, 0x9A5B + 1);
	i_single(s, 0x9A5F);
	i_single(s, 0x9A62);
	i_range(s, 0x9A64, 0x9A65 + 1);
	i_range(s, 0x9A69, 0x9A6B + 1);
	i_single(s, 0x9AA8);
	i_single(s, 0x9AAD);
	i_single(s, 0x9AB0);
	i_single(s, 0x9AB8);
	i_single(s, 0x9ABC);
	i_single(s, 0x9AC0);
	i_single(s, 0x9AC4);
	i_single(s, 0x9ACF);
	i_single(s, 0x9AD1);
	i_range(s, 0x9AD3, 0x9AD4 + 1);
	i_range(s, 0x9AD8, 0x9AD9 + 1);
	i_single(s, 0x9ADC);
	i_range(s, 0x9ADE, 0x9ADF + 1);
	i_range(s, 0x9AE2, 0x9AE3 + 1);
	i_single(s, 0x9AE6);
	i_range(s, 0x9AEA, 0x9AEB + 1);
	i_range(s, 0x9AED, 0x9AEF + 1);
	i_single(s, 0x9AF1);
	i_single(s, 0x9AF4);
	i_single(s, 0x9AF7);
	i_single(s, 0x9AFB);
	i_single(s, 0x9B06);
	i_single(s, 0x9B18);
	i_single(s, 0x9B1A);
	i_single(s, 0x9B1F);
	i_range(s, 0x9B22, 0x9B23 + 1);
	i_single(s, 0x9B25);
	i_range(s, 0x9B27, 0x9B2A + 1);
	i_range(s, 0x9B2E, 0x9B2F + 1);
	i_range(s, 0x9B31, 0x9B32 + 1);
	i_range(s, 0x9B3B, 0x9B3C + 1);
	i_range(s, 0x9B41, 0x9B45 + 1);
	i_range(s, 0x9B4D, 0x9B4F + 1);
	i_single(s, 0x9B51);
	i_single(s, 0x9B54);
	i_single(s, 0x9B58);
	i_single(s, 0x9B5A);
	i_single(s, 0x9B6F);
	i_single(s, 0x9B72);
	i_range(s, 0x9B74, 0x9B75 + 1);
	i_single(s, 0x9B83);
	i_range(s, 0x9B8E, 0x9B8F + 1);
	i_range(s, 0x9B91, 0x9B93 + 1);
	i_range(s, 0x9B96, 0x9B97 + 1);
	i_range(s, 0x9B9F, 0x9BA0 + 1);
	i_single(s, 0x9BA8);
	i_range(s, 0x9BAA, 0x9BAB + 1);
	i_range(s, 0x9BAD, 0x9BAE + 1);
	i_single(s, 0x9BB1);
	i_single(s, 0x9BB4);
	i_single(s, 0x9BB9);
	i_single(s, 0x9BBB);
	i_single(s, 0x9BC0);
	i_single(s, 0x9BC6);
	i_range(s, 0x9BC9, 0x9BCA + 1);
	i_single(s, 0x9BCF);
	i_range(s, 0x9BD1, 0x9BD2 + 1);
	i_single(s, 0x9BD4);
	i_single(s, 0x9BD6);
	i_single(s, 0x9BDB);
	i_range(s, 0x9BE1, 0x9BE4 + 1);
	i_single(s, 0x9BE8);
	i_range(s, 0x9BF0, 0x9BF2 + 1);
	i_single(s, 0x9BF5);
	i_single(s, 0x9C00);
	i_single(s, 0x9C04);
	i_single(s, 0x9C06);
	i_range(s, 0x9C08, 0x9C0A + 1);
	i_range(s, 0x9C0C, 0x9C0D + 1);
	i_single(s, 0x9C10);
	i_range(s, 0x9C12, 0x9C15 + 1);
	i_single(s, 0x9C1B);
	i_single(s, 0x9C21);
	i_range(s, 0x9C24, 0x9C25 + 1);
	i_range(s, 0x9C2D, 0x9C30 + 1);
	i_single(s, 0x9C32);
	i_range(s, 0x9C39, 0x9C3B + 1);
	i_single(s, 0x9C3E);
	i_range(s, 0x9C46, 0x9C48 + 1);
	i_single(s, 0x9C52);
	i_single(s, 0x9C57);
	i_single(s, 0x9C5A);
	i_single(s, 0x9C60);
	i_single(s, 0x9C67);
	i_single(s, 0x9C76);
	i_single(s, 0x9C78);
	i_single(s, 0x9CE5);
	i_single(s, 0x9CE7);
	i_single(s, 0x9CE9);
	i_range(s, 0x9CEB, 0x9CEC + 1);
	i_single(s, 0x9CF0);
	i_range(s, 0x9CF3, 0x9CF4 + 1);
	i_single(s, 0x9CF6);
	i_single(s, 0x9D03);
	i_range(s, 0x9D06, 0x9D09 + 1);
	i_single(s, 0x9D0E);
	i_single(s, 0x9D12);
	i_single(s, 0x9D15);
	i_single(s, 0x9D1B);
	i_single(s, 0x9D1F);
	i_single(s, 0x9D23);
	i_single(s, 0x9D26);
	i_single(s, 0x9D28);
	i_range(s, 0x9D2A, 0x9D2C + 1);
	i_single(s, 0x9D3B);
	i_range(s, 0x9D3E, 0x9D3F + 1);
	i_single(s, 0x9D41);
	i_single(s, 0x9D44);
	i_single(s, 0x9D46);
	i_single(s, 0x9D48);
	i_range(s, 0x9D50, 0x9D51 + 1);
	i_single(s, 0x9D59);
	i_range(s, 0x9D5C, 0x9D5E + 1);
	i_range(s, 0x9D60, 0x9D61 + 1);
	i_single(s, 0x9D64);
	i_range(s, 0x9D6B, 0x9D6C + 1);
	i_range(s, 0x9D6F, 0x9D70 + 1);
	i_single(s, 0x9D72);
	i_single(s, 0x9D7A);
	i_single(s, 0x9D87);
	i_single(s, 0x9D89);
	i_single(s, 0x9D8F);
	i_single(s, 0x9D9A);
	i_single(s, 0x9DA4);
	i_single(s, 0x9DA9);
	i_single(s, 0x9DAB);
	i_single(s, 0x9DAF);
	i_single(s, 0x9DB2);
	i_single(s, 0x9DB4);
	i_single(s, 0x9DB8);
	i_range(s, 0x9DBA, 0x9DBB + 1);
	i_range(s, 0x9DC1, 0x9DC2 + 1);
	i_single(s, 0x9DC4);
	i_single(s, 0x9DC6);
	i_single(s, 0x9DCF);
	i_single(s, 0x9DD3);
	i_single(s, 0x9DD9);
	i_single(s, 0x9DE6);
	i_single(s, 0x9DED);
	i_single(s, 0x9DEF);
	i_single(s, 0x9DF2);
	i_range(s, 0x9DF8, 0x9DFA + 1);
	i_single(s, 0x9DFD);
	i_range(s, 0x9E19, 0x9E1B + 1);
	i_single(s, 0x9E1E);
	i_single(s, 0x9E75);
	i_range(s, 0x9E78, 0x9E79 + 1);
	i_single(s, 0x9E7D);
	i_single(s, 0x9E7F);
	i_single(s, 0x9E81);
	i_single(s, 0x9E88);
	i_range(s, 0x9E8B, 0x9E8C + 1);
	i_range(s, 0x9E91, 0x9E93 + 1);
	i_single(s, 0x9E95);
	i_single(s, 0x9E97);
	i_single(s, 0x9E9D);
	i_single(s, 0x9E9F);
	i_range(s, 0x9EA5, 0x9EA6 + 1);
	i_range(s, 0x9EA9, 0x9EAA + 1);
	i_single(s, 0x9EAD);
	i_range(s, 0x9EB8, 0x9EBC + 1);
	i_range(s, 0x9EBE, 0x9EBF + 1);
	i_single(s, 0x9EC4);
	i_range(s, 0x9ECC, 0x9ED2 + 1);
	i_single(s, 0x9ED4);
	i_range(s, 0x9ED8, 0x9ED9 + 1);
	i_range(s, 0x9EDB, 0x9EDE + 1);
	i_single(s, 0x9EE0);
	i_single(s, 0x9EE5);
	i_single(s, 0x9EE8);
	i_single(s, 0x9EEF);
	i_single(s, 0x9EF4);
	i_range(s, 0x9EF6, 0x9EF7 + 1);
	i_single(s, 0x9EF9);
	i_range(s, 0x9EFB, 0x9EFD + 1);
	i_range(s, 0x9F07, 0x9F08 + 1);
	i_single(s, 0x9F0E);
	i_single(s, 0x9F13);
	i_single(s, 0x9F15);
	i_range(s, 0x9F20, 0x9F21 + 1);
	i_single(s, 0x9F2C);
	i_single(s, 0x9F3B);
	i_single(s, 0x9F3E);
	i_range(s, 0x9F4A, 0x9F4B + 1);
	i_range(s, 0x9F4E, 0x9F4F + 1);
	i_single(s, 0x9F52);
	i_single(s, 0x9F54);
	i_range(s, 0x9F5F, 0x9F63 + 1);
	i_range(s, 0x9F66, 0x9F67 + 1);
	i_single(s, 0x9F6A);
	i_single(s, 0x9F6C);
	i_single(s, 0x9F72);
	i_range(s, 0x9F76, 0x9F77 + 1);
	i_single(s, 0x9F8D);
	i_single(s, 0x9F95);
	i_range(s, 0x9F9C, 0x9F9D + 1);
	i_single(s, 0x9FA0);
	i_single(s, 0xF929);
	i_single(s, 0xF9DC);
	i_range(s, 0xFA0E, 0xFA2D + 1);
	i_range(s, 0xFF10, 0xFF19 + 1);
	i_range(s, 0xFF61, 0xFF9F + 1);
	i_single(s, 0xFFE5);
}

void unicode_test()
{
	cout << "Using both representations to look up Japanese Unicode characters" << endl;
	cout << "-----------------------------------------------------------------" << endl;
	CPrecisionTimer ct;
	ct.Start();
	insert_Japanese_character_set(&g_set);
	cout << "It took " << ct.Stop() << " seconds to insert all Japanese characters into the std::set" << endl;
	ct.Start();
	insert_Japanese_character_set(&g_rset);
	cout << "It took " << ct.Stop() << " seconds to insert all Japanese characters into the range_set" << endl;

	cout << "Memory used by std::set  (approx): " << g_set.size() * (sizeof(int) + sizeof(int *) * 2) + sizeof(g_set) << endl;
	cout << "Memory used by range_set (approx): " << g_rset.n_ranges() * (sizeof(int) + sizeof(int *) * 2) * 2 + sizeof(g_rset) << endl;

	cout << "Generating 100000 random characters" << endl;
	int i, N = 100000, total = g_set.size();
	std::set<int>::iterator it;
	srand(15351);
	ct.Start();
	for (i = 0; i < N; ++i)
	{
		int o = rand() % total;
		for (it = g_set.begin(); it != g_set.end() && o > 0; ++it)
		{
			--o;
		}
		int ch = *it;
	}
	cout << "Time for std::set: " << ct.Stop() " seconds " << endl;

	range_set<int>::iterator itr;
	srand(15351);
	ct.Start();
	for (i = 0; i < N; ++i)
	{
		itr = g_rset.get_nth((rand() % total) + 1);
		int ch = *itr;
	}
	cout << "Time for range_set: " << ct.Stop() " seconds " << endl;
}

int main()
{
	insertion_speed();
	random_speed();
	unicode_test();
	_getch();
	return 0;
}